/*
 *  uud.c    (a better freeware uudecode program)
 *  V2.1.3
 *  updated: 05/03/96
 *  author: Randy Ding (randyd@crl.com)
 *          check my web page for updates: http://www.crl.com/~randyd
 *
 *  This ~should~ compile "as is" for Windows NT, Dos, and UNIX.
 *  To compile on unix:
 *   1) gcc -o uud uud.c
 *   2) place uud in a binary directory included in your path, like ~/bin
 *   3) rehash if your shell is csh, tcsh, etc..
 *
 *  version history:
 *  V1.0.0  02/07/96  first release
 *  V1.0.1  02/14/96  fixed bug in query for replacing existing files
 *  V1.1.0  02/21/96  made query -q default, added -u update option
 *  V2.0.0  03/11/96  rewrite filter logic, added -t test option
 *  V2.1.0  03/14/96  delete bad files, added -k keep bad output files option
 *                    made 'no files uudecoded' an error and returns 1
 *  V2.1.1  03/16/96  (unix only) fixed bug in chmod() error reporting
 *  V2.1.2  04/04/96  code cleanup, no functional changes
 *  V2.1.3  05/03/96  added -r relax end checking option
 *
 * This program uudecodes multi-part, multi-uuencodes from stream.
 * cmdline: uud [-hvsbtkerfmuqnxpid] [infile] [-g grepfile] [-o auxfile]
 * command line options in any order are...
 * -h  display help;        -v  display version history and author contact info
 * -s  silent operation, no messages displayed to stderr
 * -b  brief operation, displays only important error messages to stderr
 * -t  range test every character decoded, abort if one found, see also -e
 * -k  keep bad output files which may also be short, delete is default
 * -e  stop after first error;    -r  relax end checking, allows missing ` line
 * -f# stop after #'th successful uudecode, 1 for -f, usually include -e
 * -m# overrides all \"begin xxx\" modes with #, 600 for -m
 * -u  update existing files without asking, -u -q -n -x are mutually exclusive
 * -q  query (default), ask before updating files, disabled if input from stdin
 * -n  never overwrite existing files, skip with error, include -e to stop
 * -x  don't overwrite existing files, skip quietly with no error
 * -p  binary output to stdout, no breaks, filenames and modes are ignored
 * -i  makes grepfile match case insensitive
 * -d  debug command line parameters, shows all to stdout and stops
 * [infile] takes input from infile, otherwise stdin
 * [-g# grepfile] skips input until #'th occur. of grepfile, see also -i -f -o
 * [-o# auxfile] renames the #'th encountered uudecode file, see also -e -f -g
 *
 *  This program takes news articles saved to file by rn or trn
 *  which have embedded multiple part uuencodings and extracts
 *  them.  The binaries will appear in the current directory.
 *  Articles with multiple part uuencodings must be saved in the
 *  input file in the proper sequence,
 *  ie. parts (1/3) then (2/3) and finally (3/3).
 *  uud.c will strip the headers and unwanted text between parts.
 *  All of the uuencodings will be decoded if more than one is
 *  present in the input stream.
 *  A report is generated to stderr for each uuencoding encountered
 *  unless the silent operation parameter is used.
 *  In silent operation, no error messages are reported either.
 *  In brief operation, only important error messages are reported.
 *  If the optional input file is not specified, stdin is assumed.
 *  1 is returned if an error is encountered.  This is the only way
 *  to detect an error (in script) if the -s parameter is used.
 */

#define copyright "uud.c  V2.1.3  Jan 1996  by Randy Ding  updated 05/03/96"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

/* The unix version does not need or define O_BINARY in <fcntl.h> */
/* This method of detecting platform may not always work, look into this */
/* if there are problems compiling. */
#ifndef O_BINARY
  #define unixver
  #define O_BINARY 0
#else
  #define pcver
#endif

#ifdef pcver
  #include <sys\stat.h>
  #include <io.h>     /* needed on pc version for open() and write() stuff */
#endif

#ifdef unixver
  #include <sys/stat.h>
#endif


/*
 * This subroutine strips all <cr> and <lf> from the end of buf[]
 * This is done because a lot of Internet-News newbies don't know how to
 * uuencode their files properly.
 * Standard unix uudecode programs usually puke if there is a ^M in the line.
 *
 * n is the size of buf[]
 */
void fixeoln(char buf[], int n)
{
    size_t len,nn;
    int done;

    if (n<1) return;
    nn = (size_t)n - 1;
    len=strlen(buf);
    if (nn<len) len=n;
    done=0;
    while (len>0 && !done) {
        switch (buf[len-1]) {
            case '\r': len--;  break;
            case '\n': len--;  break;
            case ' ':  if (len>61) len--; else done=1;  break;
            default:   done=1;
        }
    }
    buf[len]='\0';
    return;
}




/*
 * This subroutine is just like the standard fgets() except smarter.
 * When lines in the stream are longer than the buffer size, the line gets
 * truncated and the stream is purged until the actual <lf> is seen.
 * This works just like the Pascal readln(f,stringvar)
 */
char *fgetslongline(char *s, int n, FILE *stream)
{
    char buf[32];
    char *result;
    size_t nn;

    result = fgets(s,n,stream);
    if (n<2 || result==NULL) return(result);

    nn=(size_t)n - 1;
    if (strlen(s)==nn && s[n-2] != '\n') {
        while (fgets(buf,sizeof(buf),stream) != NULL)
            if (strlen(buf)+1<sizeof(buf) || buf[sizeof(buf)-2]=='\n') break;
    }
    fixeoln(s,n);
    return(result);
}




/*
 * This subroutine checks buf to see if it looks like
 * a standard uuencode line and returns a number for best guess.
 *
 * returned values are one of the following...
 *   0) unknown, possibly text or short last line
 *   1) "begin 644 filename" type of line
 *   2) "M..." etc., main body line, short last line is not detected as 2
 *   3) "`" or " ", this is usually the 2nd from last line in a uuencode
 *   4) "end", usually last line of a uuencode
 *
 * n is the size of buf[]
 */
int isgoodbuf(char buf[], int n)
{
    size_t len,nn;

    len = strlen(buf);
    nn = (size_t)n - 1;
    if (nn<len) len=nn;
    if (len==61 && buf[0]=='M') return(2);
    if (len>10 && strncmp(buf,"begin ",6)==0 && isdigit(buf[6]) &&
        isdigit(buf[7]) && isdigit(buf[8])) return(1);
    if (len>=1 && (buf[0]==' ' || buf[0]=='`')) return(3);
    if (len>=3 && strncmp(buf,"end",3)==0) return(4);
    return(0);
}




/*
 * uudecode 1 line, assumes buf[] is a valid uuencoded line
 * if test=1 then range check every character decoded, a little slower!
 * return 0 if no error
 * return 1 if general error
 * return 2 if illegal character (only when test=1)
 */
int uud1line(int tohandle, int test, unsigned char buf[])
{
    int len,b,i,j;
    size_t l;
    unsigned char c0, c1, x, *from, *to;
    static unsigned char binbuf[45];
    static const unsigned char c[3][4] = {   /* used by slow decoder below */
        {2,    4,    6   , 0 },
        {4,    2,    0   , 0 },
        {0x03, 0x0f, 0x3f, 0 }
    };

    len=(int)buf[0]-0x20; /* first char on line is # 8 bit bytes to decode */
    l=strlen(buf);
    b=0;
    /* this is the FAST! decoder for the main body lines */
    if (len==45 && l>=61) {
        from=&buf[1];
        to=binbuf;
        if (!test) while (b<len) {
            *(to++)=((*from-0x20) << 2) | (((*(from+1)-0x20) >> 4) & 0x03);
            from++;
            *(to++)=((*from-0x20) << 4) | (((*(from+1)-0x20) >> 2) & 0x0f);
            from++;
            *(to++)=((*from-0x20) << 6) | (( *(from+1)-0x20      ) & 0x3f);
            from++;
            from++;
            b=b+3;
        } else while (b<len) {
            if (*from<' ' || *from>'`') return(2);
            *(to++)=((*from-0x20) << 2) | (((*(from+1)-0x20) >> 4) & 0x03);
            from++;
            if (*from<' ' || *from>'`') return(2);
            *(to++)=((*from-0x20) << 4) | (((*(from+1)-0x20) >> 2) & 0x0f);
            from++;
            if (*from<' ' || *from>'`') return(2);
            *(to++)=((*from-0x20) << 6) | (( *(from+1)-0x20      ) & 0x3f);
            from++;
            if (*from<' ' || *from>'`') return(2);
            from++;
            b=b+3;
        }
        if (write(tohandle,binbuf,len)!=len) return(1);
        return(0);
    }
    if (len<1 || len>45) {
        if (!test) return(1);
        else return(2);
    }
    while (b<len) {    /* this is the slow decoder for the last short line */
        i=b%3;
        j=b*4/3+1;
        if ((size_t)j<l) c0=buf[j]; else c0=' ';
        if ((size_t)j+1<l) c1=buf[j+1]; else c1=' ';
        if (test)
            if (c0<' ' || c0>'`' || c1<' ' || c1>'`') return(2);
        x=((c0-0x20) << c[0][i]) + (((c1-0x20) >> c[1][i]) & c[2][i]);
        if (write(tohandle, &x, 1)!=1) return(1);
        b++;
    }
    return(0);
}




/* converts entire string to lower case, n is size of buf[] */

unsigned char *mytolower(unsigned char buf[], int n)
{
    int i;
    unsigned char ch;

    for (i=0; i<n; i++) {
        ch = buf[i];
        if (ch=='\0') break;
        if (ch>='A' && ch<='Z') buf[i]=ch+('a'-'A');
    }
    return(buf);
}




/* extracts numerical operand from command line parameter */
/* on entry s points to the "x" in "-x123" */
/* returned value is the number of chars to skip over, 3 in above example */
/* if the value cannot be converted, 0 is returned */
/* negative operands are allowed, "-x-123" returns 4 and *num=-123 */

int numerand(const unsigned char s[], int *num)
{
    int x;
    long int res;

    x=1;
    res=0;
    *num=1;
    if (s[x]=='-') {
        if (isdigit(s[++x])) *num=-1;
        else return(0);
    }
    while (x<8 && isdigit(s[x])) res=res*10+(s[x++]-'0');
    res=res*(*num);
    if (res>32767 || res<-32768) return(0);
    *num=res;
    return(x-1);
}





/* bufsize can be 81 because uuencoded files should not be wider */
/* this includes the first begin line where the filename is specified */
#define bufsize 81

char bufa[bufsize], bufb[bufsize], bufc[bufsize], bufd[bufsize];
char *buf=bufa, *bufm1=bufb, *bufm2=bufc, *bufm3=bufd;
int goodbuf=-1, goodbufm1=-1, goodbufm2=-1, goodbufm3=-1;





void printlineerror(int brief, int res, const char *s)
{
    int tohandle;

    tohandle=fileno(stderr);
    switch (res) {
        case 1:
        if (!brief) fprintf(stderr,"uudecode error in line %s..\n",s);
        else fprintf(stderr,"uudecode line error\n");
        break;

        case 2:
        if (!brief) fprintf(stderr,"illegal character in line %s..\n",s);
        else fprintf(stderr,"illegal character\n");
        break;

        case 0: break;

        default: return;
    }
    if (brief) return;

    fprintf(stderr,"  m3: \"");
    write(tohandle,bufm3,strlen(bufm3));
    fprintf(stderr,"\" (%d)\n",goodbufm3);

    fprintf(stderr,"  m2: \"");
    write(tohandle,bufm2,strlen(bufm2));
    fprintf(stderr,"\" (%d)\n",goodbufm2);

    fprintf(stderr,"  m1: \"");
    write(tohandle,bufm1,strlen(bufm1));
    fprintf(stderr,"\" (%d)\n",goodbufm1);

    fprintf(stderr,"  m0: \"");
    write(tohandle,buf  ,strlen(buf  ));
    fprintf(stderr,"\" (%d)\n",goodbuf  );

    return;
}





void rotatebufs(void)
{
    char *temp;

    temp=bufm3;
    bufm3=bufm2;
    bufm2=bufm1;
    bufm1=buf;
    buf=temp;
    strcpy(buf,"");
    goodbufm3=goodbufm2;
    goodbufm2=goodbufm1;
    goodbufm1=goodbuf;
    goodbuf=-1;
}





void clearmbufs(void)
{
    strcpy(bufm3,"");
    strcpy(bufm2,"");
    strcpy(bufm1,"");
    goodbufm3=-1;
    goodbufm2=-1;
    goodbufm1=-1;
}





/*
 * This is the main extract subroutine.
 * buf[] is used in the event that a new "begin" is encountered
 * before the ending of the previous uuencoding.  This is an error
 * condition and to keep from messing up the next uudecode which is
 * probably ok, this begin line is stuffed into the buf[] for
 * the next pass through extractuu().
 * returned values...
 *  0=success
 *  1=error
 *  2=begin line encountered before end line decoded
 *  3=eof seen
 */
int extractuu(FILE *from,   /* input stream, stdin or text file */
              int tohandle, /* handle to output, can be stdout or a file */
              int silent,
              int brief,
              int relax,
              int test,     /* range test every character */
              char *grepfile,   /* if not NULL then match the filename */
              int igrep,    /* =1 for case insensitive grepfile comparison */
              int ngrep)    /* if > 1 then searches for n'th occurrence */
{
    int amode,res;
    char fn[bufsize],gfn[bufsize];

    if (grepfile!=NULL) {
        strncpy(gfn,grepfile,sizeof(gfn)-1);
        gfn[sizeof(gfn)-1]='\0';
        if (igrep) mytolower(gfn,sizeof(gfn));
    }
    if (strlen(buf)==0) {
        retry:
        goodbuf=0;
        while (fgetslongline(buf,bufsize,from)!=NULL)
            if ((goodbuf=isgoodbuf(buf,bufsize))==1) break;
        if (goodbuf!=1) return(3);      /* indicate eof seen */
        if (grepfile==NULL) return(0);  /* return first begin line seen */
        if (sscanf(buf,"begin %d %80s",&amode,fn)!=2) goto retry;
        if (igrep) mytolower(fn,sizeof(fn));
        if (strcmp(fn,gfn)!=0) goto retry;
        if (--ngrep>0) {
            if (!silent && !brief)
                fprintf(stderr,"\"%s\" skipped, %d more to find\n",fn,ngrep);
            goto retry;
        }
        return(0);
    }

    /* here when begin line was found and is in buf[], now do decoding */
    clearmbufs();                /* clear back history buffers */
    goodbuf=1;                   /* assume begin line is in buf[] */
    if (isgoodbuf(buf,bufsize)!=goodbuf) {
        if (!silent) fprintf(stderr,"Internal Error 01\n");
        return(1);
    }
    rotatebufs();
    while (fgetslongline(buf,bufsize,from)!=NULL) {
        goodbuf=isgoodbuf(buf,bufsize);
        if (goodbufm1==2 && (goodbuf==2 || goodbufm2==2 || goodbufm2==1))
            if ((res=uud1line(tohandle,test,bufm1))!=0) {
                if (!silent) printlineerror(brief,res,"m1");
                return(1);
            }
        if (goodbuf==1) {
            if (!silent) fprintf(stderr,"begin before end\n");
            return(2);
        }

        /*  end detected...logic for what to do with 3rd from last line bufm2
         *  (for relax=0 or 1)
         *
         *  ignore this end | decode line bufm2 | just exit, already decoded
         *     (m2) (m3)    |    (m2) (m3)      |    (m2) (m3)   <-- goodbuf
         *  ----------------+-------------------+---------------------------
         *      0    0      |     0    1        |     1    x
         *      0    3      |     0    2        |     2    1
         *      0    4      |     2    0        |     2    2
         *      2    3      |                   |
         *      2    4      |                   |
         *      3    x      |                   |
         *      4    x      |                   |
         */
        if (goodbuf==4 && goodbufm1==3) {     /* end sequence seen */
            if (goodbufm2==1) return(0);
            if ((goodbufm2==0 && (goodbufm3==0 || goodbufm3>=3)) ||
                (goodbufm2>=3) || (goodbufm2==2 && goodbufm3>=3))
            {
                if (!silent) {
                    fprintf(stderr,"ignoring suspicious end\n");
                    printlineerror(brief,0,"");
                }
                goto nextline;
            }
            if ((goodbufm2==0 && (goodbufm3==1 || goodbufm3==2)) ||
                (goodbufm2==2 && goodbufm3==0))
            {
                if ((res=uud1line(tohandle,test,bufm2))!=0) {
                    if (!silent) printlineerror(brief,res,"m2");
                    return(1);
                }
            }
            return(0);
        }

        /*  end detected...logic for what to do with 2nd from last line bufm1
         *  (for relax=1)
         *
         *  ignore this end | decode line bufm2 | just exit, already decoded
         *     (m1) (m2)    |    (m1) (m2)      |    (m1) (m2)   <-- goodbuf
         *  ----------------+-------------------+---------------------------
         *      0    0      |     0    1        |     1    x
         *      0    3      |     0    2        |     2    1
         *      0    4      |     2    0        |     2    2
         *      2    3      |                   |
         *      2    4      |                   |
         *      3    x      |                   |
         *      4    x      |                   |
         */
        if (relax && goodbuf==4) {     /* end sequence seen */
            if (goodbufm1==1) return(0);
            if ((goodbufm1==0 && (goodbufm2==0 || goodbufm2>=3)) ||
                (goodbufm1>=3) || (goodbufm1==2 && goodbufm2>=3))
            {
                if (!silent) {
                    fprintf(stderr,"ignoring suspicious end\n");
                    printlineerror(brief,0,"");
                }
                goto nextline;
            }
            if ((goodbufm1==0 && (goodbufm2==1 || goodbufm2==2)) ||
                (goodbufm1==2 && goodbufm2==0))
            {
                if ((res=uud1line(tohandle,test,bufm1))!=0) {
                    if (!silent) printlineerror(brief,res,"m1");
                    return(1);
                }
            }
            return(0);
        }
        nextline:
        rotatebufs();
    }
    if (goodbufm1==2 && (goodbufm2==1 || goodbufm2==2))
        if ((res=uud1line(tohandle,test,bufm1))!=0)
            if (!silent) printlineerror(brief,res,"m1");
    if (!silent) fprintf(stderr,"eof reached before end\n");
    return(3);                 /* indicate end of input file */
}




int main(int argc, char *argv[])
{
    char *grepfile;      /* filename following -g parameter on command line */
    char *ofile;         /* filename following -o parameter on command line */
    char *infile;        /* first filename specified on command line */
    char *outfile;       /* second filename specified on command line */
    char **fileparm;     /* used to assign cmdline names by [-o <file>] */
    char ch[2];          /* general purpose character, or string of 1 char */
    char targetfile[256];   /* 255 chars max length of a pathname */
    int res,success,errors; /* gp result, success and error counters */
    int silent,brief,len,i,j,n,cmderr,help,version;
    int errstop,relax,firststop,test,overmode,topipe;
    int disparms,query,never,skip,igrep,grepnum,onum,keep;
    int amode;              /* "begin xxx" where xxx is amode, for chmod() */
    FILE *from;             /* input file variable */
    int tohandle=0;         /* output file handle, from open() */
  #ifdef pcver
    int previousmode = -1;  /* in pc version it is previous mode of stdout */
  #endif

    /* process the command line parameters, initialize defaults */
    infile=NULL;     /* input filename, stdin used if null */
    outfile=NULL;    /* output filename */
    grepfile=NULL;   /* search filename */
    ofile=NULL;      /* rename file */
    silent=0;        /* print nothing (-s) */
    brief=0;         /* print important stuff only (-b) */
    help=0;          /* show help and exit (-h) */
    version=0;       /* show version info and exit (-v) */
    errstop=0;       /* abort after first error (-e) */
    relax=0;         /* allows missing ` line just before end */
    overmode=-1;     /* numeric param following (-m) */
    grepnum=0;       /* numeric param following (-g) */
    onum=0;          /* numeric param following (-o) */
    query=1;  /* query -q option is default */
    never=0;  /* update mode is query=0, never=0, skip=0 */
    skip=0;   /* query, never, skip, and update are all mutually exclusive */
    igrep=0;         /* ignore case when searching for grepfile */
    firststop=0;     /* stop after first successfull decode (-f) */
    test=0;          /* test every character decoded for proper range, slow */
    keep=0;          /* keep bad output files, delete is default */
    topipe=0;        /* output to stdout (-p) */
    cmderr=0;        /* true when cmdline parser finds error */
    disparms=0;      /* -d option for debugging cmdline */
    fileparm=NULL;   /* temp string pointer for assigning filenames */

    /* the following loop parses the command line, options and filenames */
    i=0;
    while (++i<argc && !cmderr) {
        len=(int)strlen(argv[i]);
        if (argv[i][0]=='-') {
            j=0;
            while (++j<len && !cmderr)
                switch (toupper(argv[i][j])) {
                    case 'S': silent=1; break;
                    case 'B': brief=1; break;
                    case 'H': help=1; break;
                    case 'P': topipe=1; break;
                    case 'I': igrep=1; break;
                    case 'V': version=1; break;
                    case 'E': errstop=1; break;
                    case 'R': relax=1; break;
                    case 'T': test=1; break;
                    case 'K': keep=1; break;
                    case 'U': query=0; never=0; skip=0; break;
                    case 'Q': query=1; never=0; skip=0; break;
                    case 'N': never=1; query=0; skip=0; break;
                    case 'X': skip=1; query=0; never=0; break;
                    case 'D': disparms=1; break;
                    case 'F':
                    if ((n=numerand(&argv[i][j],&firststop))>0) j=j+n;
                    else firststop=1;
                    if (firststop<1) firststop=1;
                    break;

                    case 'M':
                    if ((n=numerand(&argv[i][j],&overmode))>0) j=j+n;
                    else overmode=600;
                    if (overmode<0 || overmode>777 || overmode%10>7 ||
                        overmode/10%10>7 || overmode/100%10>7) overmode=600;
                    break;

                    case 'G':
                    fileparm=&grepfile;
                    if ((n=numerand(&argv[i][j],&grepnum))>0) j=j+n;
                    else grepnum=1;
                    if (grepnum<1) grepnum=1;
                    break;

                    case 'O':
                    fileparm=&ofile;
                    if ((n=numerand(&argv[i][j],&onum))>0) j=j+n;
                    else onum=1;
                    if (onum<1) onum=1;
                    break;

                    default : cmderr=1; break;
                }
            continue;
        }
        if (fileparm!=NULL) {         /* assign next filename to *fileparm */
            *fileparm=argv[i];
            fileparm=NULL;    /* assign next filename to infile or outfile */
        } else {
            if (infile==NULL) infile=argv[i]; /* first specified filename */
            else outfile=argv[i];             /* second specified filename */
        }
    }
    /* end of processing command line parameters */

    /* this is for debugging the command line, use -d parameter first */
    if (disparms) {
        fprintf(stdout,"%s\n",copyright);
        fprintf(stdout,"debug comand line parameters mode...\n");
        if (infile!=NULL) fprintf(stdout,"infile= \"%s\"\n",infile);
        else fprintf(stdout,"infile=NULL\n");
        if (outfile!=NULL) fprintf(stdout,"outfile= \"%s\"\n",outfile);
        else fprintf(stdout,"outfile=NULL\n");
        if (grepfile!=NULL) fprintf(stdout,"grepfile= \"%s\"\n",grepfile);
        else fprintf(stdout,"grepfile=NULL\n");
        if (ofile!=NULL) fprintf(stdout,"ofile= \"%s\"\n",ofile);
        else fprintf(stdout,"ofile=NULL\n");
        fprintf(stdout,"-s  silent= \"%d\"\n",silent);
        fprintf(stdout,"-b  brief= \"%d\"\n",brief);
        fprintf(stdout,"-h  help= \"%d\"\n",help);
        fprintf(stdout,"-v  version= \"%d\"\n",version);
        fprintf(stdout,"-e  errstop= \"%d\"\n",errstop);
        fprintf(stdout,"-r  relax= \"%d\"\n",relax);
        fprintf(stdout,"-f  firststop= \"%d\"\n",firststop);
        fprintf(stdout,"-t  test= \"%d\"\n",test);
        fprintf(stdout,"-k  keep= \"%d\"\n",keep);
        fprintf(stdout,"-m  overmode= \"%d\"\n",overmode);
        fprintf(stdout,"-q  query= \"%d\"\n",query);
        fprintf(stdout,"-n  never= \"%d\"\n",never);
        fprintf(stdout,"-x  skip= \"%d\"\n",skip);
        fprintf(stdout,"-p  topipe= \"%d\"\n",topipe);
        fprintf(stdout,"-i  igrep= \"%d\"\n",igrep);
        fprintf(stdout,"-g  grepnum= \"%d\"\n",grepnum);
        fprintf(stdout,"-o  onum= \"%d\"\n",onum);
        fprintf(stdout,"-d  disparms= \"%d\"\n",disparms);
        fprintf(stdout,"cmderr= \"%d\"\n",cmderr);
        return(1);
    }
    if (version) {
        fprintf(stdout,"%s\n",copyright);
        fprintf(stdout,"This program is public domain freeware.\n");
        fprintf(stdout,"compiles on unix and ibmpc\n");
        #ifdef pcver
          fprintf(stdout,"compiled as pc version\n");
        #endif
        #ifdef unixver
          fprintf(stdout,"compiled as unix version\n");
        #endif
        fprintf(stdout,"first contact: randyd@crl.com\n");
        fprintf(stdout,"               http://www.crl.com/~randyd\n");
        fprintf(stdout,"version history:\n");
        fprintf(stdout,"V1.0.0  02/07/96  first release\n");
        fprintf(stdout,"V1.0.1  02/14/96  fixed bug in query for replacing existing files\n");
        fprintf(stdout,"V1.1.0  02/21/96  made query -q default, added -u update option\n");
        fprintf(stdout,"V2.0.0  03/11/96  rewrite filter logic, added -t test option\n");
        fprintf(stdout,"V2.1.0  03/14/96  delete bad files, added -k keep bad output files option\n");
        fprintf(stdout,"                  made 'no files uudecoded' an error and returns 1\n");
        fprintf(stdout,"V2.1.1  03/16/96  (unix only) fixed bug in chmod() error reporting\n");
        fprintf(stdout,"V2.1.2  04/04/96  code cleanup, no functional changes\n");
        fprintf(stdout,"V2.1.3  05/03/96  added -r relax end checking option\n");
        if (!help && !cmderr) return(1);
        else fprintf(stdout,"\n");
    }
    if (help) {
        fprintf(stdout,"%s\n",copyright);
        fprintf(stdout,"This program uudecodes multi-part, multi-uuencodes from stream.\n");
        fprintf(stdout,"cmdline: uud [-hvsbtkerfmuqnxpid] [infile] [-g grepfile] [-o auxfile]\n");
        fprintf(stdout,"command line options in any order are...\n");
        fprintf(stdout,"-h  display help;        -v  display version history and author contact info\n");
        fprintf(stdout,"-s  silent operation, no messages displayed to stderr\n");
        fprintf(stdout,"-b  brief operation, displays only important error messages to stderr\n");
        fprintf(stdout,"-t  range test every character decoded, abort if one found, see also -e\n");
        fprintf(stdout,"-k  keep bad output files which may also be short, delete is default\n");
        fprintf(stdout,"-e  stop after first error;    -r  relax end checking, allows missing ` line\n");
        fprintf(stdout,"-f# stop after #'th successful uudecode, 1 for -f, usually include -e\n");
        fprintf(stdout,"-m# overrides all \"begin xxx\" modes with #, 600 for -m\n");
        fprintf(stdout,"-u  update existing files without asking, -u -q -n -x are mutually exclusive\n");
        fprintf(stdout,"-q  query (default), ask before updating files, disabled if input from stdin\n");
        fprintf(stdout,"-n  never overwrite existing files, skip with error, include -e to stop\n");
        fprintf(stdout,"-x  don't overwrite existing files, skip quietly with no error\n");
        fprintf(stdout,"-p  binary output to stdout, no breaks, filenames and modes are ignored\n");
        fprintf(stdout,"-i  makes grepfile match case insensitive\n");
        fprintf(stdout,"-d  debug command line parameters, shows all to stdout and stops\n");
        fprintf(stdout,"[infile] takes input from infile, otherwise stdin\n");
        fprintf(stdout,"[-g# grepfile] skips input until #'th occur. of grepfile, see also -i -f -o\n");
        fprintf(stdout,"[-o# auxfile] renames the #'th encountered uudecode file, see also -e -f -g\n");
        if (!cmderr) return(1);
        else fprintf(stdout,"\n");
    }
    if (cmderr) {
        fprintf(stderr,"uud.c command line error, use -h parameter for help\n");
        return(1);
    }

    if (!silent && !brief) {
        fprintf(stderr,"%s\n",copyright);
        fprintf(stderr,"for help use: uud -h  [Ctrl-C breaks]\n");
    }
    /*
     * it is not easy to get keyboard input for query option (-q)
     * when the input is from stdin pipe, so disable the option
     * in this case...as far as I know as of this writing...
     */
    if (query && infile==NULL) {
        query=0;
        never=1;
        if (!silent && !brief) fprintf(stderr,
    "stdin is piped, changing query option to never overwrite (-q to -n)\n");
    }
    if (topipe && !brief && !silent && keep)
        fprintf(stderr,"-k will have no effect with -p option\n");
    /* if topipe, get file handle of stdout rather than opening a new file */
    if (topipe) tohandle=fileno(stdout);
    if (infile!=NULL)
        if ((from=fopen(infile,"r"))==NULL) {
            if (!silent) fprintf(stderr,"%s\"%s\"\n",
                "fatal: can't open for input ",infile);
            return(1);
        }
    /*
     * change stdout to binary mode for ibmpc version only.
     * the pc has an anoying feature where it changes all
     * <lf> characters to <cr><lf> pairs unless you change
     * the output mode of stdout to binary. Text mode is default.
     * it need not be said that this is very bad for binary output
     */
  #ifdef pcver
    if (topipe)
        if ((previousmode=setmode(tohandle,O_BINARY)) == -1) {
            if (!silent) fprintf(stderr,
                "fatal: error converting stdout to binary mode\n");
            if (infile!=NULL) fclose(from);
            return(1);
        }
  #endif
    errors=0;
    success=0;
    res=0;
    amode= -1;
    strcpy(targetfile,"");
    while (res<3) {  /* 0 success, 1 error, 2 begin before end, 3 eof seen */
        if (res<2) {     /* skip this if buf[] already contains begin line */
            strcpy(buf,"");    /* tells extractuu to search for begin line */
            if (infile!=NULL) res=extractuu(from,tohandle,
                silent,brief,relax,test,grepfile,igrep,grepnum);
            else res=extractuu(stdin,tohandle,
                silent,brief,relax,test,grepfile,igrep,grepnum);
            if (grepfile!=NULL && res==3) {
                errors++;
                if (!silent) fprintf(stderr,"\"%s\" not found\n",grepfile);
            }
            grepfile=NULL;  /* grepfile only gets used once */
        }
        /*
         * here if we have res==2 then we must be recovering from a
         * begin before end error on the previous pass, recycling buf[],
         * the extractuu() above couldn't have generated a res==2 because
         * it's looking for a begin so it must have come from the one below
         */
        if (res==0 || res==2) {
            if (sscanf(buf,"begin %d %255s",&amode,&targetfile)!=2) {
                if (!silent) fprintf(stderr,"bad begin line \"%s\"\n",buf);
                res=1;
                goto badfilename;
            }
            if (!silent && !brief)
                fprintf(stderr,"uudecoding \"%s\"\n",targetfile);
            if (overmode>=0) amode=overmode;   /* override file mode */
            /*
             * convert amode from decimal to octal
             * done this way instead of using %o in above sscanf()
             * because a non octal digit in the begin line gives
             * unpredictable results when sscanf() scans targetfile.
             * amode conversion is done here even when to pipe, just
             * don't display errors if it finds one because it don't matter
             * when the output is to pipe.
             */
            amode = amode%10 + amode/10%10*8 + amode/100%10*64;
            if (amode<0 || amode>0x1ff) {
                if (!topipe && !silent && !brief) fprintf(stderr,
                "illegal mode \"%o\", changing to 600\n",amode);
                amode=0x180;
                if (overmode>=0) overmode=600;   /* so no errors next time */
            }
            /* ignore renaming target when output is to a pipe, -p option */
            if (!topipe && ofile!=NULL) {
                if (--onum<=0) {
                    strcpy(targetfile,ofile);
                    if (!silent && !brief)
                        fprintf(stderr,"renaming target to \"%s\"\n",
                        targetfile);
                    ofile=NULL;  /* apply rename to #'th uudecode only */
                }
            }
            /* ignore target file checking when output to pipe, -p option */
            if (!topipe) {
                /* check for file existance */
                if (access(targetfile,0)==0) {        /* it exists */
                    if (never) {         /* count as an error if it exists */
                        if (!silent && brief) fprintf(stderr,"\"%s\" ",
                            targetfile);
                        if (!silent) fprintf(stderr,"Skipped, file exists.\n");
                        res=1;
                        goto badfilename;
                    }
                    if (skip) {  /* quietly skip without counting as error */
                        if (!silent && !brief)
                            fprintf(stderr,"Skipped, file exists.\n");
                        res=0;
                        goto badfilename1;
                    }
                    /* can't use query when stdin is redirected */
                    /* ie. infile is null, if infile is a file, we are ok */
                    if (query && infile!=NULL) {
                        fflush(stdin);
                        if (silent || brief)  /* since name not printed yet */
                            fprintf(stderr,"\"%s\" ",targetfile);
                        fprintf(stderr,"Target exists, update? (y|N|a|q) ");
                        scanf("%1s",ch);
                        if (toupper(*ch)=='A') {
                            if (!silent && !brief) fprintf(stderr,
                                "All, query mode disabled.\n");
                            query=0;
                            *ch='Y';
                        }
                        if (toupper(*ch)=='Q') {
                            if (!silent && !brief) fprintf(stderr,
                                "Quit, ignoring rest of input file.\n");
                            res=3;
                            goto badfilename;
                        }
                        if (toupper(*ch)!='Y') {
                            if (!silent && !brief)
                                fprintf(stderr,"Skipped.\n");
                            res=0;
                            goto badfilename1;
                        }
                    }
                    /* check if file has write access */
                    if (access(targetfile,2)!=0) {    /* write not allowed */
                        if (!silent && brief) fprintf(stderr,"\"%s\" ",
                            targetfile);
                        if (!silent) fprintf(stderr,
                            "Skipped, file exists write protected.\n");
                        res=1;
                        goto badfilename;
                    }
                    /* we determined that the file is ok to overwrite */
                    if (!silent && !brief)
                        fprintf(stderr,"updating \"%s\"\n",targetfile);
                }
                /* O_BINARY is defined above as 0 if compiled on unix */
                /* it is only required on ibmpc version because of its */
                /* annoying eoln conversion from <lf> to <cr><lf> */
                if ((tohandle = open(targetfile, O_BINARY | O_WRONLY | O_CREAT
                        | O_TRUNC, S_IREAD | S_IWRITE)) == -1) {
                    if (!silent) fprintf(stderr,"%s\"%s\"\n",
                        "fatal: can't open target file ",targetfile);
                    res=3;
                    goto badfilename;
                }
            }
            /* buf[] contains begin line, this tells extractuu to decode */
            /* rather than search for the next begin line */
            if (infile!=NULL)
                res=extractuu(from,tohandle,silent,brief,relax,test,NULL,0,0);
            else
                res=extractuu(stdin,tohandle,silent,brief,relax,test,NULL,0,0);
            if (!topipe) {
                if (close(tohandle)!=0) {
                    if (!silent) fprintf(stderr,
                        "error closing \"%s\"\n",targetfile);
                    goto badfilename;
                }
              #ifdef unixver
                if (res==0)
                    if (chmod(targetfile,amode)!=0) {
                        if (!silent && brief) fprintf(stderr,"\"%s\" ",
                            targetfile);
                        if (!silent) fprintf(stderr,
                        "chmod() error changing file mode to \"%o\"\n",amode);
                    }
              #endif
                if (res>0 && !keep) {
                    if (!silent && !brief)
                        fprintf(stderr,"deleting \"%s\"\n",targetfile);
                  #ifdef unixver
                    if (chmod(targetfile,511)!=0) {
                        if (!silent && brief) fprintf(stderr,"\"%s\" ",
                            targetfile);
                        fprintf(stderr,
                        "chmod() error changing file mode to \"%o\"\n",511);
                        goto badfilename;
                    }
                  #endif
                    if (unlink(targetfile)!=0) if (!silent)
                        fprintf(stderr,"error deleting \"%s\"\n",targetfile);
                }
            }
        }
        badfilename:
        if (res==0) success++;
        badfilename1:
        if (res==1 || res==2) errors++;
        if (errstop && errors!=0) break;
        if (firststop>0 && success>=firststop) break;
    }                                    /* <-- end of while (res<3) { ... */
    if (infile!=NULL) fclose(from);
    if (!silent && !brief && errors>0)
        fprintf(stderr,"%s%d/%d\n","total errors = ",errors,errors+success);
  #ifdef pcver
    if (topipe && previousmode!= -1) setmode(tohandle,previousmode);
  #endif
    if (errors==0 && success==0) {
        if (!silent) fprintf(stderr,"no files uudecoded\n");
        return(1);
    }
    if (errors!=0) return(1);
    return(0);
}
