*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Beauzamy bound on the maximum possible magnitude
*		of the factors of the univariate polynomial.
* Stack:	( P --> Zbound )
* Notes:
*	                             -------------
*	      3^(3/4)   3^(n/2)     / n    a[i]^2
*	B <= ---------- ------- *  / SUM ---------
*	     2*SQRT(PI) SQRT(n)   /  i=0 comb(n,i)
*
*
*	To calculate the comb(n,i) function we will utilize the
*	formula comb(n,i+1) = (n-i)/(i+1)*comb(n,i). We also
*	note that this value is symmetric so that we can take
*	the coefficients in any order we like. Also to avoid
*	unnecessary coefficient growth we will add the two
*	coefficients with the same denominator at the same time,
*	and scale all rational numbers by 2^k and then divide so
*	that we don't actually have to handle rational numbers.
*
*	Also, once the expression inside the root has been calculate
*	we raise the entire expression to 4th power to enable more
*	precision in the actual integer result. We also use an upper
*	bound on the constant coefficient as follows
*
*
*		  3^3
*		---------- * 2^32 < 734351350
*		2^4 * PI^2
*
*	We also multiply the bound by 2*an to enable true factor
*	extraction with the lc* PPZ method.
*
* Reference:
*
*	J.Symbolic Computation (1993) 15,393-413
*	'Polynomial Factorization: Sharp Bounds, Efficient Algorithms'
*	Bernard Beauzamy, Vilmar Trevisan, Paul Wang
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME Beauzamy
::
   DUP CARCOMP ZAbs		( P |an| )
   HXS 2 20 QMul SWAP		( 2*an P )

* Calculate the regular bound

   INNERCOMP			( a[n] .. a[0] #n+1 )
   HXS 2 10			( a[n] .. a[0] #n+1 comb[n,0] )
   HXS 1 0			( a[n] .. a[0] #n+1 comb[n,0] 0 )

   ( Use scale *2^16 in the norm calculation )

   3PICK #2/ ZERO_DO (DO)	( a[n-i] .. a[i] #n+1 comb[n,i] b )
	4ROLL DUP QMul		( a[n-i] .. #n+1 comb[n,i] b a[i]^2 )
	4PICK INDEX@ #2* #-
	#3+ ROLL DUP QMul	( .. #n+1 comb[n,i] b a[i]^2 a[n-i]^2 )
	QAdd DupQIsZero?
	ITE_DROP
	::
	   HXS 2 10 HXS 4 0000
	   ROT &HXS
	   4PICK QDiv QAdd QAdd		( .. #n+1 comb[n,i] b' )
	;
	SWAP3PICK INDEX@#- #1- #>Z QMul	( .. #n+1 b' comb[n,i]*[n-i] )
	INDEX@ #1+ #>Z QDiv		( .. #n+1 b' comb[n,i]*[n-i]/[i+1] )
	SWAP				( .. #n+1 comb[n,i+1] b' )
   LOOP

   3PICK ONE #AND #0= ITE		( *Handle odd/even degree* )
	SWAPDROP			( #n+1 b )
	::
	  4ROLL DUP QMul
	  DupQIsZero? casedrop SWAPDROP
	  HXS 2 10 HXS 4 0000 ROT &HXS
	  4ROLL QDiv QAdd QAdd
	;

   ( Squaring raises scale to 2^32 )

   DUP QMul				( #n+1 b^2 )

   ( Multiplying by the constant raises scale to 2^64 )

   HXS 9 6FF45CB20 QMul			( #n+1 b^2*c )

   ( 3^2n and n^2 are handled without scaling )

   HXS 2 30 3PICK #1- #2* PPow# QMul	( #n+1 b^2*c*3^2n )
   SWAP#1- #>Z DUP QMul QDiv		( #n+1 b^2*c*3^2n/n^2 )
   HXS 2 10 QAdd			( Round up the division )

   ( Remove scales and take 4th root )

   HXS 2 10 DUPDUP 4ROLL		( 1 1 1 B )
   HXS 12 000000000000000010		( * the scale 2^64 * )
   QDivRem QIsZero? ITE SWAPDROP QAdd	( 1 1 B' )
   ZSQRT ITE SWAPDROP QAdd		( 1 B'' )
   ZSQRT ITE SWAPDROP QAdd		( B''' )

* And multiply with 2*an

   QMul					( 2*an*B''' )

;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Landau-Mignotte bound on the maximum possible magnitude
*		of the factors of the univariate polynomial.
* Stack:	( P --> Zbound )
* Notes:
*
*	B <= comb(n,floor(n/2)) * norm(poly,2)
*
*	where n is the degree of the polynomial.
*
*	We also multiply the bound by 2*an to enable true factor
*	extraction with the lc* PPZ method.
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME Mignotte
::
  DUP CARCOMP ZAbs		( P |an| )
  HXS 2 20 QMul SWAP		( 2*an P )

* The regular Mignotte bound

  INNERCOMP			( a0 .. an #n+1 )
  HXS 1 0			( a0 .. an #n+1 Z0 )
  OVER ZERO_DO (DO)
     ROTDUP QMul QAdd		( a0 .. #n+1 Zsum )
  LOOP
  SWAP #1-DUP #2/		( Zsum #n #n/2 )
  COMB#				( Zsum Zcomb )
  DUP QMul QMul			( Zbound^2 )
  HXS 2 10 SWAP			( 1 B^2 )
  ZSQRT ITE SWAPDROP QAdd	( B )

* And multiply with 2*an

  QMul				( 2*an*B )

;


**********************************************************************
* Desc:		comb(n,m)
* Stack:	( #n #m --> Z )
* Notes:	We calculate the factorial n! in three parts
*		to get the proper three factorials used in the
*		formula
*
*		( n )      n!
*		(   ) = ---------
*		( m )   (n-m)! m!
*
*		Should change to use comb(n,i+1) = (n-i)/(i+1)*comb(n,i)
*
**********************************************************************
NULLNAME COMB#
::
  DUP#0= case2drop HXS 2 10		( comb[n,0]=1 )
  2DUP#= case2drop HXS 2 10		( comb[n,n]=1 )

  2DUP #-				( #n #m #n-m )
  2DUP#< IT SWAP			( #n #mid #low )

  HXS 2 10
  OVER #1+_ONE_DO (DO)
     INDEX@ #>Z QMul
  LOOP					( #n #mid #low low! )

  4UNROLL 4PICK				( low! #n #mid #low low! )
  3PICK #1+ ROT#1+			( low! #n #mid low! #mid+1 #low+1 )
  2DUP#> ITE
	:: DO INDEX@ #>Z QMul LOOP ;	( low! #n #mid mid! )
	2DROP				( low! #n #mid mid! )

  DUP 2SWAP				( low! mid! mid! #n #mid )
  #1+SWAP #1+SWAP			( low! mid! mid! #n+1 #mid+1 )
  DO INDEX@ #>Z QMul LOOP		( low! mid! n! )

  UNROT QMul QDiv			( --> comb[n,m] )
;

**********************************************************************
