	STITLE	ALG Groebner Bases
**********************************************************************
*	ALG library: Groebner Bases
**********************************************************************

EXTERNAL xGBASIS
EXTERNAL xGSIMP
EXTERNAL xGSOLVE

EXTERNAL GSOLVE
EXTERNAL REDUCESOL

EXTERNAL GBASIS
EXTERNAL SORTBASIS
EXTERNAL REDUCE1

EXTERNAL GBASIS2
EXTERNAL GBASIS3
EXTERNAL REDUCEG
EXTERNAL S-POLY
EXTERNAL HTERM

EXTERNAL EQS2R
EXTERNAL R2EQS

**********************************************************************
*
*	Algorithms
*	----------
*
* Gbasis(G)
* 
*	k = size(G)
*	Create pairs (i,j) i=1..k-1 j=i+1..k
*	InterReduce(G)
*	While (i,j)=NextPair(pairs,G)
*		h = S-poly(Gi,Gj)
*		h = Reduce(h,G)
*		If h<>0 then
*		   k=k+1
*		   Add pairs (i,k) i=1..k-1
*		   G = G & h
*		   InterReduce(G)
*	Repeat
*	return(sort(G))
*
* Reduce(r,G)
*	q = 0
*	begin
*	   while (hterm(r) divisible by some hterm(Gi))
*	       r = r - hterm(r)/hterm(Gi)*Gi
*	   q = q + hterm(r)
*	   r = r - hterm(r)
*	until r=0
*	return q
*
*	Actually we modify above to be use hmono(Gi) and hmono(r)
*	and then if the division for integer coefficients is not
*	exact we multiply both r and q by the integer denominator
*	so that we stay in ring Z all the time.
*
* InterReduce(G)
*
*	begin
*	     flag no restart
*	     for Gi in G
*		h = reduce(Gi,G-Gi)
*		if h=0 then
*		    G = G-Gi
*		    remove pairs containing Gi
*		else if Gi<>h then
*		    Gi = h
*		    flag restart loop
*		endif
*	     next
*	until no restart flag
*
*	Implementation specifically we may restart from G1 as soon
*	as a reduction occurs. Experimentally not doing this may
*	be better. The best option would be to search the largest
*	possible reduction, do it, and then search again. This
*	would mean non calling reduce() at all.
*
*  NextPair(pairs,G)
*
*	For all pairs
*	    if gcd(hterms(pair))=1 then
*		remove(pair)
*	    else
*		save pair if smallest degree(lcm) so far
*	    endif
*	Next
*	If critetion2(smallest pair) then
*	   remove(pair) and restart NextPair(pairs,G)
*	return(smallest pair)
*
*	We use total degree comparisons despite using lexicographic
*	ordering. This seems to be better for most of our examples.
*	I did a lot of experiments too, none worked too well.
*
*	Note: We don't have to do a gcd=1 test here, we could just
*	do it after each inter-reduction as a separate call.
*
*  GSolve(G)
*
*	G = GBasis(G)
*	If trivial then return trivial G
*	Test if some polynomial in G factors
*	If none factors return {G} - e.g. a single solution
*	Else we have factorization of Gi and we set G=G-Gi
*	Then for each combination G&Gij call GSolve recursively
*	Add partial results to Lg
*	And return Lg when all combinations done
*
*  ReduceSol(Lg)
*
*	For all Lgi in Lg
*	   For all Lgj in Lg (j>i)
*	     If GBasis(Lgi&Lgj) = Lgj then
*	         delete Lgj from Lg
*	     Else If GBasis(Lgi&Lgj = Lgi then
*	         delete Lgi from Lg and abort current loop
*	   Next
*	Next
*	Return(Lg)
*
**********************************************************************

**********************************************************************
* Name:		xGBASIS
* Desc:		Calculate Grobner basis for given polynomials
* Stack:	( {eqs} {vars} --> {basis} )
**********************************************************************
ASSEMBLE
	CON(1)	8	
RPL
xNAME GBASIS
::
  CK2&Dispatch
  #50
   ::	FastLib
	DUPTYPELIST ?SKIP ONE{}N	( Leqs Lvars )
	EQS2R SWAP			( Lvars' P )
	GBASIS				( Lvars' G )
	SWAP R2EQS			( Lbasis )
   ;
;
**********************************************************************
* Name:		xGSOLVE
* Desc:		Calculate factorized Grobner basis for given polynomials
* Stack:	( {eqs} {vars} --> {basis1} .. {basisn} n )
**********************************************************************
ASSEMBLE
	CON(1)	8
RPL
xNAME GSOLVE
::
  CK2&Dispatch
  #50
  ::	FastLib
	DUPTYPELIST? ?SKIP ONE{}N	( Leqs Lvars )
	EQS2R SWAP			( Lvars' P )
	GSOLVE				( Lvars' Lg )
	REDUCESOL			( Lvars' Lg' )
	DUPNULLCOMP? case2drop %0	( --> no solutions )
	ZERO UNROT			( #0 Lvars' Lg' )
	toLEN_DO (DO)			( Meta Lvars' Lg' )
	   DUPINDEX@ NTHCOMPDROP	( Meta Lvars' Lg' Lgi )
	   3PICK R2EQS			( Meta Lvars' Lg' Sgi )
	   4UNROLL ROT#1+UNROT		( Meta' Lvars' Lg' )
	LOOP
	2DROP UNCOERCE			( Meta' )
  ;
;
**********************************************************************
* Name:		GSIMP
* Desc:		Simplify polynomial with respect to side relations
* Stack:	( eq {eqs} {vars} --> eq' )
**********************************************************************
ASSEMBLE
	CON(1)	8
RPL
xNAME GSIMP
::
  CK3&Dispatch
  #A00
  ::	FastLib				( eq Leqs Lvars )
	SWAP DUPTYPELIST? ?SKIP ONE{}N
	SWAP DUPTYPELIST? ?SKIP ONE{}N	( eq Leqs Lvars )
	UNROTSWAP >HCOMP		( Lvars Leq&eqs )
	SWAP EQS2R			( Leq&eqs' Lvars' )
	OVER CARCOMP			( Leq&eqs' Lvars' eq' )
	ROT CDRCOMP			( Lvars' eq' Leqs' )
	GBASIS				( Lvars' eq' G )
	DUP CARCOMP TYPEHSTR? ITE_DROP	( Lvars' eq' )
	REDUCE1				( Lvars' eq'' )
	HXS 2 10 ROT R2A		( eq'' )
  ;
;

**********************************************************************
* Name:		GSOLVE
* Desc:		Calculate factorized Grobner basis
* Stack:	( P --> Lg )
**********************************************************************
NULLNAME GSOLVE
::
  ( Calculate initial basis )
  GBASIS					( G )
  ( Quit if trivial basis )
  DUP CARCOMP TYPEHSTR? ?SEMI			( --> G )
* Should have ONE{}N above?

  ( See if any polynomial in the basis factors                    )
  ( The guess is that the last polynomials should be tested first )

  ZEROSWAP					( #0 G )
  toLEN_DO (DO)
     "FACTORING " INDEX@ #>$ &$ DISPROW3
     DUP ISTOP-INDEX NTHCOMPDROP		( #0 G Gi )
     PFactor DROP				( #0 G Lf )
     DUPLENCOMP #1<>				( #0 G Lf factors? )
     OVER CARCOMP TWO NTHCOMPDROP #1<>		( #0 G Lf factors? power? )
     ORNOT ITE_DROP
     ::						( Does factor - abort )
       UNROT					( Lf #0 G )
       INDEX@ ROTDROPSWAP			( Lf #i G )
       ExitAtLOOP
     ;
  LOOP

  ( Done if no polynomial factorized )

  SWAP DUP#0=csedrp ONE{}N			( --> Lg )

  ( Create subproblems and recurse )		( Lf G&Gi #i )

  SWAPINCOMP get1				( Lf MetaG&Gi #i )
  #1+ROLL DROP#1- {}N				( Lf G )
  NULL{} UNROTSWAP				( Lg G Lf )
  toLEN_DO (DO)
     "SPLIT " INDEX@ #>$ &$ DISPROW3
     DUPINDEX@ NTHCOMPDROP CARCOMP		( Lg G Lf Lfi )
     3PICKSWAP >TCOMP				( Lg G Lf G&Lfi )
     GSOLVE					( Lg G Lf Lgi )
     DUP CARCOMP CARCOMP			( * Ignore trivial sols * )
     TYPEHSTR? ITE_DROP
     :: 4ROLLSWAP &COMP UNROT ;			( Lg' G Lf )
  LOOP
  2DROP						( Lg' )

  ( Don't return a null solution! )

  DUPNULLCOMP? casedrop { { HXS 2 10 } }
;

**********************************************************************
* Name:		REDUCESOL
* Desc:		Remove redundant solutions from solution set
* Stack:	( Lg --> Lg' )
* Notes:	{ 0 } --> {{0}}		always true
*		{ 1 } --> {}		no solutions
**********************************************************************
NULLNAME REDUCESOL
::
  ( Abort if trivial )
  DUP CARCOMP					( Lg Lg1 )
  DupQIsZero? casedrop ONE{}N			( --> {{0}} )
  ZIsOne? casedrop NULL{}			( --> {} )

  ( Sort solutions - shortest ones first because they're )
  ( more likely to be non-zerodimensional                )

  ' :: LENCOMP SWAP LENCOMP #< ;
  SortList

  ( Remove solutions which are redundant - e.g. if the basis )
  ( of the joined systems equals one of the solutions.       )

  {{ Lg }}

  Lg LENCOMP #1+_ONE_DO (DO)
  ::
     Lg INDEX@ NTHELCOMP			( Lgi TRUE / FALSE )
     NOT?SEMI					( *We're done* )
     ISTOP-INDEX #1=case DROP			( *Last solution* )
     ( Compare with all solutions after it )
     ISTOP@ INDEX@ #1+ DO
     ::
        "NON-REDUNDANT: " Lg LENCOMP #>$ &$ DISPROW3
        Lg INDEX@ NTHELCOMP			( Lgi Lgj TRUE / Lgi FALSE )
        NOT?SEMI				( *No more pairs* )
        2DUP &COMP				( Lgi Lgj Lgij )
        GBASIS					( Lgi Lgj G )
        3PICKOVER EQUALcase
        ::					( *Lgi redundant* )
          2DROP					( Lgi )
          Lg INNERCOMP				( Lgi MetaLg )
          DUP JINDEX@ #- #2+ROLL DROP#1-	( Lgi MetaLg' )
          {}N Lg!				( Lgi )
          JINDEX@ #1- JINDEXSTO			( *Removed 1 - no increment* )
          ExitAtLOOP				( *Abort inner loop* )
        ;
        EQUAL NOT?SEMI				( *Not equal to either one* )
        Lg INNERCOMP				( Lgi MetaLg )
        DUPINDEX@ #- #2+ROLL DROP#1-		( Lgi MetaLg' )
        {}N Lg!					( Lgi )
        INDEX@ #1- INDEXSTO			( *Removed 1 - no increment* )
     ;
     LOOP
     DROP					( *Drop Lgi* )
  ;
  LOOP
  Lg ABND					( --> Lg' )
;
**********************************************************************
* Name:		GBASIS2
* Desc:		Calculate Grobner basis from given set of polynomials.
*		Intended for overflow cases - accordingly not optimized
*		since we're not likely to be able to finish many
*		calculations anyway.
* Stack:	( P --> G )
**********************************************************************
NULLNAME GBASIS2
::
  "OVERFLOW - SLOW METHOD" DISPROW3
  GBASIS3
  NULL$ DISPROW2
  DUPNULLCOMP? casedrop { HXS 1 0 }		( *Fix trivial cases* )
;

NULLNAME GBASIS3
::
  ( Handle trivial cases )
  DUPLENCOMP DUPTWO #<case DROP			( --> G )
  ZEROZERO
  {{ G k tmp Gn }}

  ( Create pairs )
  ZERO						( #0 )
  k ONE_DO (DO)					( pairs )
    k #1+ INDEX@ #1+ DO
       JINDEX@ INDEX@ ROT #2+			( pairs' )
    LOOP
  LOOP

  ( Inter-reduce )
  REDUCEG					( pairs' )

  ( Check if interreduction left no pairs or G is now trivial )
  DUP#0= k TWO #< ORcase
  :: NDROP G ABND ;

  ( Main Buchberger loop )
  BEGIN						( pairs )
	DUP#0<>
  WHILE
	"GBASIS: " k #>$ &$
	" PAIRS: " &$ OVER #2/ #>$ &$ DISPROW1
	#2- UNROT				( pairs' #i #j )
	G ROT NTHCOMPDROP			( pairs' #j Gi )
	G ROT NTHCOMPDROP			( pairs' Gi Gj )
	S-POLY					( pairs' spoly )
	G REDUCE1				( pairs' h )
	DupQIsZero? caseDROP			( --> pairs' )
	DUPTYPEHSTR? casedrop			( * Handle no solution case * )
	:: NDROP ZERO				( #0 )
	   { HXS 2 10 } G!			( * G = {1} * )
	   ONE k!				( * k = 1 * )
	;
	PPZ DROP				( * Remove content * )
	G SWAP >TCOMP G!			( * G=G&h * )
	k #1+ k!				( * k=k+1 * )
	k ONE_DO (DO)				( * Add new pairs * )
	   INDEX@ k ROT #2+
	LOOP
	REDUCEG					( * Inter-reduce * )
  REPEAT
  DROP						( * Drop #0 pairs * )
  G ABND					( G )
  SORTBASIS					( G' )
;
*---------------------------------------------------------------------
* Name:		REDUCEG
* Desc:		Inter-reduce basis
* Stack:	( pairs --> pairs' )
* Lams:		G k tmp
*---------------------------------------------------------------------
NULLNAME REDUCEG
::
  ( Handle trivial G )
  k TWO #<case
  :: NDROP ZERO ;				( --> #0 )

  ( Loop until all reductions done )
  BEGIN
     TRUE tmp!					( * No reductions yet * )
     k #1+_ONE_DO (DO)
     ::
	"GBASIS: " k #>$ &$
	" PAIRS: " &$ OVER #2/ #>$ &$ DISPROW1
	"REDUCE: " INDEX@ #>$ DISPROW2
	G INDEX@ NTHELCOMP			( pairs Gi T / pairs F )
	NOTcase ISTOP@				( * No more in this pass * )
        G INNERCOMP				( pairs Gi MetaG )
	DUPINDEX@ #- #2+ROLL DROP#1- {}N	( pairs Gi G-Gi )
	2DUP REDUCE1				( pairs Gi G-Gi Gi' )
	DUPTYPEHSTR? NOTcase
	::					( * Got polynomial back * )
	   PPZ DROP				( pairs Gi G-Gi Gi'' )
	   ROTOVER EQUAL			( pairs G-Gi Gi'' same? )
	   case2drop ONE			( * No reduction * )
	   SWAPINCOMP get1			( pairs MetaG' Gi'' )
	   OVER INDEX@#- #3+ UNROLL #1+		( pairs MetaG'' )
	   {}N G!				( * Update G * )
	   INDEX@ #1= case ONE			( * No restart if 1st poly * )
	   FALSE tmp! ONE			( * Else flag restart * )
	;
	( Reduced to some constant - possibly zero )
	QIsZero? NOTcase2drop
	::					( * Reduce to 1! * )
	  { HXS 2 10 } G!			( * G=1 * )
	  ONE k!				( * k=1 * )
	  NDROP ZEROZERO ExitAtLOOP		( * Clear pairs * )
	;
	SWAPDROP G!				( * Smaller G * )
	k #1- k!				( * k=k-1 * )
	( Update pairs )
	ZERO BEGIN				( pairs #i )
	   2DUP#>
	WHILE
	   #2+					( pairs #i+2 )
	   DUP #2+ROLL OVER #2+ROLL		( pairs #i+2 i j )
	   OVER INDEX@ #=
	   OVER INDEX@ #= ORcase		( * If to be removed * )
	   :: 2DROP #2- SWAP #2- SWAP ;		( pairs' #i )
	   DUPINDEX@ #> IT #1- SWAP		( pairs #i+2 j' i )
	   DUPINDEX@ #> IT #1- SWAP		( pairs #i+2 i' j' )
	   3PICK #2+UNROLL OVER #2+UNROLL	( pairs' )
	REPEAT
	DROPZERO				( pairs' +loop )
     ;  
     +LOOP
     tmp k TWO #< OR				( * See if done * )
  UNTIL
;
*---------------------------------------------------------------------
* Name:		S-POLY
* Desc:		Calculate s-polynomial
* Stack:	( p q --> spoly(p,q) )
*---------------------------------------------------------------------
NULLNAME S-POLY
::
  OVER HTERM					( p q Mp )
  OVER HTERM					( p q Mp Mq )
  2DUP QGcd					( p q Mp Mq GCD )
  DUPTYPEHSTR? case :: 5DROP HXS 1 0 ;		( * Criterion 1 check * )
  DUPUNROT QDiv					( p q Mp GCD Mq/GCD )
  5ROLL QMul					( q Mp GCD Mq/GCD*p )
  4UNROLL QDiv					( Mq/GCD*p q Mp/GCD )
  QMul QSub					( Mq/GCD*p-q*Mp/GCD )
  PTrim						( spoly )
  DupQIsZero? ?SEMI
  PPZ DROP					( remove content )
;
*---------------------------------------------------------------------
* Name:		HTERM
* Desc:		Get highest lexicographic term
* Stack:	( p --> hterm(p) )
*---------------------------------------------------------------------
NULLNAME HTERM
CODE
		GOSBVL	=SAVPTR

* Check if constant input and thus hterm(p) = p

		A=DAT1	A
		D0=A				D0 = ->P
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+
		GOVLNG	=GETPTRLOOP

* Allocate memory for hterm - takes always less than P itself

+		GOSUBL	QSKIP
		CD0EX				C[A] = ->Ptail
		A=DAT1	A			A[A] = ->P
		C=C-A	A			C[A] = Psize
		GOSBVL	=GETTEMP

* Find highest term		

		A=C	A			A[A] = Psize
		GOSBVL	=D1=DSKTOP
		C=DAT1	A			C[A] = ->P
		B=C	A			B[A] = ->P
		A=A+C	A			A[A] = ->Ptail
		CD0EX				D0 = ->P
		D1=C				D1 = ->out
		R0=C	A			R0[A] = ->out
		R1=A	A			R1[A] = ->Ptail

		LC(5)	=DOLIST
-		D0=D0+	5			Skip DOLIST
		A=DAT0	A
		?A=C	A
		GOYES	-			Skip all DOLIST's
		A=B	A
		AD0EX				D0 = ->P
		B=A	A			B[A] = ->hterm

* The following code is generic - it will extract any given term
* Remains of the day when I still had total degree ordering, but
* I figure it might come useful in other contexts too - for example
* for extracting the N'th term

* First output leading DOLIST's

--		A=DAT0	A			Copy DOLIST / DOHSTR
		DAT1=A	A
		D0=D0+	5
		D1=D1+	5
		LC(5)	=DOLIST
		?A=C	A
		GOYES	--			All leading DOLIST's
		LC(2)	=DOHSTR
		?A#C	A
		GOYES	+			SEMI - back up in output
		C=DAT0	A			C[A] = Zlen
		AD0EX
		D0=A
		?A>=B	A
		GOYES	++			Past hterm - done DOLIST's
		A=A+C	A
		D0=A				D0 = ->Ztail
		D1=D1+	5			Total -5 to counter DOHSTR
+		D1=D1-	10			Back up 10 due to SEMI
		GONC	--

* Now the coefficient itself

++		GOSBVL	=MOVEDOWN

* Then zero coefficients and SEMI's until the end is found
		
		C=R1	A
		B=C	A			B[A] = ->Ptail
--		A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		GOYES	+			Accept and copy SEMI
		GOSUBL	QSKIP			Skip {} or Z
		LCZ0				And output Z0 instead
		DAT1=C	11
		D1=D1+	11
		GONC	--
+		DAT1=A	A			Copy SEMI
		D0=D0+	5
		D1=D1+	5
		AD0EX
		D0=A
		?A<B	A
		GOYES	--			Until reached ->Ptail

* Shrink memory and return hterm

		AD1EX
		D0=A				D0 = ->output tail
		C=R0	A			C[A] = ->output
		A=A-C	A
		A=A+CON	A,6
		DAT0=A	A			Fix hterm link
		D0=D0+	5
		GOSBVL	=recover
		GOVLNG	=GPOverWrR0Lp		( --> hterm(p) )
ENDCODE

*---------------------------------------------------------------------
* Name:		SORTBASIS
* Desc:		Sort Grobenr basis to triangular form
* Stack:	( G --> G' )
* Note:		The elements must be different!
*---------------------------------------------------------------------
NULLNAME SORTBASIS
::
  ( Compare objects based on the degree of polynomials )
  '
  CODE
		C=DAT1	A		C[A] = ->Q
		D1=D1+	5
		D=D+1	A
		RSTK=C
		GOSBVL	=PopASavptr	A[A] = ->P
		C=RSTK
		D0=A			D0 = ->P
		D1=C			D1 = ->Q
--		A=DAT0	A		If different prolog then compare them
		C=DAT1	A
		?A#C	A
		GOYES	gbaspred
		LC(2)	=DOLIST		If not list then abort and leave
		?A#C	A		random order
		GOYES	gbaspred
		CD0EX
		RSTK=C			RSTK = ->P
		D0=C
		GOSUBL	QLEN
		R0=A	A		R0 [A] = degree(P)+1
		CD1EX
		D1=C
		D0=C
		GOSUBL	QLEN		C[A] = degree(Q)+1
		A=R0	A		A[A] = degree(P)+1
		?A#C	A
		GOYES	gbaspred	If different then done
		C=RSTK
		D0=C			D0 = ->P
		D0=D0+	5		Advance to next variable
		D1=D1+	5
		GONC	--		And loop until difference occurs
gbaspred	?A<=C	A
		GOYES	+
+		GOVLNG	=GPPushT/FLp
  ENDCODE
  SortList
;
**********************************************************************
* Name:		EQS2R
* Desc:		Convert list of equations to normal polynomial form
* Stack:	( {eqs} {vars} --> Lp Lvars )
**********************************************************************
NULLNAME EQS2R
::
  ( Remove topmost '=' from all objects )
  SWAP INNERCOMP				( {vars} eq1 .. eqn #n )
  DUP ZERO_DO (DO)
	DUP #1+ROLL				( {vars} .. eqn #n eq1 )
	DUPTYPEIDNT? IT ONESYMBN
	DUPTYPELAM? IT ONESYMBN
	DUPTYPECOL? IT
	:: { xi xPI xCONSTANTe }
	   OVER EQUALPOSCOMP
	   #0=?SEMI ONESYMBN
	;
	DUPTYPESYMB? NcaseTYPEERR
	DUP DUPLENCOMP NTHCOMPDROP		( {vars} .. eqn #n eq1 op )
	EQ: x= IT
	:: INNERCOMP swapop: x- SYMBN ;		( {vars} .. eqn #n eq1' )
	SWAP					( {vars} .. eqn eq1' #n )
  LOOP
  {}N						( {vars} {eqs'} )

  ( Collect variables )
  SWAPDUP LENCOMP UNROT				( #vars {eqs'} {vars} )
  1LAMBIND ZERO
  BEGIN
     NEXTCOMPOB
  WHILE
     syminner PVarSub
  REPEAT
  ( Sort variables )
  SWAP 1GETLAM					( {eqs'} #vars {vars'} )
  INNERCOMP get1 SWAPOVER#-
  DUPUNROT #2+UNROLL {}N			( {eqs'} {oldvars} {newvars} )
  RESOROMP IsV>V? SortList
  INNERCOMP top& {}N				( {eqs'} {vars'} )
  1PUTLAM

  ( Transform equations to polynomials )
  ZERO BEGIN
      NEXTCOMPOB
  WHILE
      syminner A2Rsub
      QIsZero? case ErrInfRes
      PTrim UNROT
  REPEAT
  LENCOMP {}N 1GETABND				( Lp Lvars )
;

**********************************************************************
* Name:		R2EQS
* Desc:		Convert polynomial list to algebraics
* Stack:	( Lp Lvars --> Leqs )
**********************************************************************
NULLNAME R2EQS
::
  OVER NULLCOMP? caseDROP			( --> {} )
  SWAP ZERO
  BEGIN
     NEXTCOMPOB
  WHILE
     HXS 2 10 5PICK R2A 4UNROLL
  REPEAT
  SWAPDROP LENCOMP {}N				( --> Leqs )
;

**********************************************************************
* Name:		REDUCE1
* Desc:		Reduce polynomial with respect to set of polynomials
* Stack:	( p G --> q )
**********************************************************************
NULLNAME REDUCE1
::
   ( Handle special cases )
   OVER TYPEHSTR? caseDROP			( * Constant p * )
   DUPNULLCOMP? caseDROP			( * Empty G * )

   ( Try fast code first )

   CODE
		GOSBVL	=SAVPTR
* Initialize flags
		ST=1	sDESTROY		Destroy $ if overflow
		ST=0	sMODULO			Not modulo calculation

* Check polynomials are short

		C=DAT1	A
		D0=C				D0 = ->G
		GOSUBL	QSHORT?
		GONC	redfail
		D1=D1+	5
		C=DAT1	A
		D0=C				D0 = ->P
		GOSBVL	QSHORT?
		GONC	redfail

* Allocate *all* memory

		GOSBVL	=GARBAGECOL
		GOSBVL	=ROOM
		A=C	A
		LC(5)	500			Require 500 nibbles
		?A>=C	A
		GOYES	+
redmem		GOVLNG	=GPMEMERR
redfail		GOVLNG	=GPPushFLoop

+		CSRB.F	A
		C=A-C	A
		R4=C	A			R4[A] = free
		GOSBVL	=MAKE$N

		AD0EX
		R0=A	A			R0[A] = ->G1
	
* Count variables = max[vars(P),vars(Gi)]
		D1=D1+	5
		C=DAT1	A
		D0=C
		D1=D1-	5
		GOSUBL	=CountVars
		R3=A	A			R3[A] = vars(P)
		C=DAT1	A
		D0=C
		GOSUBL	=CountVars
		A=A-1	A			A[A] = vars(Gi)
		C=R3	A
		?A<=C	A
		GOYES	+
		R3=A	A			R3[A] = vars
+
* Convert base polynomials to distributed form

		C=DAT1	A
		D0=C
		D0=D0+	5			D0 = ->G1
		C=R0	A
		D1=C	A			D1 = ->output
redpop		C=R3	A
		B=C	A			B[A] = vars
		GOSUBL	=Distribute
		GOC	redmem1
		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	redpop

* Output 16-nibble stopper for basis polynomials and q=0

		GOSUBL	CSLOT
		C=C+CON	A,16
		A=R4	A
		A=A-C	A			free = free - 16
		R4=A	A
		GONC	+
redmem1		GOTO	redmem
+		GOSBVL	=WIPEOUT

* Pop r=p after that

		AD1EX
		D1=A	
		R1=A	A			R1[A] = ->r
		GOSBVL	=D0=DSKTOP
		D0=D0+	5
		A=DAT0	A
		D0=A				D0 = ->P
		C=R3	A
		B=C	A			B[A] = vars
		GOSUBL	Distribute
		GOC	redmem1
		CD1EX
		GOSBVL	=CSLW5
		C=R4.F	A
		R4=C	W			R4[A1] = ->free

* Reduce r

		GOSUBL	REDUCE

* Output q

		A=R0	A
		A=A-CON	A,10
		R0=A	A			R0[A] = ->$

* Move q up to end of work mem
+		A=A+CON	A,5
		D1=A
		C=DAT1	A
		A=A+C	A
		D1=A				D1 = ->$tail
		AD0EX
		B=A	A			B[A] = ->q
		AD0EX
		GOSUBL	DISTSKIP
		CD0EX
		D0=C
		C=C-B	A			C[A] = qsize
		GOSBVL	=MOVEUP
* Convert it to recursive form

		A=R0	A
		A=A+CON	A,10			A[A] = ->out
		CD1EX
		D0=C				D0 = ->q (new)
		D1=A				D1 = ->out
		C=C-A	A
		R4=C	A			R4[A] = free
		C=R3	A
		B=C	A
		GOSUBL	=Recurse
		GONC	+
		GOTO	redmem
+		AD1EX
		D0=A
		GOSBVL	=Shrink$
		A=R0	A
		A=A+CON	A,10
		GOSBVL	=GETPTR
		D1=D1+	5
		DAT1=A	A
		D1=D1-	5
		GOVLNG	=OverWrTLoop	
  ENDCODE

  ?SEMI

  ( Assembly code failed due to overflow - reduce in SRPL )

* Note: This is the main bottleneck of the SRPL code.
*	Possible improvement would be to to check divisibility
*	in assembly language - shouldn't be too hard.
*	I guess the best solution would be to make the assembly
*	language code split r in half according to which
*	term will reduce, and return the reducing polynomial too.
*	The first half of r would be added to q as is, then the
*	code would follow the reduction below until it is time
*	to call the 'reduction' finder again.

  SWAP NULL{} HXS 1 0 ZERO
  {{ r Mr q i }}
  INNERCOMP					( MetaG )
  BEGIN
      r HTERM Mr!				( Update hterm[r] )
      DUP					( MetaG j )
      BEGIN
          i #1+ i!
          "Slow reduce " i #>$ &$ DISPROW3
          #1-DUP #3+PICK			( MetaG j' f )
          Mr OVER HTERM				( MetaG j' f Mr Mf )
          PPZ 4UNROLL				( MetaG j' fc f Mr Mf )
          QDiv? ITE
          ::
            ROT RCanno				( MetaG j' f M fc )
            q OVER QMul q!			( q=q*fc )
            r QMul UNROT QMul QSub		( MetaG j' r*fc-M*f )
            PTrim r!				( MetaG j' )
            r HTERM Mr!				( Update hterm[r] )
            DROPDUP				( MetaG j'' )
          ;
          2DROP					( MetaG j' )
          DUP#0= r QIsZero? OR
      UNTIL
      DROP
      q Mr QAdd	q!				( q=q+hterm[r] )
      r Mr QSub PTrim r!			( r=r-hterm[r] )
      r QIsZero?
 UNTIL
 NDROP q ABND
 NULL$ DISPROW3
;

**********************************************************************
* Name:		GBASIS
* Desc:		Calculate Grobner basis for given set of polynomials
* Stack:	( P --> G )
**********************************************************************
NULLNAME GBASIS
::
  ( Check trivial cases )
  DUPNULLCOMP ?SEMI				( Empty P )
  PTrim						( Remove zero Pi )
  DupQIsZero? case ONE{}N			( Identically true )
  DUPTYPEHSTR? casedrop { HXS 2 10 }		( No solution )

  ( Try fast algorithm in assembly )

  CODE

* Flag definitions

sRESTART	EQU 7		Flag to indicate restart in inter-reduce
sGOTPAIR	EQU 7		Flag to indicate minimum found in nextpair
sREDUCED	EQU 8		Flag to indicate reduction occurred
sGCD1		EQU 8		Flag to indicate gcd of terms is 1
sMODULO		EQU 9		Flag to indicate modular calculations
sDESTROY	EQU 11		Flag to indicate $ destruction if overflow
RPL

		GOSBVL	=SAVPTR
* Abort if less than 1 polynomial
		A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QLEN
		R2=A	A		R2[A] = k
		C=0	A
		C=C+1	A		C[A] = 1
		?A>C	A
		GOYES	+
		GOVLNG	=GPPushTLoop	Done if k=1

* Check polynomials are short
+		A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QSHORT?
		GOC	+
		GOVLNG	=GPPushFLoop	Not short - failure
+

* Allocate *all* memory - a full restart if GC occurs can be very costly
* in comparison to a single GC call

		GOSBVL	=GARBAGECOL
		GOSBVL	=ROOM
		A=C	A
		LC(5)	500		Require 500 nibbles
		?A>=C	A
		GOYES	+
gbasmem		GOVLNG	=GPMEMERR
+		CSRB.F	A
		C=A-C	A
		R4=C	A		R4[A] = free
		GOSBVL	=MAKE$N
		D0=D0-	5
		A=DAT0	A
		CD0EX
		A=A+C	A		A[A] = ->$tail
		C=C+CON	A,5		C[A] = ->$start
		GOSBVL	=ASLW5
		A=C	A
		R0=A	W		R0[A] = ->$body	R0[A1]=->$tail
	
* Work memory structure:
*
*		G1		first basis polynomial
*		G2
*		...
*		Gk		last basis polynomial
*		BSS 16		16 nibble zero to indicate end of basis
*		q		reduction result
*		r		polynomial being reduced
*		[free]		free memory
*		BSS 10		10 nibble zero to indicate end of pairs
*		pairM		last pair in pairs list
*		...
*		pair1		first pair in pairs list

* Init flags

		ST=1	sDESTROY	Destroy $ if overflow
		ST=0	sMODULO		Not modular calculation

* Init vars

		A=DAT1	A
		D0=A
		GOSUBL	CountVars
		A=A-1	A		G is a list itself
		R3=A	A		R3[A] = vars

* Pop polynomials

		A=DAT1	A
		D0=A
		D0=D0+	5		D0 = ->P1
		A=R0	A
		D1=A			D1 = ->G1 (output)
--		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Distribute
gbasmem0	GOC	gbasmem
		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	--		Pop until end of list

* Output 16 nibble stopper for basis

		C=R4	A
		C=C-CON	A,16
		R4=C	A
		GOC	gbasmem0
		A=0	W
		DAT1=A	W		Output stopper

* Now:
*	R0	->G1	->pairs
*	R2	k
*	R3	vars
*	R4	free


* Create initial list of pairs to end of free memory
* For i=1 to n-1
*   For j=i+1 to n
*     Out [i,j]
*   Next
* Next

		A=R0	W
		GOSBVL	=ASRW5
		D1=A			D1 = ->pairs
		A=0	A	
		A=A+1	A		i=1
--		B=A	A
		B=B+1	A		j=i+1
-		GOSUB	MKPAIR		Allocate 10 nibbles
		C=B	A
		D1=D1-	5
		DAT1=C	A
		D1=D1-	5
		DAT1=A	A		[i,j]
		B=B+1	A		j=j+1
		C=R2	A		k
		?B<=C	A
		GOYES	-		Until j>k
		A=A+1	A		i=i+1
		?A<C	A
		GOYES	--		Until i>=k

		GOSUB	MKPAIR		Allocate 10 nibbles for stopper
		D1=D1-	10
		A=0	W
		DAT1=A	10		Output stopper

* Inter-reduce basis

INTERREDUCE

		GOSUBL	GBASINFO	Display basis info

		ST=0	sRESTART	Flag no restart condition yet
		A=0	A
		A=A+1	A		i=1
		GOSBVL	=ASLW5
		A=R3.F	A
		R3=A	W		R3[A1] = i

INTERLOOP	
		GOSUBL	INTERINFO	Display inter-reduction info

* Find the polynomial to reduce - Gi

		C=R3	W
		GOSBVL	=CSRW5		C[A] = i
		GOSUB	FINDGI		CS if index too high
		GONC	+

* No Gi - restart loop or interreduce done
		?ST=1	sRESTART
		GOYES	INTERREDUCE
		GOTO	INTERDONE

* Set r=Gi and G=G-Gi

+		AD0EX
		B=A	A		B[A] = ->Gi
		AD0EX
		GOSUB	DISTSKIP
		CD0EX
		D=C	A		D[A] = ->Gi+1
		CD0EX
-		A=DAT0	W		Skip to end of basis
		?A=0	W
		GOYES	+
		GOSUB	DISTSKIP
		GONC	-
+		D0=D0+	16		D0 = ->q
		AD0EX
		D1=A			D1 = ->q
		GOSUB	MKSLOT		Set q=0
		GOSBVL	=WIPEOUT	D1 = ->r
		A=B	A
		D0=A			D0 = ->Gi
		C=D	A		C[A] = ->Gi+1
		C=C-A	A		C[A] = size(Gi)
		AR4EX	A		Ensure memory for a duplicate
		?A>=C	A
		GOYES	+
		GOTO	gbasmem		Not enough memory - error
+		AR4EX	A
		RSTK=C	A		RSTK = size(Gi)
		GOSBVL	=MOVEDOWN	Copy Gi to r 
		A=B	A
		AD1EX			D1 = ->Gi
		C=D	A
		D0=C			D0 = ->Gi+1
		C=A-C	A		C[A] = size to move down
		GOSBVL	=MOVEDOWN	       to fill the slow left by gi
		AD1EX			->free
		C=RSTK			C[A] = size(Gi)
		C=A-C	A		C[A] = ->r
		GOSBVL	=ASLW5
		A=R4.F	A
		R4=A	W		R4[A1] = ->free
		R1=C	A		R1[A] = ->r

* Reduce Gi with respect to G-Gi and remove the content of the result

		GOSUB	REDUCE		D0 = ->q
		GOSUB	DISTPPZ		D0 = ->q'

* Dispatch based on q

		A=DAT0	W
		?A#0	W
		GOYES	inter#

* Reduced to zero - remove Gi from the set. The basis is already reduced so
* what remains is removing the pairs containing Gi and decreasing the
* index j for pairs containing Gj with j>i.

inter0		GOSUB	DESLOT		deallocate q=0
		A=R3	A
		GOSBVL	=ASRW5
		B=A	A		B[A] = i
		A=R0	W
		GOSBVL	=ASRW5
		D0=A			D0 = ->pairs (input)
		D1=A			D1 = ->pairs (output)
remloop		D0=D0-	5
		C=DAT0	A		C[A] = J
		D0=D0-	5
		A=DAT0	A		A[A] = I
		?A=B	A		If either index is i then remove pair
		GOYES	remthis
		?C=B	A
		GOYES	remthis
		?A<B	A		Else decrement if index>i
		GOYES	+
		A=A-1	A		A[A] = i-1
+		?C<B	A
		GOYES	+
		C=C-1	A		C[A] = j-1
+		D1=D1-	5		Output the pair which is kept
		DAT1=C	A
		D1=D1-	5
		DAT1=A	A
		?A#0	A
		GOYES	remloop		Until we encountered zero stopper
		GONC	remdone		which is of course < i
remthis		C=R4	A		deallocate pair
		C=C+CON	A,10
		R4=C	A
		GONC	remloop

* Removed the pair - now we re-enter the loop with the same index
* since we removed Gi

remdone		GOTO	INTERLOOP	Re-enter inter-reduction loop
		
* Gi reduced to nonzero value - reinsert Gi to basis

inter#		?ST=0	sREDUCED	If Gi changed then flag a restart
		GOYES	+		of inter-reduction
		ST=1	sRESTART
+		AD0EX
		B=A	A		B[A] = ->q
		AD0EX
		GOSUB	DISTSKIP
		CD0EX
		D=C	A		D[A] = ->qtail
		C=R3	A
		GOSBVL	=CSRW5		C[A] = i
		GOSUB	FINDGI		D0 = ->Gi
		A=B	A		A[A] = ->q
		C=D	A		C[A] = ->qtail
		C=C-A	A		C[A] = qsize
		AR4EX	A		Make sure enough memory for a
		?A>=C	A		copy of the reduced Gi
		GOYES	+
		GOTO	gbasmem
+		AR4EX	A
		RSTK=C			RSTK = qsize
		A=A+C	A		A[A] = ->qtail
		C=C+A	A		C[A] = ->qtail'
		RSTK=C			RSTK = ->qtail'
		D1=C			D1 = ->qtail'
		C=A	A		C[A] = ->qtail
		AD0EX
		C=C-A	A		C[A] = size to move up
		GOSBVL	=MOVEUP		Make room for Gi
		C=RSTK
		D0=C			C[A] = ->qtail'
		C=RSTK
		GOSBVL	=MOVEUP		Move reduced Gi back to basis

* Now continue inter-reduction with i=i+1. We don't bother with a check
* to see if Gi reduced to a constant - if no overflow happens it will
* be detected very fast anyway

		A=R3	W
		GOSBVL	=ASRW5
		A=A+1	A		i=i+1
		GOSBVL	=ASLW5
		A=R3.F	A
		R3=A	W
		GOTO	INTERLOOP

* Inter-reduction has been done

INTERDONE

* Fix free pointer

		A=R0	A
		D0=A			D0 = ->G1
-		A=DAT0	W		Skip to end of basis
		?A=0	W
		GOYES	+
		GOSUB	DISTSKIP
		GONC	-
+		D0=D0+	16		D0 = ->free
		AD0EX
		GOSBVL	=ASLW5
		A=R4.F	A
		R4=A	W		R4[A1] = ->free

* Main Buchberged algorithm loop

		GOSUBL	GBASBLANK	Blank inter-reduction info

BUCHLOOP
		GOSUBL	GBASINFO	Display basis info
		GOSUB	NEXTPAIR	Pick next best pair
		GONC	+		CC if got one
		GOSUBL	GBASINFO	Else we're done - update display
		GOTO	GBASISDONE	to show "PAIRS: 0" and exit

+		GOSUBL	SPOLY		Calculate s-polynomial to ->r location
		GONC	+		Continue if spoly<>0

* Hmm, shouldn't the next one include substaction by CSLOT since q
* is before r ?

buchfix		AD0EX			Else fix things and loop
		GOSBVL	=ASLW5		R4[A1] = ->r = ->free
		A=R4.F	A
		A=A+CON	A,16
		R4=A	W
		GOTO	BUCHLOOP

* Reduce s-polynomial
* But this one is valid since it points to D0. Perhaps we should
* include a call at BUCHLOOP to simply fix ->free and free simply based
* on current gbasis and pairs to make things easier?

+		GOSUB	REDUCE		D0 = ->q
		A=DAT0	W
		?A=0	W
		GOYES	buchfix
		
* S-polynomial reduced to nonzero value

* Add new polynomial to base

		GOSUB	DISTPPZ		Remove content
		A=R0	A
		D0=A			D0 = ->G1
		D=0	A
-		D=D+1	A		Count polynomials in basis
		GOSUB	DISTSKIP
		A=DAT0	W
		?A#0	W
		GOYES	-		D[A] = size(basis)
		AD0EX
		B=A	A		B[A] = ->basis tail
		D1=A			D1 = ->basis tail
		D0=A
		D0=D0+	16		D0 = ->q
		GOSUB	DISTSKIP	D0 = ->qtail
		A=B	A		A[A] = ->basis tail
		A=A+CON	A,16		A[A] = ->q
		CD0EX			C[A] = ->qtail
		D0=A			D0 = ->q
		C=C-A	A		C[A] = qsize
		GOSBVL	=MOVEDOWN	Append q to basis
		AD1EX			Update ->free
		GOSBVL	=ASLW5
		A=R4.F	A
		A=A+CON	A,16		Decallocate q
		R4=A	A

* How about the stopper for the new basis, does it come from
* r=0 automatically? Hmm??

* Hmm, I do think it would be safer & easier to simply recalculate things

* Add new pairs
		A=R0	A
		GOSBVL	=ASRW5
		D0=A			D0 = ->pairs
-		D0=D0-	10		Find end of pairs list
		A=DAT0	A
		?A#0	A
		GOYES	-
		D=D+1	A		k+1
		C=0	A
		C=C+1	A		i=1..k
-		GOSUB	MKPAIR		Allocate pair
		DAT0=C	A
		D0=D0+	5
		CDEX	A
		DAT0=C	A
		D0=D0-	15
		CDEX	A
		C=C+1	A		i=i+1
		?C<D	A
		GOYES	-		Until i>=k
		A=0	W
		DAT0=A	10		The stopper was already allocated
		GOTO	INTERREDUCE	And inter-reduce

* We could insert a test above to see if the new basis polynomial
* has lower degree than any other polynomial in the basis - if so
* then we can ignore inter-reduction.
* Alternatively we could simply reduce polynomials with respect to
* the new polynomial only to see if any reduction cascades can happen,
* but this is a bit too complex in assembly (where would we keep
* the remaining basis with the current code for REDUCE? )


* Buchberger algorithm finished - output basis

GBASISDONE

* Move basis to end of memory

		A=R0	A
		D0=A			D0 = ->G1
-		A=DAT0	W		Find end of basis
		?A=0	W
		GOYES	+
		GOSUB	DISTSKIP
		GONC	-
+		D0=D0+	16		D0 = ->basis tail

* Hmm, could use R0[A1] = ->pairs here

		A=R0	A
		D1=A			D1 = ->G1
		D1=D1-	5
		C=DAT1	A
		AD1EX
		A=A+C	A
		D1=A			D1 = ->$tail
	
		A=R0	A
		CD0EX
		D0=C
		C=C-A	A		C[A] = size(G)
		GOSBVL	=MOVEUP
	
* Then convert to recursive form (embedded in a list)

		AD1EX
		C=R0	A
		D1=C			D1 = ->output
		D0=A			D0 = ->input
		C=C-A	A
		R4=C	A		R4[A] = free
		GOSUB	MKPAIR		Allocate list
		LC(5)	=DOLIST		Output list
		DAT1=C	A
		D1=D1+	5
* We prefer { 0 } over {}

		A=DAT0	W
		?A#0	W
		GOYES	recloop
		LCZ0
		DAT1=C	11
		D1=D1+	11
		GONC	++

recloop		A=DAT0	W		Recurse until terminator
		?A=0	W
		GOYES	++
		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Recurse
		GONC	recloop
		GOTO	gbasmem
++		LC(5)	=SEMI		Output semi (MKPAIR earlier)
		DAT1=C	A
		D1=D1+	5
		AD1EX
		D0=A
		A=R0	A		A[A] = ->{}	
		A=A-CON	A,10
		R0=A	A		R0[A] = ->$
		GOSBVL	=Shrink$
		GOSBVL	=GETPTR
		A=R0	A
		A=A+CON	A,10
		DAT1=A	A		Overwrite with ->{}
		GOVLNG	=PushTLoop

*---------------------------------------------------------------------
* Pick best pair from pairs list.
* If we encounter any pair with GCD=1 then we remove it
* Exit:	CC - got pair		D0 = ->Gi	D1 = ->Gj
*	CS - no pairs left
*---------------------------------------------------------------------

NEXTPAIR	ST=0	sGOTPAIR	No minimum yet
		A=0	A
		A=A+1	A
		R2=A	A		R2[A] = i = 1

* Need 1 slot for LCM

		GOSUB	MKSLOT		Allocate it
		GOSUB	DESLOT		And deallocate it already

* During the loop:
*	R2[A]  = i
*	R2[A1] = mindegree (current)
*	R2[A2] = minpair (current)

PAIRLOOP
		GOSUB	PAIRLCM		LCM for current pair
		GONC	+		CC: Pair exists - continue
		GOTO	CKPAIR		CS: Pairs done - final checks

+		?ST=0	sGCD1		If gcd<>1 then check if new minimum
		GOYES	minpair?

* GCD was 1 - remove pair
rempair		GOSUB	REMPAIR		Remove pair i
		GOTO	PAIRLOOP	And loop
	
* Test if pair is a new minimum. B[A] = total degree(LCM)
minpair?
		?ST=0	sGOTPAIR	If no minimum yet then this one
		GOYES	+		will be the first guess

		C=R2	W
		GOSBVL	=CSRW5		C[A] = mindegree
		?B<C	A
		GOYES	+		Smaller than old minimum!
		C=R2	W
		GONC	++		Not smaller - try next pair

* Set new minimum

+		ST=1	sGOTPAIR	We have minimum pair
		C=R2	A		minpair = i
		GOSBVL	=CSLW5
		C=B	A		mindegree = degree
		GOSBVL	=CSLW5
		C=R2.F	A
++		C=C+1	A		pair++
		R2=C	W
		GOTO	PAIRLOOP	Loop until all pairs tested

* Tested all pairs - now that all gcd=1 pairs have been removed we test
* criterion 2 on the minimum. This way we don't spend too much time on a
* usually unsuccesful (for us anyway?) test which would complicate things
* if done in the actual minimum finder loop. Besides, now that all gcd=1
* pairs are removed the criterion should be more likely to succeed (?)

* If there is a polynomial Gk in the basis not included in the minimum
* pair i,j which divides the LCM, and the associated pairs i,k and j,k
* have been tested (not in pairs list) then the minimum pair is known
* to reduce to zero. Thus we remove it and jump to NEXTPAIR to find a
* new minimum pair. Experimentally if this occurs it tends to occur
* many times in sequence (using beeps to indicate it).

CKPAIR		?ST=0	sGOTPAIR	Found no minimum - thus no pairs left
		RTNYES			Indicate it with carry set

		C=R2	W
		GOSBVL	=CSRW5
		R2=C	W		R2[A1] = minpair
		GOSBVL	=CSRW5
		R2=C.F	A		R2[A] = minpair

		GOSUB	PAIRLCM		Calculate LCM of minpair

* Now test if any Gi divides the LCM

		A=0	A
		A=A+1	A
		R2=A.F	A		R2[A] = i = 1

CRIT2LOOP	C=R2	W
		GOSBVL	=CSRW5		C[A] = pair
		GOSUB	GETPAIRC	A[A] = ->pair
		D1=A
		A=DAT1	A		A[A] = mini
		D1=D1+	5
		C=DAT1	A
		B=C	A		B[A] = minj
		C=R2	A		C[A] = i

		?A=C	A		If i=mini or k=minj then
		GOYES	crit++		Gk is not to be tested
		?B=C	A
		GOYES	crit++

* Test if Gi divides LCM

		GOSUB	FINDGI		D0 = ->Gi
		GONC	+		CC if Gi exists
		GOTO	CRIT2OK		Else we have tested all Gi - accept

+		C=R4	W
		GOSBVL	=CSRW5		C[A] = ->free
		D1=C			D1 = ->lcm
		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
		C=R3	A
		D=C	A		D[A] = vars
-		A=DAT1	A		A[A] = lcm^
		C=DAT0	A		C[A] = Gi^
		D0=D0+	5
		D1=D1+	5
		?A<C	A		If lcm^ < Gi^ then Gi does not
		GOYES	crit++		divide LCM and we try next Gi
		D=D-1	A
		?D#0	A
		GOYES	-		Test all variables
		GONC	++		Gi divides LCM - check pairs

* Try if next Gi divides LCM

crit+++		AR2EX	A
crit++		C=R2	A
		C=C+1	A		i = i+1
		R2=C	A
		GOTO	CRIT2LOOP

* Gi divides LCM - check if pairs have been tested

++		C=R2	W
		GOSBVL	=CSRW5		C[A] = minpair
		GOSUB	GETPAIRC	A[A] = ->pair
		D0=A
		A=DAT0	A		A[A] = mini
		D0=D0+	5
		C=DAT0	A
		B=C	A		B[A] = minj
		C=R4	W
		GOSBVL	=CSRW5
		D0=C			D0 = ->pairs
--		D0=D0-	5
		C=DAT0	A		C[A] = J
		D0=D0-	5
		?C=0	A		If end of pairs then pairs have
		GOYES	crit!		been tested - remove minpair!

		?A=C	A		If J=mini or J=minj then test I too
		GOYES	++
		?B=C	A
		GOYES	++

		AR2EX	A		If J=i then test I too
		?A=C	A
		GOYES	+

		AR2EX	A
		GONC	--		This pair is fine - continue loop

+		AR2EX	A		J=i - test I too
		C=DAT0	A		C[A] = i
		?A=C	A		If equal to mini or minj then
		GOYES	crit++		pair is tested - try next Gi
		?B=C	A
		GOYES	crit++
		GONC	--		Else pair is fine - continue loop

++		C=DAT0	A		J=mini or J=minj - test I too
		AR2EX	A
		?A=C	A		If I=i then pair is tested 
		GOYES	crit+++		- try next Gi

		AR2EX	A
		GONC	--		Else pair is fine - continue loop

* We tested all pairs and found no i,k or j,k pair - thus the spolynomial
* of this pair would reduce to zero. Remove it and restart

crit!		A=R2	W
		GOSBVL	=ASRW5
		R2=A	A		R2[A] = minpair
		GOSUB	REMPAIR		Remove it
*		LC(2)	#C2
*		GOSBVL	=RCKBp
		GOTO	NEXTPAIR	Restart search

* Found a valid minimum pair - return respective polynomials and CC

CRIT2OK		C=R2	W
		GOSBVL	=CSRW5
		R2=C	A		R2[A] = minpair
		GOSUB	GETPAIR		A[A] = ->pair
		D1=A
		C=DAT1	A		C[A] = mini
		GOSUB	FINDGI		D0 = ->Gi
		D1=D1+	5
		C=DAT1	A		C[A] = minj
		AD0EX
		D1=A			D1 = ->Gi
		GOSUB	FINDGI		D0 = ->Gj
		CD0EX
		RSTK=C			RSTK = ->Gj
		CD1EX
		RSTK=C			RSTK = ->Gi
		GOSUB	REMPAIR		Remove minpair
		C=RSTK
		D0=C			D0 = ->Gi
		C=RSTK
		D1=C			D1 = ->Gj
		RTNCC			CC: got pair

*---------------------------------------------------------------------
* Calculate LCM of pair in R2[A] to free memory (allocated elsewhere)
* CS if index of pair is too big
*---------------------------------------------------------------------
		GOSUB	GETPAIR		A[A] = ->pair
		D1=A
		C=DAT1	A		C[A] = i
		?C=0	A
		RTNYES			CS: No such pair
		GOSUB	FINDGI
		AD0EX			A[A] = ->Gi
		D1=D1+	5
		C=DAT1	A		C[A] = j
		D1=A			D1 = ->Gi
		GOSUB	FINDGI		D0 = ->Gj
		B=0	A		totaldeg
		C=R3	A
		D=C	A		D[A] = vars
		C=R4	W
		GOSBVL	=CSRW5		C[A] = ->free
		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
		C=C+CON	A,16
		ST=1	sGCD1		Assume GCD=1
--		RSTK=C			RSTK = ->output
		A=DAT0	A		A[A] = Gi^
		C=DAT1	A		C[A] = Gj^
		D0=D0+	5
		D1=D1+	5
		?A>=C	A		The LCM is the maximum value
		GOYES	+
		ACEX	A		A[A]=lcm  C[A]=gcd
+		?C=0	A		If minimum<>0 then gcd<>1
		GOYES	+	
		ST=0	sGCD1
+		B=B+A	A		totaldeg += lcm^
		C=RSTK			Output lcm^
		CD0EX
		DAT0=A	A
		D0=D0+	5
		CD0EX
		D=D-1	A
		?D#0	A
		GOYES	--		Until all variables passed
		RTNCC			CC: Got LCM

*---------------------------------------------------------------------
* Remove pair with index R2[A]
*---------------------------------------------------------------------
REMPAIR		GOSUB	GETPAIR
		D0=A			D0 = ->pair (input)
		D1=A			D1 = ->pair (output)
-		D0=D0-	10		Move data up
		A=DAT0	10
		DAT1=A	10
		D1=D1-	10
		?A#0	A
		GOYES	-		until terminator found
		C=R4	A
		C=C+CON	A,10		deallocate removed pair
		R4=C	A
		RTN
*---------------------------------------------------------------------
* Calculate location of pair with index R2[A]
*---------------------------------------------------------------------
GETPAIR		C=R2	A		C[A] = i
GETPAIRC	A=R0	W
		GOSBVL	=ASRW5		A[A] = ->pairs
		C=C+C	A
		A=A-C	A
		C=C+C	A
		C=C+C	A
		A=A-C	A		A[A] = ->pair = ->pairs -10*i
		RTN
		
*---------------------------------------------------------------------
* Get indexed polynomial from basis - CS if no such polynomial
*---------------------------------------------------------------------
FINDGI		?C=0	A
		RTNYES			No 0'th polynomial
		A=R0	A
		D0=A			D0 = ->G1
--		A=DAT0	W
		?A=0	W
		RTNYES			CS: No indexed polynomial
		C=C-1	A		index--
		?C=0	A
		GOYES	++		CC: Got ->Gi
		RSTK=C
		GOSUB	DISTSKIP
		C=RSTK
		GONC	--
++		RTNCC			CC: D0 = ->Gi


*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Skip distributed polynomial in D0
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
DISTSKIP	C=R3.F	A		C[A] = vars
		A=C	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		C=C+CON	A,16		C[A] = slotsize
-		A=DAT0	W		Skip until terminator found
		?A=0	W
		GOYES	+
		AD0EX
		A=A+C	A
		AD0EX
		GONC	-
+		D0=D0+	16		Skip terminator
		RTNCC
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Remove content of distributed polynomial in D0
* CS if polynomial is zero
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
DISTPPZ		A=DAT0	W
		A=0	S
		R2=A	W		R2[W] = |lc|
		?A=0	W
		RTNYES			CS: Zero polynomial
		CD0EX
		RSTK=C			RSTK = ->poly
		CD0EX
* Calculate GCD's until all coeffs included or gcd=1

--		GOSUB	CSLOT		C[A] = slotsize
		AD0EX
		A=A+C	A
		AD0EX
		A=DAT0	W
		?A=0	W
		GOYES	++		All coefficients included
		A=0	S		A[W]=abs(coeff)
		C=R2	W		C[W]=oldgcd
		GOSUBL	SHORTGCD
		R2=A	W		R2[W]=newgcd
		A=A-1	W
		?A#0	W
		GOYES	--		Repeat if gcd<>1

* Include sign of lc in gcd

++		C=RSTK
		D=C	A		D[A] = ->poly
		D0=C			D0 = ->poly
		A=DAT0	W		A[S] = sign(lc)
		P=	14
		A=R2.F	WP		A[W] = sign(lc)*content
		P=	0
		R2=A	W		R2[W] = s*content

* Divide coefficients by gcd

		A=A-1	W
		?A=0	W
		GOYES	++		Avoid division if s*content=1
--		A=DAT0	W		Else divide all coefficients
		?A=0	W
		GOYES	++
		C=R2	W
		GOSUBL	ZSDIV
		DAT0=A	W
		GOSUB	CSLOT
		AD0EX
		A=A+C	A
		AD0EX
		GONC	--
++		C=D	A
		D0=C			D0 = ->poly
		RTNCC			CC: Division succesful

*---------------------------------------------------------------------
MKPAIR		AR4EX	A		Allocate a pair = 10 nibbles
		A=A-CON	A,10
		AR4EX	A
		RTNNC
		GOLONG	gbasmem
*---------------------------------------------------------------------
DESLOT		GOSUB	CSLOT		Deallocate a slot
		AR4EX	A
		A=A+C	A
		AR4EX	A
		RTN
*---------------------------------------------------------------------
MKSLOT		GOSUB	CSLOT		Allocate a slot
		AR4EX
		A=A-C	A
		AR4EX
		RTNNC
		GOLONG	gbasmem
*---------------------------------------------------------------------
CSLOT		AR3EX	A		Calculate slotsize
		C=A	A
		C=C+C	A
		C=C+C	A
		C=C+A	A		5*vars
		AR3EX	A
		C=C+CON	A,16		C[A] = slotsize
		RTN


*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Reduce r with respect to basis
* Input:	R0[A]	->G1		1st polynomial in basis
*			->q		an empty slot must follow
*		R1[A]	->r		the polynomial to reduce
*		R3[A]	vars		number of variables
*		R4[A]	free		amount of free memory
*		R4[A1]	->free		address of free memory (after ->r)
* Out:		D0	->q		reduced result
*		sREDUCED indicates if any reductions occurred
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

REDUCE

* Flag no reductions have occurred yet

		ST=0	sREDUCED

* Set R0[A1] to contain ->G1 and use R0[A] to indicate the next polynomial
* in the basis with which we are trying to reduce

		A=R0	A
		GOSBVL	=ASLW5
		R0=A	W

* Make sure q=0

		GOSUB	CSLOT
		A=R1	A		A[A] = ->r
		A=A-C	A		A[A] = ->q
		D1=A
		GOSBVL	=WIPEOUT	

* Main reduction loop

redloop

* Test in order G1-Gk

		GOSUB	SetG1		R0[A] = ->G1

* Inner loop:
*  If Mr divides Mf then
*    r=r-Mr/Mf*f ; f=G1
*  Else
*    f=f++
*  End
*  Until no more f or r=0

reduce1
		A=R1	A
		D0=A			D0 = ->Mr
		A=R0	A
		D1=A			D1 = ->Mf

* Need 1 slot for Mr/Mf

		GOSUB	MKSLOT
		A=R4	W
		GOSBVL	=ASRW5
		B=A	A		B[A] = ->Mr/Mf = ->free

* Check if G empty - must do the check since the gbasis algorithm
* may reduce the basis to a single polynomial - trying to reduce
* that leaves an empty basis behind

		A=DAT1	W
		?A=0	W
		GOYES	badMf		No more polynomials - skip Mr/Mr

* Calculate Mr/Mf to the slot

		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
		B=B+CON	A,16
		C=R3	A
		D=C	A		D[A] = vars
-		A=DAT0	A		A[A] = Mr^
		C=DAT1	A		C[A] = Mf^
		D0=D0+	5
		D1=D1+	5
		A=A-C	A		A[A] = Mr/Mf^
		GOC	badMf		Not divisible - try next f
		C=B	A		Save power
		CD1EX
		DAT1=A	A
		D1=D1+	5
		CD1EX
		B=C	A
		D=D-1	A
		?D#0	A
		GOYES	-		Loop for all variables
		GONC	okMf		Was divisible - continue

* Not valic reduction, try next one
badMf		GOSUB	DESLOT		deallocate Mr/Mf slot
		A=R0	A		A[A] = ->f
		D0=A
		A=DAT0	W		This prevents skipping in an empty
		?A=0	W		basis - otherwise ->q output will
		GOYES	+		not be valid
		GOSUB	DISTSKIP	Skip f
+		C=DAT0	W		Read possible terminator
		AD0EX
		R0=A.F	A		R0[A] = ->f'
		?C=0	W		If terminator then reduction done
		GOYES	+
		GOTO	reduce1		Else try with next f
+		GOTO	reduce+

* Mr/Mf powers ok - calculate coefficients
okMf		ST=1	sREDUCED	flag reduction happened
		A=R1	A
		D0=A			D0 = ->Mr
		A=R0	A
		D1=A			D1 = ->Mf
		A=DAT0	W		A[W] = cr
		C=DAT1	W		C[W] = cf
		GOSUBL	ZSCANNO		"divide"
		R2=C	W		R2[W] = denominator
		C=R4	W
		GOSBVL	=CSRW5
		D1=C			D1 = ->free
		A=-A-1	S
		DAT1=A	W		Set -Mr/Mf for addition purposes
* We work in ring Z, thus we have to multiply q with the denominator

		C=R2	W		C[W] = denominator
		C=C-1	A
		?C=0	W		Avoid multiplication if *1
		GOYES	++
		A=R0	W
		D0=A			D0 = ->f
-		GOSUB	DISTSKIP	Skip to end of basis
		A=DAT0	W
		?A#0	W
		GOYES	-
		D0=D0+	16		D0 = ->q
-		A=DAT0	W		Multiply each coefficient in q
		?A=0	W
		GOYES	++		done
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL
		DAT0=A	W
		GOSUB	CSLOT
		AD0EX
		A=A+C	A
		AD0EX
		GONC	-
++

* Now we calculate denominator*r-Mr/Mf*f

* First we move r and -Mr/Mf to end of work memory and then we overwrite
* the old r with the result. Note that we use ->free and free to calculate
* the end of work memory instead of using the actual $ inside which we
* are working. This prevents the list of pairs used in GBASIS from being
* overwritten. Also note that -Mr/Mf is right after r, thus we manage
* with a single move

		GOSUB	CSLOT
		AD1EX
		A=A+C	A
		D0=A			D0 = ->Mr/Mf tail
		B=A	A		B[A] = ->Mr/Mf tail
		A=R4	A
		A=A+B	A		A[A] = ->free_end
		D1=A			D1 = ->freetail
		C=A-C	A		C[A] = ->Mr/Mf'
		CBEX	A		B[A] = ->Mr/Mf'
		A=R1	A		A[A] = ->r
		C=C-A	A		C[A] = size to move
		GOSBVL	=MOVEUP
		CD1EX
		CD0EX
		CD1EX
		C=B	A
		D=C	A

* Now we calculate the reduced r. During the loop
*	D0	->r
*	D1	->output
*	R0[A]	->fc1
*	R2[W]	denominator
*	D[A]	->Mr/Mf
* Note in particular that R0[A] can be altered since we plan to start
* reductions from G1 again with a SetG1 call later on.


redadd		GOSUB	MKSLOT		Allocate result slot

* Calculate -Mr/Mf*f term into the slot

		A=R0	A
		AD0EX			D0 = ->fc1
		R0=A	A		R0[A] = ->r
		A=DAT0	W		A[W] = lr
		?A#0	W
		GOYES	++		f not traversed yet - do -Mr/Mf*f

* f traversed - output remaining denominator*r
		A=R0	A
		D0=A			D0 = ->r
--		A=DAT0	W		A[W] = lr
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL		lr*denom
		DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		?A=0	W
		GOYES	+		If was at end of r then exit loop
		C=R3	A		C[A] = vars
-		A=DAT0	A		A[A] = r^
		DAT1=A	A		denominator is constant!
		D0=D0+	5
		D1=D1+	5
		C=C-1	A
		?C#0	A
		GOYES	-		Loop until all variables done
		GONC	--
+		GOTO	redaddok	Addition done - re-enter reduction

* f not traversed yet - calculate -Mr/Mf*f
++		CD1EX
		CDEX	A		D[A] = ->output
		CD1EX			D1 = ->Mr/Mf
		C=DAT1	W		C[W] = Lr/Lf
		CD1EX
		CDEX	A		D[A] = ->Mr/mf
		CD1EX			D1 = ->output
		GOSUBL	ZSMUL		A[W] = lr*Lr/Lf
		DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		C=R3	A
		B=C	A		B[A] = vars
		C=D	A
		C=C+CON	A,16		C[A] = ->Mr/Mf power 1
-		A=DAT0	A		A[A] = f^
		D0=D0+	5
		CD1EX
		RSTK=C			RSTK = ->output
		C=DAT1	A		C[A] = Mr/Mf^
		D1=D1+	5
		A=A+C	A		A[A] = f*Mr/Mf^
		C=RSTK
		CD1EX			D1 = ->output
		DAT1=A	A
		D1=D1+	5
		B=B-1	A
		?B#0	A		Multiply all variables
		GOYES	-

* Now see if denom*r is less, greater or equal in lexico ordering
* compared to -Mr/Mf*f term

		AR0EX	A
		AD0EX			D0 = ->r
		AR0EX	A
redaddf		A=DAT0	W		Check if r done
		?A#0	W
		GOYES	+
		GOTO	redadd		If so then vars are ok
+		GOSUB	CSLOT
		AD1EX
		A=A-C	A
		AD1EX			D1 = ->Mr/Mf*f term
		C=R3	A
		B=C	A		B[A] = vars
		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
-		A=DAT0	A		A[A] = r^
		C=DAT1	A		C[A] = Mr/Mf*f^
		D0=D0+	5
		D1=D1+	5
		?A#C	A
		GOYES	redadd#		Different - one must win then
		B=B-1	A
		?B#0	A
		GOYES	-		Test all variables if necessary

* Equal powers - add coefficients, advance r

redadd=		GOSUB	CSLOT
		RSTK=C			RSTK = slotsize
		AD0EX
		A=A-C	A
		AD0EX			D0 = ->r
		AD1EX
		A=A-C	A
		AD1EX			D1 = ->Mr/Mf*f
		A=DAT0	W		A[A] = lr
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL		A[W] = denominator*lr
		C=DAT1	W		C[W] = Lr/Lf*lf
		GOSUBL	ZSADD		A[W] = result coefficient
		C=RSTK
		B=C	A		B[A] = slotsize
		AD0EX
		A=A+B	A		advance r
		AD0EX
		C=R4	A		C[A] = free
		C=C+B	A		deallocate r slot
		?A#0	W
		GOYES	+
		C=C+B	A		cancellation - deallocate output slot
+		R4=C	A		R4[A] = free'
		?A=0	W
		GOYES	+		If cancellation then skip output
		DAT1=A	W		Output coefficient
		AD1EX
		A=A+B	A		Skip output slot
		AD1EX
+		GOTO	redadd		Continue addition loop

* Not equal powers - skip to end of slots

redadd#		B=B-1	A		vars--
		?B=0	A
		GOYES	+
		D0=D0+	5
		D1=D1+	5
		GONC	redadd#

* Now test which term is higher

+		?A>C	A
		GOYES	redadd>		r is greater

* r power is smaller - -Mr/Mf*f term is already in place

redadd<		GOSUB	CSLOT		skip back in r
		AD0EX
		A=A-C	A
		AD0EX
		GOTO	redadd		and continue addition loop
	
* r power is greater - move Mr/Mf*f term up 1 slot and copy r slot to
* the output slot previously occupied by Mr/Mf*f. This way we don't have
* to recalculate Mr/Mf*f terms again - terms coming from r are much simpler
* of course since the only possible multiplier for it is a constant and
* this we don't have to worry about recalculating r data too much.

redadd>		GOSUB	MKSLOT		Allocate -Mr/Mf*f
		B=C	A		B[A] = slotsize
		CD0EX
		RSTK=C			RSTK = ->rtail
		CD0EX
		AD1EX
		D0=A
		A=A+C	A
		D1=A
		GOSBVL	=MOVEUP		Move ->Mr/Mf*f term up
		C=RSTK
		D0=C			D0 = ->rtail
		C=B	A		C[A] = slotsize
		GOSBVL	=MOVEUP		Copy r slot to output

* The actual result is denominator*r so we fix the copied r slot here

		C=B	A
		RSTK=C			RSTK = slotsize
		A=DAT1	W		A[W] = lr
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL
		DAT1=A	W		Output lr*denominator
		C=RSTK			C[A] = slotsize
		AD1EX
		A=A+C	A		A[A] = ->out++
		A=A+C	A		A[A] = -Mr/Mf*f tail
		AD1EX
		AD0EX
		A=A+C	A		A[A] = ->r++
		AD0EX
		A=R4	A
		A=A+C	A		deallocate r slot
		R4=A	A

* Now continue with the known -Mr/Mf*f

		GOTO	redaddf	

* Addition done, fix data pointers and free memory counter
* We allocated a slot which turned out to be zero:	+slot
* We still need the terminator though:			-16
* And deallocate Mr/Mf term too:			+slot
*							-----
*							+2*slot-16
* Hmm, my earlier comments say +2*slot ??

redaddok	AD1EX			A[A] = ->free (new)
		GOSBVL	=ASLW5		A[A1] = ->free (new)
		GOSUB	CSLOT		C[A] = slotsize
		A=R4.F	A		A[A] = free
		A=A+C	A
		A=A+C	A
		R4=A	W		R4[A] = free'	R4[A1] = ->free'

* Restart reduction for new r unless r=0

		A=R1	A
		D0=A			D0 = ->r
		A=DAT0	W
		?A=0	W
		GOYES	reddone		r=0 - we're done
		GOSUB	SetG1		Reduce starting from G1
		GOTO	reduce1		And re-enter reduction loop

* Leading term of r no longer reduces - update as follows
*	q = q + Mr
*	r = r - Mr
* In lex order we can simply flip the leading term of r to the other side
* of the q terminator and update ->r pointer accordingly

reduce+

		A=R1	A
		D0=A			D0 = ->r
		GOSUB	CSLOT
		A=A-C	A
		D1=A			D1 = ->qslot
		GOSBVL	=MOVEDOWN	Append r term to end of q
		A=0	W
		DAT1=A	W		Make sure q is terminated
		C=DAT0	W		C[W] = lr
		AD0EX
		R1=A.F	A		R1[A] = ->r'

* Prepare next loop unless r=0

		?C=0	W
		GOYES	reddone		r=0 - we're done
		GOTO	redloop		Else re-enter reduction loop

* Done - return q

reddone		A=R4	A		deallocate r=0
		A=A+CON	A,16		Should this be +slot-16 ??
		R4=A	A		Damned allocations are a mess..
		A=R0	W		Restore R0[A] to point to basis
		GOSBVL	=ASRW5
		R0=A	A		R0[A] = ->G1
		D0=A			And return ->q in D0
-		A=DAT0	W		Skip to end of basis
		?A=0	W
		GOYES	+
		GOSUB	DISTSKIP
		GONC	-
+		D0=D0+	16		D0 = ->q
		RTN

*---------------------------------------------------------------------
* Set G1 to be tested
*---------------------------------------------------------------------
SetG1		A=R0	W		A[A] = ->f	A[A1] = ->G1
		GOSBVL	=ASRW5
		R0=A.F	A		R0[A] = ->G1
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Calculate s-polynomial of given polynomials
* In:	D0	->Gi
*	D1	->Gj
*	R3	vars
*	R4	free	->free
* Out:	Free memory contains empty slot for q and spoly follows it
*	CS if spoly is zero
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SPOLY

* Calculate GCD of leading coefficients

		A=DAT0	W		A[W] = lci
		C=DAT1	W		C[W] = lcj
		A=0	S
		C=0	S		unsigned GCD
		GOSUBL	SHORTGCD
		R2=A	W		R2[W] = GCD

* Save pointers to leading terms

		AD0EX
		D0=A
		GOSBVL	=ASLW5
		AD1EX
		D1=A
		R1=A	W		R1[A] = ->Mj	R1[A1] = ->Mi

* Allocate empty slot for q

		C=R4	W
		GOSBVL	=CSRW5		C[A] = ->free
		CD1EX
		RSTK=C			RSTK = ->Gj
		GOSUB	MKSLOT
		GOSBVL	=WIPEOUT	q=0
		C=RSTK
		CD1EX			D1 = ->Gj
		D=C	A		D[A] = ->r = ->output

* These flags indicate whether the respective terms have been calculated

sTERMI	EQU 0
sTERMJ	EQU 1

* We need two slots for the above terms

		GOSUB	MKSLOT
		GOSUB	MKSLOT
		ST=0	sTERMI		Neither one calculated yet
		ST=0	sTERMJ

SPOLYLOOP	?ST=1	sTERMI		If we have atleast one term
		GOYES	+		precalculated then we're not
		?ST=1	sTERMJ		done yet
		GOYES	+

		A=DAT0	W		Else check if Gi and Gj have both
		?A#0	W		been traversed
		GOYES	+
		A=DAT1	W
		?A#0	W
		GOYES	+

* spolynomial done - recalculate variables

		C=D	A
		D0=C			D0 = ->spoly tail
		A=0	W
		DAT0=A	W		Terminate spoly
		D0=D0+	16		D0 = ->free (new)
		GOSUB	DESLOT		Deallocate Gi term
		GOSUB	DESLOT		Deallocate Gj term
		AD0EX
		GOSBVL	=ASLW5		A[A1] = ->free (new)
		A=R4.F	A
		A=A-CON	A,16		Allocate the spoly terminator
		AR4EX	W		Note: 2*slot always > 16
		GOSBVL	=ASRW5		A[A] = ->free (old) (aka ->q)
		A=A+C	A		A[A] = ->spoly
		R1=A.F	A		R1[A] = ->spoly (aka ->r) 
		D0=A			D0 = ->spoly
		GOSUB	DISTPPZ		Remove content
		A=DAT0	W
		?A=0	W
		RTNYES			CS: spoly=0
		RTNCC			CC: spoly<>0

* spolynomial not done yet - calculate next term in it

+		GOSUB	Spoli		Establish term from Gi
		GOSUB	Spolj		Establish term from Gj
		CD0EX
		RSTK=C			RSTK = ->Gi
		CD1EX
		RSTK=C			RSTK = ->Gj
		C=D	A
		D0=C			D0 = ->termi
		GOSUB	CSLOT
		C=C+D	A
		D1=C			D1 = ->termj

* Now check which term is greater - or add them if equal variable powers

		C=R3	A
		B=C	A		B[A] = vars
		D0=D0+	16		skip coefficient fields
		D1=D1+	16
--		A=DAT0	A		A[A] = termi^
		C=DAT1	A		C[A] = termj^
		D0=D0+	5
		D1=D1+	5
		?A#C	A
		GOYES	spoly#		Different powers - one is greater
		B=B-1	A
		?B#0	A
		GOYES	--		Test all variables if necessary
		GOTO	spoly=		All equal - add coefficients

* Different powers - output greater term to spoly

spoly#		B=B-1	A		First skip to end of terms though
		?B=0	A
		GOYES	+
		D0=D0+	5
		D1=D1+	5
		GONC	spoly#

* Now dispatch based on the powers

+		?A<C	A
		GOYES	spoly<

* Term from Gi greater - it is already in it's right place at the tail
* of spoly. However we must move term from Gj up 1 slot and flag
* we need a new term from Gi

spoly>		GOSUB	MKSLOT		Allocate new Gj slot
		D=D+C	A		D[A] = ->out'
		GOSBVL	=MOVEDOWN	Move Gj slot up
		ST=0	sTERMI		Flag new term needed
		GOTO	SPOLYNXT	And loop

* Term from Gj greater. We swap the locations of the terms from Gi and Gj
* to get Gj appended to spoly and flag we need a new term from Gj.

spoly<		GOSUB	MKSLOT		Allocate new Gj slot
		AD1EX
		D0=A			D0 = ->tail
		A=A+C	A
		D1=A
		C=C+C	A
		GOSBVL	=MOVEUP		Move Gi *and* Gj slots up
		GOSUB	CSLOT
		D=D+C	A		D[A] = ->out'
		AD0EX
		D1=A
		A=A+C	A
		A=A+C	A
		D0=A
		GOSBVL	=MOVEDOWN	Move Gj slot to old Gi slot
		ST=0	sTERMJ		Flag new term needed
		GOTO	SPOLYNXT	And loop

* Equal powers - add coefficients

spoly=		A=DAT0	W		A[W] = termi coefficient
		C=D	A
		D0=C
		C=DAT0	W		C[W] = termj coefficient
		GOSUBL	ZSADD		A[W] = coeffi+coeffj
		DAT0=A	W		Output it
		ST=0	sTERMI		Flag new terms from Gi Gj needed
		ST=0	sTERMJ
		?A=0	W		*If* cancellation then ignore
		GOYES	SPOLYNXT	the result slot after all
		GOSUB	MKSLOT		Else allocate it
		D=D+C	A		And update ->out'

SPOLYNXT	C=RSTK
		D1=C			D1 = ->Gj
		C=RSTK
		D0=C			D0 = ->Gi
		GOTO	SPOLYLOOP	Loop until Gi and Gj traversed

*---------------------------------------------------------------------
* Calculate spoly term from Gi to first slot in output
* Note:		LCM(term1,term2)/term1 * term
*		For individual powers this means either one of
*		    pow1-pow1+pow	pow1>pow2
*		    pow2-pow1+pow	pow2>pow1
*		The easy way to do it is to calculate pow2-pow1
*		and set it to zero if the result is negative.
*---------------------------------------------------------------------
Spoli		?ST=1	sTERMI		Abort if we already have it
		RTNYES

		A=DAT0	W		Else check if Gi has been traversed
		?A#0	W
		GOYES	++		Not yet - must calculate

		C=D	A		Else clear the Gi term
		CD1EX
		RSTK=C
		GOSUB	CSLOT
		GOSBVL	=WIPEOUT
		C=RSTK
		D1=C
		RTN

++		ST=1	sTERMI		Flag we will have the term after this
		C=R1	A		C[A] = ->Mj
		CD0EX
		A=DAT0	W		A[W] = lcj
		CD0EX
		C=R2	W		C[W] = GCD
		GOSUBL	ZSDIV		A[W] = lcj/GCD
		C=DAT0	W		C[W] = ci
		GOSUBL	ZSMUL		A[W] = lcj/GCD*ci
		C=D	A
		CD1EX
		DAT1=A	W		Output the coefficient
		D1=D1+	16
		CD1EX
		D=C	A
		C=R3	A		Next output the powers
		B=C	A		B[A] = vars
		D0=D0+	16
		A=R1	W
		C=A	A		C[A] = ->Mj
		GOSBVL	=ASRW5		A[A] = ->Mi
		A=A+CON	A,16		skip coefficient fields
		C=C+CON	A,16
--		CD1EX
		RSTK=C			RSTK = ->Gj
		C=A	A		C[A] = ->Mi
		A=DAT1	A		A[A] = Mj^
		D1=D1+	5
		CD1EX
		RSTK=C			RSTK = ->Mj
		C=DAT1	A		C[A] = Mi^
		D1=D1+	5
		A=A-C	A		Divide
		GONC	+
		A=0	A		With underflow check!
+		C=DAT0	A		C[A] = Gi^
		D0=D0+	5
		A=A+C	A		A[A] = result^
		C=D	A		Output the power
		CD1EX
		DAT1=A	A
		D1=D1+	5
		CD1EX
		D=C	A
		C=RSTK
		A=C	A		A[A] = ->Mj
		C=RSTK			C[A] = ->Gj
		CD1EX
		ACEX	A
		B=B-1	A
		?B#0	A
		GOYES	--		Loop for all variables
		GOSUB	CSLOT
		D=D-C	A		D[A] = ->term from Gi
		RTN

*---------------------------------------------------------------------
* Calculate spoly term from Gj to second slot in output
*---------------------------------------------------------------------
Spolj		?ST=1	sTERMJ		Abort if we have it already
		RTNYES

		A=DAT1	W		Else check if Gj has been traversed
		?A#0	W
		GOYES	++

		CD1EX			If so then zero the second slot
		RSTK=C			RSTK = ->Gj
		GOSUB	CSLOT
		CDEX	A
		C=C+D	A
		D1=C
		C=C-D	A
		CDEX	A
		GOSBVL	=WIPEOUT
		C=RSTK
		D1=C
		RTN

++		ST=1	sTERMJ		Flag we will have the term now
		GOSUB	CSLOT
		D=D+C	A
		C=R1	W
		GOSBVL	=CSRW5		C[A] = ->Mi
		CD0EX
		A=DAT0	W		A[W] = lci
		CD0EX
		C=R2	W
		C=-C-1	S		C[W] = negate Gj terms!
		GOSUBL	ZSDIV
		C=DAT1	W		C[W] = cj
		GOSUBL	ZSMUL
		C=D	A
		CD0EX
		DAT0=A	W		Output coefficient
		D0=D0+	16
		CD0EX
		D=C	A
		C=R3	A		Now calculate powers
		B=C	A		B[A] = vars
		D1=D1+	16
		C=R1	W
		A=C	A		A[A] = ->Mj
		GOSBVL	=CSRW5		C[A] = ->Mi
		A=A+CON	A,16		skip coefficient fields
		C=C+CON	A,16
--		CD0EX
		RSTK=C			RSTK = ->Gi
		C=A	A		C[A] = ->Mj
		A=DAT0	A		A[A] = Mi^
		D0=D0+	5
		CD0EX
		RSTK=C			RSTK = ->Mi
		C=DAT0	A		C[A] = Mj^
		D0=D0+	5
		A=A-C	A
		GONC	+
		A=0	A
+		C=DAT1	A		C[A] = Gj^
		D1=D1+	5
		A=A+C	A		A[A] = result^
		C=D	A		Output result power
		CD0EX
		DAT0=A	A
		D0=D0+	5
		CD0EX
		D=C	A
		C=RSTK
		A=C	A		A[A] = ->Mi
		C=RSTK			C[A] = ->Gi
		CD0EX
		ACEX	A
		B=B-1	A
		?B#0	A
		GOYES	--		Loop for all powers
		GOSUB	CSLOT
		D=D-C	A
		D=D-C	A		D[A] = ->term from Gi
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Display information about current basis:
* "GBASIS: xx PAIRS: yy"
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
GBASINFO	A=R0	A
		D0=A			D0 = ->G1
		B=0	A
-		A=DAT0	W		Count polynomials in basis
		?A=0	W
		GOYES	+
		B=B+1	A		basis++
		GOSUBL	DISTSKIP
		GONC	-
+		A=R0	A
		GOSBVL	=ASRW5
		D0=A			D0 = ->pairs
		D=0	A
-		D0=D0-	10		Count pairs
		A=DAT0	A
		?A=0	A
		GOYES	+
		D=D+1	A		pairs++
		GONC	-

+		GOSUB	Disp$		C[A] = ->display string
		R1=C	A		R1[A] = ->display$
		D0=C
		LCSTR	'GBASIS: '	"GBASIS: "
		DAT0=C	16
		D0=D0+	16
		C=B	A		"xx"
		GOSUB	OutDec
		LCSTR	' PAIRS: '	" PAIRS: "
		DAT0=C	16
		D0=D0+	16
		C=D	A		"yy"
		GOSUB	OutDec

		C=R1	A
		D1=C			D1 = ->text
		GOSBVL	=D0->Row1	D0 = ->row1
		D0=D0-	5
		A=DAT0	A
		D0=D0+	5
		GOSBVL	=w->W
		C=A	A
		D=C	A		D[A] = display width
		LC(5)	22		C[A] = chars
		B=0	A		B[A] = X=0
		GOVLNG	=$5x7

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Display inter-reduction info:
* "REDUCE: xx"
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
INTERINFO	GOSUB	Disp$
		D=C	A		D[A] = ->text
		D0=C
		LCSTR	'REDUCE: '	"REDUCE: "
		DAT0=C	W
		D0=D0+	16
		C=R3	W
		GOSBVL	=CSRW5		C[A] = i
		GOSUB	OutDec		"xx"
		C=D	A
		D1=C			D1 = ->text	
		GOSBVL	=D0->Row1	D0 = ->row1
		D0=D0-	5
		A=DAT0	A
		D0=D0+	5
		GOSBVL	=w->W
		C=A	A
		D=C	A		D[A] = display width
		C=C+C	A
		C=C+C	A
		C=C+C	A		Skip 8 rows
		AD0EX
		A=A+C	A
		AD0EX	
		LC(5)	22		C[A] = 22 chars
		B=0	A		B[A] = X = 0
		GOVLNG	=$5x7
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Clear display line 2
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
GBASBLANK	GOSBVL	=D0->Row1	D0 = ->row1
		D0=D0-	5
		A=DAT0	A
		D0=D0+	5
		GOSBVL	=w->W		A[A] = display width
		A=A+A	A
		A=A+A	A
		A=A+A	A		A[A] = size of 8 rows
		CD0EX
		C=C+A	A
		D1=C			D1 = ->textrow2
		C=A	A		C[A] = size of 8 rows
		GOVLNG	=WIPEOUT
*---------------------------------------------------------------------
* Return C[A] = ->text in uart_buffer and clear 22 char area
*---------------------------------------------------------------------
Disp$		D0=(5)	=IRAM@		Get address of uart_buffer
		C=DAT0	S
		D0=(5)	=G_uart_buffer
		C=C+C	S
		GOC	+
		D0=(5)	=uart_buffer
+		LCASC	' '		Clear 22 chars
		LA(2)	22-1
-		DAT0=C	B
		D0=D0+	2
		A=A-1	B
		GONC	-
		LC(5)	44
		AD0EX
		C=A-C	A		C[A] = ->text
		RTN
*---------------------------------------------------------------------
* Convert hex integer in C[A] to decimal and output it to D0
*---------------------------------------------------------------------
OutDec		A=0	W
		A=C	A		A[W] = hex
		GOSBVL	=HXDCW		A[W] = dec
		SETHEX
		?A=0	W
		GOYES	+		No trimming if dec=0
-		P=P-1
		ASL	W
		?A=0	S
		GOYES	-		Shift until highest nibble nonzero
		P=P-1			One less nibble to display
+		C=P	15		C[S] = nibbles to write
		P=	0
-		LCASC	'9'		Conversion loop
		ASLC			Next digit from A[S]
		ACEX	P
		?C<=A	P
		GOYES	+
		C=C+CON	B,7
+		DAT0=C	B
		D0=D0+	2
		C=C-1	S		C[S] nibbles to write
		GONC	-
		RTNCC
  ENDCODE
 
  ( P FALSE / G TRUE )

  case SORTBASIS		( If got result then sort it )

  GBASIS2			( Else call the slow code )
;
**********************************************************************
