**********************************************************************
* Desc:		Linear factor extractor using Newton's method
*		for finding roots.
* Stack:	( P --> Lf P' )
* Notes:
*
*	Original problem:
*
*	   n               n
*	  ---             ---     i
*	k | | (x-r ) =  k >   (a x ) = 0	, r  is rational
*	  | |     i       ---   i                  i
*	  i=0             i=0
*
*	Transformed problem:
*
*	 n             n
*	---           ---     i n-i
*	| | (x-kr ) = >   (a x k   )		, kr  is integer
*	| |      i    ---   i                       i 
*	i=0           i=0
*
*	Transforming roots back is then trivial: divide by k
*
**********************************************************************
NULLNAME PTrialNewton
::
  NULL{} SWAP				( Lf P )
  Univar? NOT?SEMI			( Not univariate - fail )

  BEGIN
	?ATTNQUIT

	DUPLENCOMP THREE #<case		( If already linear )
	:: RDROP ;			( then abort )

	DUP CARCOMP DupZIsOne? ITE	( Dispatch on leading coefficient )

	::				( leading coefficient = 1 )
	  DROP Newton			( Lf P root TRUE / Lf P FALSE )
	  NOTcase :: 2RDROP ;		( No root found - abort )
	  HXS 2 10 SWAP			( Lf P Z1 root )
	;

	::
	  OVER INNERCOMP		( Lf P Zk Meta )
	  get1 DUP3PICK			( Lf P Meta Zk Zk #n )
	  ONE_DO (DO)			( Lf P Meta Zk Zk^i )
		ISTOP@ #2+ROLL
		OVER QMul 4UNROLL	( Lf P Meta' Zk Zk^i )
		OVER QMul		( Lf P Meta' Zk Zk^i+1 )
	  LOOP
	  DROP OVER#2+UNROLL {}N	( Lf P Zk P' )
	  Newton			( Lf P Zk P' root TRUE / .. )
	  NOTcase2drop :: 2RDROP ;	( No root found - abort )
	  SWAPDROP			( Lf P Zk root )
	  2DUP ZGcd			( Lf P Zk root Zgcd )
	  ROTOVER QDiv UNROT QDiv	( Lf P Zk' Zr' )
	;

	QNeg TWO{}N			( Lf P Pf )
	ROTOVER >TCOMP UNROT		( Lf' P Pf )
	QDiv
  AGAIN
;

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
* Desc:		Find root for univariate polynomial using
*		Newton's method
* Stack:	( P --> P root TRUE / P FALSE )
* Notes:
*		Note how EXITLOOP is done within Newton so that the
*		profiler won't be confused when NewtonClose is executed.
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
NULLNAME Newton
::
  DUP PDer HXS 2 10				( P D Z )

  FIFTY ZERO_DO (DO)				( Try max 50 iterations )

	3PICKOVER PEval				( P D Z Pz )
	DupQIsZero? case			( Found root - exit )
	:: EXITLOOP DROPSWAPDROP TRUE ;		( --> P Z TRUE )

	3PICK3PICK PEval			( P D Z Pz Dz )
	DupQIsZero? casedrop			( Found extrema - abort )
	:: EXITLOOP 3DROP FALSE ;		( --> P FALSE )

	QDiv					( P D Z Pz/Dz )
	DupQIsZero? casedrop			( Handle very small step )
	:: EXITLOOP NewtonClose ;
	QSub					( P D Z' )

  LOOP
  2DROPFALSE					( P FALSE )
;

* ( P D Z --> ? )
NULLNAME NewtonClose
::
   SWAPDROP					( P Z )
   HXS 2 10 QAdd 2DUP PEval			( P Z+1 P[Z+1] )
   QIsZero? caseTRUE
   HXS 2 20 QSub 2DUP PEval			( P Z-1 P[Z-1] )
   QIsZero? caseTRUE
   HXS 2 30 QAdd 2DUP PEval			( P Z+2 P[Z+2] )
   QIsZero? caseTRUE
   HXS 2 40 QSub 2DUP PEval			( P Z-2 P[Z-2] )
   QIsZero? caseTRUE
   DROPFALSE

;
**********************************************************************
