	STITLE	ALG P Factor
**********************************************************************
*		ALG library: Polynomial Factorization Routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*  PFactor	( P --> Lf Z )
*  PSqff	( P --> Lf )
*  PSQFF	( P --> Lsqff )
*
**********************************************************************


**********************************************************************
* Desc:		Do a complete factorization of P
*		The result is trimmed.
*
* Stack:	( P -> Lfk Z )
*
*		where
*
* 		Lfk = { {F1 k1} {F2 k2} ... {Fn kn} }
*		Fi  = non-const factor i
*		ki  = power of factor i
*	        Z   = constant factor
*
**********************************************************************
NULLNAME PFactor
::
  ( * remove integer content * )

  PPZ SWAP				( Z P' )

  ( * full factorization of P * )
 
  TRUE 
  { LAM FullFact? }
  BIND
    PHFctr				( Z Lf )
  ABND  
  SWAP					( Lf Z )
;

**********************************************************************
* Desc:		Square-free & trivial factorization, including 
*		integer content, of P taken positive
*
* Stack:	( P -> Lfk )
*
*		where
*
* 		Lfk = { {F1 k1} {F2 k2} ... {Fn kn} }
*		Fi  = const factor i
*		ki  = power of factor i
*
* Note:		Factors of same power are not necessarily merged or
*		adjacent, but all Fi's are square-free 
**********************************************************************
NULLNAME PSqff
::
  FALSE
  { LAM FullFact? }
  BIND
    PHFctr				( Z Lf )
  ABND  
;

**********************************************************************
* Desc:		Heuristic factorization of polynomial taken positive
*		LAM FullFact? must be bound
*		
*		LAM FullFact? = TRUE   => full factorization
*		LAM FullFact? = FALSE  => square-free & trivial 
*					  factorization only
* Stack:	( P --> Lf )   
*		
*		Lf = { {F1 #k1} ... {Fn #kn} }
* 
**********************************************************************
NULLNAME PHFctr
::
   DUP PSetSign			( |P| )
   DUPTYPELIST? NOTcase PNFctr  ( -> Lf )

   ( * remove trivial factors * )

   PFactTriv SWAP		( Lft P' )

   ( * factor primitive part * )

   PPP SWAP			( Lft PC PP )
   PHFctr1 SWAP			( Lft Lfpp PC )

   ( * done if content is unit * )

   DupZIsOne? casedrop &COMP	( -> Lft&Lfpp )

   ( * factor content and make factors * )
   ( * constant in main variable       * )

   PHFctr			( Lft Lfpp Lfpc )

   FIVE				( Lft Lfpp Lfpc offset )
   BEGIN
     NEXTCOMPOB
   WHILE			( Lft Lfpp Lfpc offset pi )
     INCOMPDROP SWAP		( Lft Lfpp Lfpc offset #ki fi )
     ONE CKMkP SWAP		( Lft Lfpp Lfpc offset fi' ki )
     TWO{}N UNROT		( Lft Lfpp pi' Lfpc offset )
   REPEAT			( Lft Lfpp p1'..pn' Lfpc )
   LENCOMP {}N			( Lft Lfpp Lfpc' )

   ( * combine all factors * )

   SWAP &COMP &COMP		( Lf )
;

**********************************************************************
* Desc:		Heuristic factorization of primitive polynomial
*		LAM FullFact? must be bound
*		
*		LAM FullFact? = TRUE   => full factorization
*		LAM FullFact? = FALSE  => square-free & trivial 
*					  factorization only
* Stack:	( P --> Lf )   
*		
*		Lf = {} 			if P is constant
*		Lf = { {F1 #k1} ... {Fn #kn} }  o/w
* 
**********************************************************************
NULLNAME PHFctr1
::
  DUPTYPELIST? NOTcasedrop NULL{}	( -> {} )

  ( * square-free factorization * )

  PSQFF					( Lsqf )

  ( * factor non trivial square-free factors * )

  ZEROSWAP				( Meta Lsqf )
  toLEN_DO (DO) 
    ::
      DUPINDEX@ NTHCOMPDROP		( Meta Lsqf Fi )
      DUPTYPELIST? NOTcaseDROP

      LAM FullFact? ITE PHFctr0 ONE{}N	( Meta Lsqf Lfi )

      ( * add factors to Meta * )
      
      FIVE				
      BEGIN				( Meta Lfsq Lfi offset )
        NEXTCOMPOB
      WHILE				( Meta Lfsq Lfi offset fij )
        INDEX@ TWO{}N			( Meta Lfsq Lfi offset {fij i} )
        5UNROLL 4ROLL #1+ 4UNROLL	( Meta' Lfsq Lfi offset )
      REPEAT
      DROP				( Meta Lfsq )
    ;
  LOOP
  DROP {}N				( Lf )  
;

**********************************************************************
* Desc:		Heuristic factorization of primitive square-free 
*		non constant polynomial
* Stack:	( P --> Lf ) 
*		
*		Lf = { F1 .. Fm }
**********************************************************************
NULLNAME PHFctr0
::
  DUPLENCOMP THREE #<case ONE{}N	( deg<2 -> { P } )

  ( * remove linear factors * )

  Univar? ITE UFactLin PFactSpl		( Lf P' )

  DUPTYPELIST? NOTcaseDROP		( -> Lf )
  DUPLENCOMP FIVE #<case >TCOMP		( deg < 4 -> Lf&{P} )	

  ( * Z-adic * )

  DUP PNMax ZAbs			( Lf P pmax )		
  HXS 2 20 QMul				( Lf P 2*pmax )
  HXS 2 40 QAdd				( Lf P z )
  2DUP PEval				( Lf P z Pz )
  PHFctr				( Lf P z Lfz )
  LiftFact				( Lf Lf1 )
  &COMP					( Lf' )
;

**********************************************************************
* Desc:		Factorization of positive integer as polynomial
* Stack:	( Z -> Lf )  
* Note:				
*		Lf = {} 			if Z is 1
*		Lf = { {Z1 #k1} ... {Zn #kn} }  o/w
**********************************************************************
NULLNAME PNFctr
::
  DupQIsZero? caseSIZEERR
  DupZIsOne? casedrop NULL{}		( -> {} )

  NFactorSpc				( L )
  ZEROSWAP				( #0 L )
  INNERCOMP				( #0 l1..ln #n )
  BEGIN					( Meta l1..ln #n )
    #1-SWAP ONE				( Meta l1..li #i z 1 )
    BEGIN				( * count duplicates * )
      OVER5PICK EQUAL			( Meta l1..li #i z 1 flag )
    WHILE				( Meta l1..li-1 z #i z #k )
      SWAPDROPSWAP			( Meta l1..li-1 z #k #i )
      #1-UNROT #1+			( Meta l1..li-1 #i-1 z #k+1 )
    REPEAT				( Meta l1..li #i z #k )
    TWO{}N				( Meta l1..li #i f={z #k} )
    psh1&rev				( Meta&f l1..li #i )
  #0=UNTIL (UNTIL)			( Meta #0 )
  DROP {}N				( Lf )
;	

**********************************************************************
* Desc:		Compute the square-free factorization of primitive P
* 		The result is trimmed (provided P was)
* Stack:	( P --> Lsqff )
**********************************************************************
NULLNAME PSQFF
::
  ONESWAP			( #1 P )
  DUP PDer			( #1 P P' )
  OVER QGcd			( #1 P R1 )
  SWAPOVER			( #1 R1 P R1 )
  QDiv SWAP			( #1 T1 R1 )
  BEGIN				( S1..Si-1 i Ti Ri )
	?ATTNQUIT
	DUPTYPELIST?		( S1..Si-1 i Ti Ri flag )
  WHILE
  	ROT#1+UNROT		( S1..Si-1 i+1 Ti Ri )
	2DUP QGcd		( S1..Si-1 i+1 Ti Ri Ti+1 )
	SWAPOVER QDiv		( S1..Si-1 i+1 Ti Ti+1 Ri+1 )
	ROT 3PICK		( S1..Si-1 i+1 Ti+1 Ri+1 Ti Ti+1 )
	QDiv			( S1..Si-1 i+1 Ti+1 Ri+1 Si )
	4UNROLL			( S1..Si-1 Si i+1 Ti+1 Ri+1 )
  REPEAT			( S1..Se-1 e Te Re )
  DROPSWAP {}N			( Le )
;

**********************************************************************
* Name:		UFactLin
* Desc:		( P --> Lf P' )		
* Notes:
*		IPRZL algorithm by R. Loos, "Rational zeros of 
*		integral polynomials by p-adic extensions", 
*		SIAM J. of Computing 12, pp 286-293 (1983)
*
*		P must be square-free univariate
**********************************************************************

NULLNAME UFactLin
::
  DUP CARCOMP				( A lc )
  OVER DUPLENCOMP NTHCOMPDROP		( A lc tc )
  3PICK PDer				( A |c tc B )
  HXS 2 B0				( A lc tc B p )

  ZEROZERO DUPDUP			( A lc tc B p Ap Bp bf pj )

  {{ A lc B p Ap Bp bf pj }}

* Search suitable prime and find roots

  BEGIN					( - )
     p					( p )
     BEGIN
        Prime+ DUP lc ZGcd		( p gcd[p,lc] )
	ZIsOne?			
     UNTIL

     DUP p!				( p )
     A OVER QMod Ap!			( * compute A mod p * )
     B OVER QMod Bp!			( * compute B mod p * )

     ZEROSWAP				( Meta p )
     HXS># ZERO_DO (DO)			( [From 0 to p-1] )
       ::
          INDEX@ #>Z			( Meta i )
	  Ap OVER p PEvalMod		( Meta i Ap[i] )
	  QIsZero? NOTcaseDROP		( -> Meta )
	  Bp OVER p PEvalMod		( Meta i Bp[i] )
	  QIsZero? NOTcase SWAP#1+	( -> Meta&i )

*
* Unlucky homomorphism:  A[i] mod p AND A'[i] mod p have a common zero 
* 						    
* -> discar work so far and try with next prime
*
	  DROP NDROPFALSE		( FALSE )
	  ZEROISTOPSTO			( exit DO-loop )
       ;				
     LOOP  				( Meta / FALSE )
     DUPTYPEBINT? IT TRUE		( Meta TRUE / FALSE )
  UNTIL

* Exit if no roots were found

  DUP#0=csedrp :: NULL{} A ABND ;	( --> {} A )

* Lift the roots

  lc tc QMul HXS 2 20 QMul ZAbs bf!

  BEGIN
	p bf ZNLT?
  WHILE
	p DUP QMul pj!			( pj = p^2 )
	A pj QMod Ap!			( Ap = A mod pj )
	B p QMod Bp!			( Bp = B mod p )
	DUP ZERO_DO (DO)
	   DUP #1+ROLL
	   Ap OVER pj PEvalMod
	   p QDiv			( Ap[root] mod pj / p )
	   Bp 3PICK p PEvalMod		( Bp[root] mod p )
	   p ModDiv			( modular division )
	   p QMul QSub			( root' )
	   SWAP
	LOOP
	pj p!				( p=pj )
  REPEAT

* Check which ones are actual factors of A

  DUP ZERO_DO (DO)
      ::
	DUP #1+ROLL			( root )
	lc p QMod			( lc mod p )
	lc ROT QMul p QMod QNeg		( -lc*root mod p )
	TWO{}N PPZ DROP			( Pf )
	lc OVER CARCOMP QMod		( Pf lc mod Pflc )
	QIsZero? NOTcasedrop #1- 	( * not a root * )
        A OVER QDiv?			( Pf [ a' TRUE / FALSE ] )
	NOTcasedrop #1-			( * not a root * )
	A! SWAP
      ;
  LOOP

* Return found factors and remaining A

  {}N A PTrim
  ABND
;

**********************************************************************
* Desc:		Remove simple (linear) factors from P
* Stack:	( P -> L P' )
* Notes:	Finds linear factors of the form AX+B where
*			X is current main variable
*			A/B are constants/polynomials
**********************************************************************
NULLNAME PFactSpl		

DEFINE @sfP   13GETLAM		( P to factor )
DEFINE !sfP   13PUTLAM
DEFINE @sfr1  12GETLAM		( Number of factors in lead coefficient )
DEFINE !sfr1  12PUTLAM
DEFINE @sfF1  11GETLAM		( Factors of lead coefficient )
DEFINE !sfF1  11PUTLAM
DEFINE @sfM1  10GETLAM		( Multiplicities of lead factors )
DEFINE !sfM1  10PUTLAM
DEFINE @sfn1  9GETLAM		( Number of unique lead factors )
DEFINE !sfn1  9PUTLAM
DEFINE @sfr2  8GETLAM		( Number of factors in last coefficient )
DEFINE !sfr2  8PUTLAM
DEFINE @sfF2  7GETLAM		( Factors of last coefficient )
DEFINE !sfF2  7PUTLAM
DEFINE @sfM2  6GETLAM		( Multiplicities of last factors )
DEFINE !sfM2  6PUTLAM
DEFINE @sfn2  5GETLAM		( Number of unique last factors )
DEFINE !sfn2  5PUTLAM
DEFINE @sfb1  4GETLAM
DEFINE !sfb1  4PUTLAM
DEFINE @sfb2  3GETLAM
DEFINE !sfb2  3PUTLAM
DEFINE @sfL   2GETLAM		( List of found factors )
DEFINE !sfL   2PUTLAM

::
  DUP CARCOMP				( P c1 )
  PHFctr LCGetCnt			( P r1 F1 M1 )
  DUPLENCOMP				( P r1 F1 M1 n1 )
  5PICK DUPLENCOMP NTHCOMPDROP		( P r1 F1 M1 n1 c2 )
  PHFctr LCGetCnt			( .. n1 r2 F2 M2 )
  DUPLENCOMP				( .. n1 r2 F2 M2 n2 )
  ZEROZERO NULL{}			( .. n2 b1 b2 L )
  TWELVE DUP CACHE			( - )
  ::
   BEGIN
    @sfM1 @sfn1 @sfb1 ONE		( M1 n1 b1 1 )
    CbInit				( C1 )
    BEGIN				( C1 )
    ::
     @sfF1 CbProd			( C1 P1 )
     ZERO !sfb2				( C1 P1 )
     BEGIN				( C1 P1 )
      @sfM2 @sfn2 @sfb2 ONE		( C1 P1 M2 n2 b2 1 )
      CbInit				( C1 P1 C2 )
      BEGIN				( C1 P1 C2 )
      ::
	?ATTNQUIT
        @sfF2 CbProd			( C1 P1 C2 P2 )
        @sfP 4PICK ROT			( C1 P1 C2 P P1 P2 )
        PDivSpl?			( C1 P1 C2 [Q P' T/F] )
        NOTcase 
         :: 
            @sfM2 @sfn2			( C1 P1 C2 M2 n2 )
	    CbNext 			( repeat with next C2 )
         ;    
        @sfL ROT			( C1 P2 C2 P' L Q )
        >TCOMP !sfL			( C1 P2 C2 P' )
        DUP !sfP			( C1 P2 C2 P' )
	LENCOMP THREE #<case 
         ::				( deg[P] < 2 )
            3DROP			( - )
            3RDROP 3RDROP		( exit all loops )
         ;
        @sfr1 @sfb1 #- !sfr1		( update r1 )
        @sfr2 @sfb2 #- !sfr2		( update r2 )
        @sfM2 OVER CbRm !sfM2		( update M2 )
        @sfM1 4PICK CbRm !sfM1		( update M1 )
        3PICK @sfM1			( C1 P2 C2 C1 M1' )
        CbValid?			( C1 P2 C2 flag )
        case
         ::				( C1 still valid )  
            @sfM2 @sfn2			( C1 P2 C2 M2 n2 )
            @sfb2 @sfr2			( C1 P2 C2 M2 n2 b2 r2 )
            CbReset			( repeat with reset C2 )
         ;
        2DROP				( C1 )
        3RDROP				( break out C2 loops )
        @sfM1 @sfn1 @sfb1 @sfr1		( C1 M1 n1 b1 r1 )
        CbReset				( repeat with reset C1 )  
      ;
      NOT_UNTIL (UNTIL)			( C1 P1 )
      @sfb2 #1+DUP !sfb2		( C1 P1 b2 )
      @sfr2 #>				( C1 P1 flag )
     UNTIL
     DROP				( C1 )
     @sfM1 @sfn1			( C1 M1 n1 )
     CbNext				( [C1' TRUE / FALSE] )
    ;
    NOT_UNTIL (UNTIL)
    @sfb1 #1+DUP !sfb1			( b1 )
    @sfr1 #>				( flag )
   UNTIL				( - )
  ;
  @sfL @sfP				( L P' )
  ABND  
;

**********************************************************************
* Desc:		Check whether P is divisible by linear factor aX+/-b
* Stack:	( P a b -> Q P' TRUE / FALSE )
**********************************************************************
NULLNAME PDivSpl?
::                         
  THREE NDUP TWO{}N		( P a b P Q )
  QDiv?				( P a b [P' TRUE/FALSE] ) 
  case ::
          4UNROLL		( P' P a b )
          TWO{}N		( P' P Q )
          UNROTDROP		( Q P' )
          TRUE			( return Q P' TRUE )
        ; 
  QNeg TWO{}N  			( P Q' )
  SWAPOVER			( Q' P Q' )
  QDiv? caseTRUE		( return Q' P' TRUE )
  DROPFALSE			( FALSE )
;

**********************************************************************
* Desc:		Get total, count-list and fctr-list from Lfk
* Stack:	( Lfk --> #m Lf Lc )  
*
*		Lfk = { {F1 #k1} ... {Fn #kn} }
*		Lf  = { F1 ... Fn }
*		Lc  = { #k1 ... #kn }
*		#m  = #k1 + ... + #kn 
*		
**********************************************************************
NULLNAME LCGetCnt
::
   DUP FIVE
   BEGIN			( L ... L offset )
     NEXTCOMPOB
   WHILE			( l ... l offset {fi ki} )
     CARCOMP UNROT		( l ... fi l offset )
   REPEAT			
   LENCOMP {}N			( l lf )
   SWAP FIVE			( lf l )
   BEGIN			( lf ... l offset )
     NEXTCOMPOB
   WHILE			( lf ... l offset {fi ki} )
     TWO NTHCOMPDROP UNROT	( lf ... ki l offset )
   REPEAT			
   LENCOMP {}N			( lf lc )
   ZEROSWAP FIVE		( lf #0 lc start-offset )
   BEGIN			( lf #m lc offset )
     NEXTCOMPOB
   WHILE			( lf #m lc offset #k )
     4ROLL #+ UNROT		( lf #m' lc offset )
  REPEAT			( lf #m' lc )
  ROTSWAP			( #m lf lc )
;

**********************************************************************
* desc:		lift n-1 z-adic factorization into n factorization
* stack:	( p z R --> L )  
*		
*		R = { {f1 #k1} ... {fn #kn} }
*		L = { f1 .. fm }
**********************************************************************
NULLNAME LiftFact

DEFINE @lfcfact	13GETLAM
DEFINE @lfcP	12GETLAM
DEFINE !lfcP	12PUTLAM
DEFINE @lfcz	11GETLAM
DEFINE @lfcr	10GETLAM
DEFINE !lfcr	10PUTLAM
DEFINE @lfcF	9GETLAM
DEFINE @lfcM	8GETLAM
DEFINE !lfcM	8PUTLAM
DEFINE @lfcn	7GETLAM
DEFINE @lfcz/2	6GETLAM
DEFINE @lfcb	5GETLAM
DEFINE !lfcb	5PUTLAM
DEFINE @lfct	4GETLAM
DEFINE !lfct	4PUTLAM
DEFINE @lfcs	3GETLAM
DEFINE !lfcs	3PUTLAM
DEFINE @lfcL	2GETLAM
DEFINE !lfcL	2PUTLAM

::				( P z R )
  RESOROMP LFCtest		( avoid excessive SKIPOB/DOROMP for speed )
  4UNROLL LCGetCnt		( sub P z r F M )
  DUPLENCOMP			( sub P z r F M n )
  5PICK HXS 2 20 QDiv		( sub P z r F M n z/2 )
  ONEONE ZERO NULL{}		( sub P z r F M n z/2 b t s L )

  TWELVE DUP CACHE		( - )
    BEGIN
      @lfcb #2*			( 2b )
      @lfcr 2DUP#=		( 2b r flag )
      IT :: ONE !lfcs ;		( 2b r )
      #> 			( flag )
    NOT WHILE
      @lfcb ONE LFCInit		( C )
      BEGIN
	@lfcfact EVAL		( C' TRUE | FALSE )
      NOT_UNTIL (UNTIL)
      @lfcb #1+ !lfcb
    REPEAT
    @lfcL @lfcP			( L P )
    DUP PSetSign		( L |P| )
    >TCOMP			( L' )
  ABND
;

*---------------------------------------------------------------------
NULLNAME LFCtest
::
   ?ATTNQUIT
   DUP @lfct NTHCOMPDROP	( if 2b=r and Ct=0 )
   @lfcs #< casedrpfls		( then exit the loop )
   @lfcF CbProd			( C P )
   DUP PNMax			( C P pmax )
   @lfcz/2 ZNLT?		( C P flag )
   casedrop LFCNext		( repeat with Cnext )
   @lfcz PZadic			( C Q )
   @lfcP PLCZ			( C Q lp )
   OVER PLCZ			( C Q lp lq )
   QDivRem SWAPDROP		( C Q rem[lp,lq] )
   QIsZero? NOT			( C Q flag )
   casedrop LFCNext		( repeat with Cnext )
   @lfcP OVER			( C Q P Q )
   QDiv? NOT			( C Q [P' FALSE / TRUE] )
   casedrop LFCNext		( repeat with Cnext )
   PTrim !lfcP			( C Q )
   DUP PSetSign			( C |Q| )
   @lfcL SWAP			( C L |Q| )
   >TCOMP !lfcL			( C )
   @lfcM OVER			( C M C )
   CbRm DUP !lfcM		( C M' )
   ZERO ' #<> POSCOMP		( C #t )
   !lfct			( C )
   @lfcr @lfcb #-		( C r' )
   DUP !lfcr			( C r' )
   @lfcb #2*			( C r' 2b )
   2DUP#=			( C r' 2b flag )
   IT :: ONE !lfcs ;		( C r' 2b )
   #< casedrpfls		( exit with FALSE )
   LFCReset			( C' TRUE / FALSE )
;
*---------------------------------------------------------------------

NULLNAME CbRm			( M C --> M' )
::
  DUPLENCOMP			( M C n )
  DUP#0=csedrp DROP		( return {} )
  #1+_ONE_DO (DO)		( M C )
     OVER INDEX@ NTHCOMPDROP	( M C mi )
     OVER INDEX@ NTHCOMPDROP	( M C mi ci )
     #- UNROT			( ai M C )
  LOOP				( a1..an M C )
  DROP LENCOMP {}N		( M' )
;

*---------------------------------------------------------------------

NULLNAME CbProd		( C F --> C P )
* Powers --> polynomial from factors
::
  OVER >R >R			( C )
  HXS 2 10			( C Z1 )

LABEL LFCprodloop		( We use direct jumps to simplify ticR use )
  ticR NOTcase :: RDROP ;	( C P | C P F )
  RSWAP 'R RSWAP		( C P F k )
  DUP#0= IT
  CODE
	D1=D1+	10
	D=D+1	A
	D=D+1	A		( C P )
pfll1	A=PC			Loop until end of list
	LC(5)	(LFCprodloop)-(*)
	A=A+C	A
	D0=A
	GOVLNG	=Loop
  ENDCODE
  PPow# QMul			( C P' )
  CODE
	GOTO	pfll1		Loop until end of list
  ENDCODE
;

*---------------------------------------------------------------------

NULLNAME LFCInit	( b k --> C[k..n] )
::
  @lfcM @lfcn		( b k M n )
  2SWAP			( M n b k )

  NULLNAME CbInit		( M n b k --> C[k..n] )
  CODE
		GOSBVL	=POP2#
		R2=A			b
		R3=C			k
		GOSBVL	=POP#
		R4=A			n
		?A#0	A
		GOYES	+		Continue if n<>0
		GOVLNG	=Loop		Else return {} (M)

+		GOSBVL	=SAVPTR
		A=A+1	A		n+1
		C=R3			k
		A=A-C	A		n+1-k elements in output
		C=A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A		 5(n+1-k)
		C=C+C	A		10(n+1-k)
		C=C+CON	A,10		DOLIST & SEMI
		GOSBVL	=GETTEMP
		GOSBVL	=D1=DSKTOP
		C=DAT1	A		->M
		R0=C
		CD0EX
		DAT1=C	A		Overwrite ->M with ->C
		D1=C			->output
		LC(5)	=DOLIST
		DAT1=C	A
		D1=D1+	5
		GOSUB	LfcInit
		AD1EX
		GOSBVL	=D1=DSKTOP
		GOTO	LfcRecover
  ENDCODE
;

*---------------------------------------------------------------------

NULLNAME LFCNext			( C --> C' TRUE | FALSE )
::
  @lfcM @lfcn				( C M n )

  NULLNAME CbNext			( C M n --> C' TRUE | FALSE )
  CODE
		GOSBVL	=POP#		A[A] = n
		R4=A	A		n = n
		R3=A	A		k = n
		A=0	A
		R2=A	A		b = 0
		R1=A	A		a = 0
		GOTO	intoLFCnext
  ENDCODE
;

*---------------------------------------------------------------------

NULLNAME CbNextSub		( C M n a b k --> C' TRUE | FALSE )
CODE
		GOSBVL	=POP2#		A[A] = b	C[A] = k
		R3=C	A		k = k
		R2=A	A		b = b
		GOSBVL	=POP2#		A[A] = n	C[A] = a
		R1=C	A		a = a
		R4=A	A		n = n

intoLFCnext	GOSBVL	=PopASavptr
		C=DAT1	A
		R0=C	A		->C = ->C
		D1=A			->M = ->M

		A=R4	A		A[A] = n
		?A=0	A
		GOYES	lfcnextfail	Abort if n==0

* Now:
*	D1	->M
*	R0	->C
*	R1	a
*	R2	b
*	R3	k
*	R4	n


lfcnextloop	C=R0	A
		D0=C			->C
		C=R3	A		k
		GOSUB	NthComp#Drop	D0 = ->ck
		GOSUB	PopList#
		B=A	A		ck
		A=R1	A		a
		C=R2	A		b
		?A<=C	A		a < b ?
		GOYES	lfcnextwhile
		?B#0	A		ck<>0 ?
		GOYES	lfcnextok
lfcnextwhile	C=C+B	A		b = b + ck
		R2=C	A
		AD1EX
		D1=A
		D0=A			->M
		C=R3	A		k
		GOSUB	NthComp#Drop	->mk
		GOSUB	PopList#	mk
		C=R1	A
		C=C+A	A		a = a + mk
		R1=C	A
		C=R3	A
		C=C-1	A		k = k - 1
		R3=C	A
		?C#0	A		Repeat if k-1 >= 1
		GOYES	lfcnextloop

lfcnextfail	GOVLNG	=GPOverWrFLp	Return FALSE

****************************************
* Here:	R0   = ->C	R3 = k
*	R2   = b	R4 = n
*	B[A] = ck
****************************************

* Create list of degrees:
*	{ C[1..k-1] ck' Cini[k+1..n] }
*	     k-1     1    [n+1-k-1]	total = n

lfcnextok	B=B-1	A		ck'
		C=R2	A
		C=C+1	A		b = b + 1
		R2=C	A

		C=B	A
		R1=C	A		ck

		GOSBVL	=GETPTR		Protect M from GC destruction
		D1=D1-	5		now that we are going to use it
		D=D-1	A
		GOSBVL	=SAVPTR

		C=R4	A		n
		A=C	A
		C=C+C	A
		C=C+C	A
		C=C+A	A		5*n
		C=C+C	A		10*n
		C=C+CON	A,10		And list delimiters
		GOSBVL	=GETTEMP
		GOSBVL	=D1=DSKTOP
		C=DAT1	A		->M
		R0=C	A
		D1=D1+	5
		C=DAT1	A		->C
		CD0EX
		DAT1=C	A		Overwrite ->C by ->C'
		D1=C			->output
* Copy C[1..k-1] to output
		CD0EX
		B=C	A		->C		
		D0=C
		C=R3			k
		GOSUB	NthComp#Drop	Find end address of k-1'th #
		CD0EX			->Ck
		C=C-B	A		size to copy (includes DOLIST)
		CBEX	A
		D0=C
		C=B	A
		GOSBVL	=MOVEDOWN
		A=R1
		GOSUB	PushList#	Add ck'
		C=R3
		C=C+1	A		k = k+1
		R3=C
		GOSUB	LfcInit		Init remaining elements
		AD1EX
		GOSBVL	=D1=DSKTOP
		LC(5)	=TRUE
		DAT1=C	A
		D1=D1+	5
LfcRecover	D0=A			->output
		LC(5)	=SEMI
		DAT0=C	A
		D0=D0+	5
		C=DAT1	A		->C'
		A=A-C	A		list size - 5
		A=A+CON	A,5+6
		DAT0=A	A
		D0=D0+	5
		GOSBVL	=recover
		GOVLNG	=GETPTRLOOP	

****************************************
* Input:	R0[A] = ->M
*		R2[A] = b
*		R3[A] = k
*		R4[A] = n
*		D1    = ->output
* Exit:		D1    = ->outputtail
****************************************

LfcInit		A=R0
		D0=A			->M
		C=R3			k
		GOSUB	NthComp#Drop	->mk
		C=R4			n
		A=R3			k
		C=C-A	A		n-k = elements-1
		D=C	A		counter
		C=R2
		B=C	A		b
lfcinitloop	GOSUB	PopList#	A[A] = mi
		?A<=B	A
		GOYES	lfcinit10
		A=B	A		min(mi,b) = b
lfcinit10	B=B-A	A		b = b - min(mi,b)
		GOSUB	PushList#
		D=D-1	A		counter--
		GONC	lfcinitloop
		RTN
****************************************
* Find Nth system binary from list
****************************************
NthComp#Drop	D=C	A		N
		D0=D0+	5		->ob1
		D=D-1	A
		RTNC
		LC(5)	=DOBINT
nthcomp#drplp	D=D-1	A
		RTNC
		A=DAT0	A		prolog / ROM address
		D0=D0+	5
		?A#C	A
		GOYES	nthcomp#drplp	Skipped ROM address
		D0=D0+	5
		GONC	nthcomp#drplp	Skipped #
****************************************
* Pop # from list in D0.
****************************************
PopList#	A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		RTNYES			CS: End of list
		D0=D0+	5
		LC(4)	=DOBINT
		?A=C	A
		GOYES	poplist#
		AD0EX
		D0=D0+	5
		C=DAT0	A		integer
		D0=A
		A=C	A
		RTNCC			CC: Got integer
poplist#	A=DAT0	A		integer
		D0=D0+	5
		RTNCC			CC: Got integer
****************************************
* Push # into output in D1, as ROM pointer if possible
****************************************
PushList#	LC(5)	43		FORTYTHREE is the last one in sequence
		?A<=C	A
		GOYES	pushrom#
		LC(5)	=DOBINT
		DAT1=C	A
		D1=D1+	5
		GONC	pushrom10
pushrom#	LC(5)	=ZERO
		A=A+A	A		2*
		C=C+A	A		+2*
		A=A+A	A
		A=A+A	A		8*
		A=A+C	A		+10*
pushrom10	DAT1=A	A
		D1=D1+	5
		RTNCC
ENDCODE

*---------------------------------------------------------------------

NULLNAME LFCReset		( C --> C' TRUE / FALSE )
::
  @lfcM @lfcn			( C M n )
  @lfcb @lfcr			( C M n b r )
  COLA

  NULLNAME CbReset		( C M n b r --> C' TRUE / FALSE )
  ::
    3PICK #0=case
	:: 4DROP TRUE ; 	( return C TRUE )
    ONE				( C M n b r k )
    BEGIN
      DUP 5PICK #<		( C M n b r k f1 )
      7PICK 3PICK NTHCOMPDROP	( C M n b r k f1 ck )
      7PICK 4PICK NTHCOMPDROP	( C M n b r k f1 ck mk )
      2DUP#> NOT		( C M n b r k f1 ck mk f2 )
      4ROLL AND			( C M n b r k ck mk flag )
    WHILE
      4ROLL SWAP#-		( C M n b k ck r' )
      4ROLL ROT#-		( C M n k r' b' )
      2DUP#< case
       ::
         ROT			( C M n r' b' k )
 	 2RDROP CbNextSub	( return CbNext )
       ;
      SWAPROT #1+		( C M n b' r' k+1 )
    REPEAT			( C M n b r k ck mk )
    2DROP SWAPDROP		( C M n b k )
    5ROLL OVER			( M n b k C k )
    ONE SWAP#1- SUBCOMP		( M n b k C[1..k-1] )
    5UNROLL CbInit		( C[1..k-1] C'[k..n] )
    &COMP			( C'[1..n] )
    TRUE			( C' TRUE )
  ;
;

*---------------------------------------------------------------------

NULLNAME CbValid?               ( C M -> TRUE/FALSE )
::
  ONE                  
  BEGIN                     ( C M k )
    2DUP NTHELCOMP          ( C M k [mk TRUE/FALSE] )
  WHILE
    4PICK 3PICK             ( C M k mk C k )
    NTHCOMPDROP             ( C M k mk ck )
    #< NOTcase #1+          ( repeat with k+1 )
    3DROP FALSE             ( FALSE )
    2RDROP                  ( return )
  REPEAT
  3DROP TRUE                ( TRUE )
;

**********************************************************************
* Desc:		Extract all trivial power factors of P
* Stack:	( P --> P' Lf )
*		Lf = { {F1 #k1} .. {Fj #kj} }
**********************************************************************
NULLNAME PFactTriv
::
  PFactPowCnt				( P Lk flag )
  NOTcasedrop NULL{}			( --> P {} )
  SWAPOVER PDivLk 			( Lk P' )
  ZERO ROT				( P' #0 Lk )
  toLEN_DO (DO)				( P' f1 .. fj #j Lk )
    ::
       DUPINDEX@ NTHCOMPDROP		( P' f1 .. fj #j Lk #k )
       DUP#0=csDROP			( * continue if 0 * )
       INDEX@ ONE MkSplP		( P' f1 .. fj #j Lk #k Fi )
       SWAP TWO{}N			( P' f1 .. fj #j Lk fi )
       UNROT SWAP#1+SWAP		( P' f1 .. fj fi #j+1 Lk )
    ;
  LOOP 
  DROP {}N
;

**********************************************************************
* Desc:		Calculate trivial GCD of two polynomials
* Stack:	( P1 P2 --> P1' P2' gcd )
* Notes:	Integer content is included
**********************************************************************
* NULLNAME PGcdTriv
* ::
* 
* * Extract common trivial factors
* 
*   SWAP Variables		( P2 P1 #n1 )
*   ROT Variables			( P1 #n1 P2 #n2 )
*   ROT #MAX			( P1 P2 #n )
*   DUP4UNROLL			( #n P1 P2 #n )
*   #1+_ONE_DO (DO)
*       INDEX@ VarFactor		( a1 .. ai #n P1 P2 #m2 )
*       ROT INDEX@ VarFactor	( a1 .. ai #n P2 #m2 P1 #m1 )
*       ROT #MIN			( a1 .. ai #n P2 P1 #m )
*       4UNROLL			( a1 .. ai ai+1 #n P2 P1 )
*       SWAP			( a1 .. ai ai+1 #n P1 P2 )
*   LOOP
* 
* * Divide out trivial factors
* 
*   PFactPowCnt ITE PDivLk DROP		( a1 .. an #n P1 P2' )
*   SWAP
*   PFactPowCnt ITE PDivLk DROP		( a1 .. an #n P2' P1' )
* 
* * Extract integer contents
* 
*   PPZ ROT PPZ ROT ZGcd			( a1 .. an #n P1 P2 zgcd )
* 
* * Multiply together
* 
*   THREE psh				( P1 P2 zgcd #3 a1 .. an #n )
*   {}N PMulLk				( P1 P2 zgcd #3 pgcd )
*   SWAPDROP QMul				( P1 P2 gcd )
* ;

**********************************************************************
* Desc:		Calculate trivial power factors in P
* Stack:	( P --> P Lk flag )
*		Lk = { #k1 .. #kn }
* Notes:	flag is TRUE if any of the powers is nonzero
**********************************************************************
NULLNAME PFactPowCnt
CODE

sCNTZERO	EQU 9	* Flag to indicate a nonzero trivial factor

		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			D0 = ->P
		GOSUBL	QVARS
		R4=A	A		R4[A]=vars
		A=A+A	A		2*vars
		C=A	A
		C=C+C	A		4*vars
		C=C+C	A		8*vars
		C=C+A	A		10*vars
		C=C+CON	A,10		+ DOLIST + SEMI
		GOSBVL	=GETTEMP
		LC(5)	=DOLIST		Start list output
		DAT0=C	A
		AD0EX
		R2=A	A		R2[A] = ->list
		A=A+CON	A,5
		R3=A	A		R3[A] = ->output
		GOSBVL	=D1=DSKTOP
		ST=1	sCNTZERO	No nonzero power yet
		C=0	A
		C=C+1	A		var=1
--		RSTK=C
		A=DAT1	A
		D0=A			D0 = ->P
		A=C	A		A[A] = var
		GOSUBL	CountVarPow	A[A] = power
		?A=0	A
		GOYES	+
		ST=0	sCNTZERO	Flag nonzero power found
+		C=R3	A
		D0=C			D0 = ->output
		LC(5)	=DOBINT
		DAT0=C	A
		D0=D0+	5
		DAT0=A	A
		D0=D0+	5
		CD0EX
		R3=C	A		R3[A] = ->output'
		C=RSTK
		C=C+1	A		var++
		A=R4	A		A[A] = vars
		?C<=A	A
		GOYES	--		Continue until all vars counted for
		C=R3	A
		D0=C
		LC(5)	=SEMI		End list output
		DAT0=C	A		
		A=R2	A		A[A] = ->list
		GOSBVL	=GPPushA
		?ST=0	sCNTZERO	Push TRUE if nonzero power found
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE

* ::
*    FALSE ZERO 3PICK			( P flag #0 P )
*    Variables				( P F #0 P #v )
*    #1+_ONE_DO (DO)			( P a1 .. ai flag #i P )
* 	INDEX@ VarFactor		( P a1 .. ai flag #i P #n )
* 	4UNROLL SWAP#1+SWAP		( P a1 .. ai ai+1 flag #i+1 P )
* 	4PICK #0=?SKIP
* 	:: ROTDROP TRUE UNROT ;		( P a1 .. ai ai+1 flag' #i+1 P )
*    LOOP
*    DROPSWAP				( P a1 .. an #n flag )
*    OVER#2+UNROLL			( P flag a1 .. an #n )
*    {}N SWAP				( P Lk flag )
* ;

**********************************************************************
* Desc:		Divide polynomial by its trivial powers
* Stack:	( P Lk --> P' )
* Algorithm:	Copy elements of P to output
*		When SEMI is encountered skip back to remove Z0's
*		The number of Z0's == Lk[compositelevel]
**********************************************************************
NULLNAME PDivLk
::
  CODE
		GOSBVL	=SAVPTR

		D1=D1+	5		Allocate duplicate P
		A=DAT1	A
		D0=A			->P
		GOSBVL	=SKIPOB
		CD0EX
		A=DAT1	A
		C=C-A	A
		GOSBVL	=GETTEMP
		AD0EX			->output
		R0=A	A		->output

		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		R1=C	A		->Lk
		D1=D1+	5
		C=DAT1	A		->P
		DAT1=A	A		Overwrite ->P
		D0=C			->input
		D1=A			->output
		B=0	A		complevel

pdivlkloop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	pdivlklist
		LC(2)	=DOHSTR
		?A#C	A
		GOYES	pdivlksemi
		D0=D0+	5
		C=DAT0	A		Zlen
		D0=D0-	5
		C=C+CON	A,5
		GOSBVL	=MOVEDOWN	Copy Z
		GONC	pdivlkloop

pdivlklist	B=B+1	A		complevel++
pdivlkcont	DAT1=C	A		Copy DOLIST
		D0=D0+	5
		D1=D1+	5
		?B#0	A
		GOYES	pdivlkloop
		CD1EX
		D0=C			->Ptail
		A=R0	A		->P
		C=C-A	A		Psize
		C=C+CON	A,5+1		+link+GCnib
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=recover	Remove extra memory
		GOVLNG	=GETPTRLOOP

pdivlksemi	C=R1	A		->Lk
		CD0EX
		RSTK=C			->input saved
		C=B	A
		GOSUB	NthComp#Drop
		GOSUB	PopList#	A[A] = power of this variable
		C=A	A
		C=C+C	A		2*
		A=A+C	A		3*
		C=C+C	A		4*
		C=C+C	A		8*
		A=A+C	A		11*
		CD1EX
		C=C-A	A		Skip back over the Z0's
		CD1EX
		C=RSTK
		D0=C			->input restored
		LC(5)	=SEMI
		B=B-1	A		complevel--
		GONC	pdivlkcont
  ENDCODE
  DROP PTrim
;

**********************************************************************
* Desc:		Multiply trivial powers into polynomial form
* Stack:	( Lk --> P )
* Notes:	Ignores trailing zeros in Lk.
*		Returns Z1 if all are zeros.
**********************************************************************
* NULLNAME PMulLk
* CODE
* 		GOSBVL	=SAVPTR
* 
* 		A=DAT1	A		* Count number of variables to use
* 		D0=A			D0 = ->Lk
* 		D0=D0+	5
* 		B=0	A		total=0
* 		D=0	A		zeros=0
* -		GOSUB	PopList#	A[A]=Lki
* 		GOC	+		End of list
* 		B=B+1	A		total++
* 		D=D+1	A		zeros++
* 		?A=0	A
* 		GOYES	-
* 		D=0	A		zeros=0
* 		GONC	-
* +		C=B	A
* 		C=C-D	A		C[A] = variables (-trailing zeros)
* 
* 		?C#0	A
* 		GOYES	+		Nonzero - continue
* 		A=0	W
* 		A=A+1	A
* 		GOLONG	PushN1Loop	Return Z1
* 
* +		R4=C	A		R4[A] = variables
* 		A=DAT1	A		* Calculate needed room
* 		D0=A
* 		D0=D0+	5
* 		B=0	A		room=0
* 		D=C	A		D[A]=vars
* --		GOSUB	PopList#	A[A]=Lki
* 		B=B+A	A		room = room+Lki
* 		A=A+A	A
* 		B=B+A	A		room = room+3*Lki
* 		A=A+A	A
* 		A=A+A	A
* 		B=B+A	A		room = room+11*Lki (Z0 room)
* 		B=B+CON	A,10		room = room + DOLIST + SEMI
* 		D=D-1	A
* 		?D#0	A
* 		GOYES	--		Until all vars counted for
* 		B=B+CON	A,12		room = room + room(Z1)
* 		C=B	A
* 		GOSBVL	=GETTEMP	D0 = ->P	D1 = ->Ptail
* 
* 		CD1EX
* 		RSTK=C			RSTK = ->Ptail
* 		GOSBVL	=D1=DSKTOP
* 		A=DAT1	A		A[A] = ->Lk
* 		AD0EX
* 		R0=A	A		R0[A] = ->P
* 		D1=A			D1 = ->P
* 		D0=D0+	5		D0 = ->Lk1
* 		A=R4	A		A[A] = vars
* 		LC(5)	=DOLIST		* Output leading DOLIST's
* -		DAT1=C	A
* 		D1=D1+	5
* 		A=A-1	A
* 		?A#0	A
* 		GOYES	-		Until all vars counted for
* 
* 		LCZ1			* Output Z1
* 		DAT1=C	12
* 
* 		C=RSTK
* 		D1=C			D1 = ->Ptail
* 		C=R4	A
* 		D=C	A		D[A] = vars
* --		LC(5)	=SEMI		* Output SEMI for var
* 		D1=D1-	5
* 		DAT1=C	A
* 		GOSUB	PopList#	A[A] = power
* 		A=A-1	A
* 		GOC	+		No power for this var
* 		LCZ0			* Output power for var
* -		D1=D1-	11
* 		DAT1=C	11
* 		A=A-1	A
* 		GONC	-		Until full power output
* +		D=D-1	A
* 		?D#0	A
* 		GOYES	--		Until all vars counted for
* 
* 		GOVLNG	=GPOverWrR0Lp
* ENDCODE
 
 
**********************************************************************
* Desc:		Calculate what power of the given variable
*		is a factor in P.
* Stack:	( P #var --> P #n )
* Algorithm:
*		When compositelevel == #var
*			then count trailing zeros
*			take minimum with the old result
*		When composite level < #var
*			then any Z <> 0 means power = 0
**********************************************************************
NULLNAME VarFactor
CODE
		GOSBVL	=POP#		A[A] = #var
		GOSBVL	=SAVPTR
		C=DAT1	A
		D0=C			->P
		GOSUB	CountVarPow
		A=A+1	A		Return 0 instead of -1 for
		GOC	+		too high variable number
		A=A-1	A
+		GOVLNG	=PUSH#ALOOP

* Input:	A[A] = #var
*		D0   = ->P
* Output:	A[A] = #pow

CountVarPow	R0=A	A		R0[A] = #var
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+
		A=0	A		No polynomial - return 0
		RTN
+
		A=0	A		Init minimum to -1
		A=A-1	A
		R1=A	A		R1[A] = min
		B=0	A		complevel

cntpowlist	D0=D0+	5
		B=B+1	A		complevel++
		C=R0	A		#var
		?B=C	A
		GOYES	cntpownow	Count trailing zeros now
cntpowloop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	cntpowlist
		LC(2)	=DOHSTR
		?A#C	A
		GOYES	cntpowsemi
		D0=D0+	5
		C=DAT0	A		Zlen
		AD0EX
		A=A+C	A		Skip Z
		D0=A
		C=C-CON	A,7		CRY set if Z0
		GOC	cntpowloop
		A=0	A		#pow must be zero since Z <> 0
		RTN			(complevel known to be < #var)

cntpowsemi	D0=D0+	5
		B=B-1	A		complevel--
		?B#0	A
		GOYES	cntpowloop	Continue until traversed all
		A=R1	A		min
		A=A+1	A		CRY set if -1
		RTNC			If so then return 0
		A=A-1	A		Else return min
		RTN

cntpownow	D=0	A		zeros = 0
		C=A	W		copy high nibs for Z0 comparisons
		LCZ0
cntpowtrim	A=DAT0	11		First trim the sub poly
		D0=D0+	11		just in case P is untrimmed
		?A=C	W
		GOYES	cntpowtrim
		D0=D0-	11		->first non-zero coeff

cntpownowlp	LC(5)	=DOHSTR		Make sure low 5 nibbles are ok

cntpowzerlp	A=DAT0	11
		?A#C	W
		GOYES	+		Not Z0 - either SEMI or set zeros=0
		D0=D0+	11		Skip Z0
		D=D+1	A		zeros++
		GONC	cntpowzerlp	Count all consecutive zeros
+		LC(5)	=SEMI
		?A=C	A
		GOYES	+		Found SEMI - return zeros counter
		D=0	A		Found nonzero coeff - zeros = 0
		GOSBVL	=SKIPOB		Skip the coeff (possibly not Z!)
		GONC	cntpownowlp	And try zeros again

+		C=R1	A		Check if zeros count is bigger
		?C<=D	A		than current min
		GOYES	cntpowsemi	If so then just continue
		C=D	A
		R1=C	A		Else set new min
		?C#0	A		And continue if var factor is
		GOYES	cntpowsemi	still possible
		A=0	A		Else return 0 as the power
		RTN
ENDCODE
**********************************************************************

