	STITLE	ALG Berlekamp factorization
**********************************************************************
*	ALG library: Berlekamp factorization
**********************************************************************

**********************************************************************
* Desc:		Berlekamp's algorithm for finding modular factors
*		of a univariate polynomial with a leading frontend
*		for finding linear factors faster.
* Stack:	( P --> P FALSE / P Lf #prime TRUE )
* Notes:
*		The input polynomial must be square free, otherwise
*		the polynomial is not fully factored.
*
*		Due to memory restrictions byte sized coefficients
*		are used and the following restrictions were
*		imposed: prime<128 and degree<256. If the conditions
*		are not met FALSE is returned.
*
* Algorithm:
*
*		Allocate sufficient memory to hold
*
*		   Slot 1:  All factors found so far and the
*		            remaining polynomial being factored
*		   Slot 2:  Workspace for polynomial arithmetic
*		   Slot 3:  Workspace for matrix arithmetic
*
*		1. Find a prime which does not divide the leading
*		   coefficient of the polynomial such that
*		   the polynomial is square free mod prime.
*		2. Evaluate polynomial at all values smaller
*		   than the prime to find all linear factors.
*		3. If a degree=2 factor remains we're done
*		4. Form the Q-matrix for the remaining polynomial
*		   and solve it. [Knuth, Maple factor/Berlekamp etc]
*		5. Split the remaining polynomial using the
*		   found vectors.
*
*		Note that if this returns FALSE we may still
*		try to find linear factors with the common
*		modular method and then try again to see if
*		we can now work on the nonlinear ones. Seems
*		rather unlikely the extra work required is
*		needed though, the restrictions imposed are
*		not really that bad.
*
**********************************************************************
NULLNAME Berlekamp
CODE

* The first prime we will attempt is 5. Increasing this value
* has a strong influence on the total running time of the Berlekamp
* algorithm because
*
*	1. Forming the Q-matrix has loops from 1 to p
*	2. Splitting the Q-matrix has loops from 1 to p
*
* However since the actual lifting is done in system RPL it might
* make sense to increase this value to get faster increase in the
* modulo during the lifting. Possibly we could check the degree
* of the input polynomial, combine with some heuristics on the
* coefficients of the input polynomial, and determine the start
* value based on those. In particular a larger prime might make
* a lot more sense for low degree polynomials with relatively
* small coefficients.

BERPRIME1	EQU 11		* Start value if degree < 20
BERPRIME2	EQU 5		* Start value if degree >= 20
BERPRIMESWITCH	EQU 20		* The limiting degree

BERPRIMELIMIT	EQU 128		* Imposed by arithmetic on bytes
BERMAXDEGREE	EQU 256		* Imposed by arithmetic on bytes


* Check the degree of the polynomial

		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			D0 = ->P
		GOSUBL	QLEN		A[A] = degree+1
		A=A-1	A
		R3=A	A		R3[A] = degree

		LC(5)	BERMAXDEGREE
		?A<C	A
		GOYES	+		Degree is small enough

berlefail	GOVLNG	=GPPushFLoop	( --> P FALSE )

+

* Memory allocation logic:
*
*	Slot 1:	Has to be able to hold all the factors.
*		Maximum space is taken when all factors are
*		linear, in which case the room taken is
*		3*deg+1 bytes, the +1 being for the 00 terminator.
*	Slot 2:	Maximum space required during the modular
*		GCD calls. We need room for two input polynomials,
*		quotient, remainder. The room needed for changing
*		the order of the polynomials is the quotient slot
*		which is not needed.
*		4*(deg+1) bytes
*	Slot 3:	(deg+1)(deg+3) size byte matrix
*
*	Total:	2[ 3*deg+1 + 4*(deg+1) + (deg+1)(deg+3) ]
*		=  2(D^2+11D+8)
*		<= 2(D^2+11D+10)
*		=  2(D+1)(D+10)
*
*	When we exit we convert the found factors to a list of lists.
*	For this we use slots 2 and 3. The maximum memory needed is when
*	all the factors are linear, in which case we will need max
*	5+D*(5+3*13+5)+5 = 49*D+10 nibbles <= 25*D+5 bytes.
*	Slots 2-3 take (D+7)*(D+1) bytes which is actually smaller
*	for small degree, the difference being greatest at 75 bytes
*	for degrees 9 and 10. We can cover for this simply by adding
*	a safety value to the allocated size, say 200 bytes.


		A=A+1	A		A[A] = deg+1
		C=A	A
		C=C+CON	A,9		C[A] = deg+10
		GOSBVL	=MUL#		B[A] = (deg+1)(deg+10)
		B=B+B	A
		LC(5)	2*200		200 bytes safety
		C=C+B	A		C[A] = 2(deg+1)(deg+1)+safety
		GOSBVL	=MAKE$N

* Now choose a prime p and return the monic modular polynomial
* at the start of the allocated memory

		GOSUB	BerlePrime	work memory now contains monic image
		GOC	berlefail	Unless we failed to find a prime!

* First find linear factors by polynomial evaluation - this decreases
* the matrix size in the Q-matrix method and thus significantly
* increases the speed. BerlePrime leaves the address of the
* polynomial in R0[A]

		GOSUB	BerleLinear
		GOC	BerleExit	CS if was fully linearized

* Now the start of work memory contains a sequence of linear polynomials,
* possibly none of them though, followed by whatever polynomial remains,
* again possibly none, followed by a zero terminator.
* BerleLinear leaves the address of the remaining polynomial in R0[A]
* and returns CS if that polynomial is linear too

		GOSUB	InitQMatrix	Initialize Q-matrix
		GOSUB	SolveQMatrix	Solve Q-matrix	A[A]=r
		?A=0	A
		GOYES	BerleExit	Irreducible - we're done
		R1=A	A		R1[A] = r
		GOSUBL	TopTempob	A[A] = ->$
		AR0EX	A		R0[A] = ->$
		R2=A	A		R2[A] = ->table
		GOSUB	BerleSplit	Split polynomial to r+1 factors

* Done - return the found factors in a list and the prime itself

BerleExit	GOSUBL	TopTempob	A[A] = ->$
		R0=A	A		R0[A] = ->$

* Store the leading coefficient in the first factor

		GOSUB	BerleUnMonic

* Skip to the end of the factor list

		A=R0	A
		D0=A
		D0=D0+	10		D0 = ->factor1
-		C=0	A
		C=DAT0	B		C[B] = n(factor_i)
		?C=0	A
		GOYES	+
		C=C+1	A
		C=C+C	A
		AD0EX
		A=A+C	A
		AD0EX
		GONC	-
+		D0=D0+	2		Skip the 0-terminator too

* Convert to regular polynomials embedded in a list

		AD0EX
		R1=A	A		R1[A] = ->output
		D1=A
		A=R0	A
		D0=A
		D0=D0+	10		D0 = ->factor1
		LC(5)	=DOLIST		Embedd into a list
		DAT1=C	A
		D1=D1+	5
		A=DAT0	B		A[B] = n(factor_i)
--		B=A	B
		D0=D0+	2
		LC(5)	=DOLIST		Start polynomial
		DAT1=C	A
		D1=D1+	5
-		A=0	W
		A=DAT0	B		A[W] = coeff
		D0=D0+	2
		C=0	W
		C=R4.F	B		C[B] = prime
		C=C-A	A		C[B] = -coeff
		?C>A	A
		GOYES	+		Choose smaller
		A=C	A		-coeff is smaller
		C=C-1	S		set negative sign
+		GOSUBL	SHORTOUT
		B=B-1	B
		?B#0	B
		GOYES	-
		LC(5)	=SEMI		Terminate factor
		DAT1=C	A
		D1=D1+	5
		A=DAT0	B
		?A#0	B
		GOYES	--		Continue until terminator
		DAT1=C	A		Terminate list of factors
		D1=D1+	5

* Move the created list to start of string

		CD1EX			C[A] = ->tail
		A=R1	A		A[A] = ->start
		D0=A
		C=C-A	A		C[A] = size of list
		A=R0	A
		D1=A
		D1=D1+	10		D1 = ->$body
		GOSBVL	=MOVEDOWN

* Shrink work memory and exit

		AD1EX
		D0=A			D0 = ->tail (new)
		GOSBVL	=Shrink$
		A=R0	A
		A=A+CON	A,10
		GOSBVL	=GPPushA	Push the list
		GOSBVL	=SAVPTR
		A=R4	A
		R0=A	A		R0[A] = prime
		GOVLNG	=Push#TLoop	( --> P list #prime TRUE )

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Store the leading coefficient into the first
*		found factor to make the factorization valid.
* Entry:	R0[A] = ->$
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
BerleUnMonic	GOSBVL	=D0=DSKTOP
		A=DAT0	A
		D0=A
		D0=D0+	5		D0 = ->lcoeff
		C=0	W
		C=R4.F	B		C[W] = p
		GOSUBL	ZMODPSHORT	A[B] = lcoeff mod p		
		C=R0	A
		D0=C
		D0=D0+	12		->lcoeff(factor1)
		DAT0=A	B		Set it
		D0=D0-	2		->factor1
		GOLONG	bermonicutil	Utilize BERMONIC for the other coeffs

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Choose prime which does not divide lcoeff(poly)
*		and poly mod p is square free.
* Entry:	Stack level 1 = ->poly
*		R3[A] = degree
*		R0[A] = ->$work memory
* Exit:		R0[A] = ->polynomial (monic image of poly on stack)
*		CS: prime exceeds limit
*		CC: R4[W] = prime
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

* Currently: switch = 20, prime1 = 11, prime2 = 5

BerlePrime	C=0	W		Clear high nibbles to ease math
		A=R3	A		A[A] = degree(poly)
		LC(2)	BERPRIMESWITCH
		?A>=C	A
		GOYES	+
		LC(2)	BERPRIME1	Use prime 1 if degree(poly)<limit
		GONC	++
+		LC(2)	BERPRIME2	Use prime 2 if degree(poly)>=limit
++		R4=C	W		R4[W] = prime

berprimeloop	GOSBVL	=D1=DSKTOP
		A=DAT1	A
		D0=A			D0 = ->poly
		D0=D0+	5		D0 = ->lcoeff
		C=R4	W		C[W] = prime
		GOSUBL	ZMODPSHORT	A[B] = modulo
		?A#0	B
		GOYES	bertestprime	Nonzero - check poly is squarefree

berprimenext	A=R4	A		A[A] = prime
		GOSUBL	PRIME[A]+	A[A]=R4[A] = next prime
		LC(5)	BERPRIMELIMIT
		?A<C	A
		GOYES	berprimeloop	Continue if not maximum yet
		RTNSC			CS: Failure, prime too big

* Pop the polynomial mod p

bertestprime	A=DAT1	A
		D0=A			D0 = ->poly
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->$body
		GOSUBL	BERPOPPOLY	D1 = ->poly[p] tail

* Calculate the derivative mod p

		A=R0	A
		D0=A
		D0=D0+	10		D0 = ->poly[p]
		GOSUBL	BERDERMOD	D0 = ->der	D1 = ->dertail

* Calculate gcd in the workspace after the derivative

		C=R0	A
		CD0EX			C[A] = ->der
		D0=D0+	10		D0 = ->poly
		CD1EX			D1 = ->der	C[A] = ->dertail
		GOSUBL	BERGCDMOD	CS if GCD is constant (1)
		GONC	berprimenext	If not then try next prime

* Make the image monic

		A=R0	A
		A=A+CON	A,10		A[A] = ->poly
		R0=A
		D0=A
		GOLONG	BERMONIC	Returns CC if poly is not constant

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Find linear factors of a polynomial by evaluating
*		it in all points (Zp[X]).
* Entry:	R0[A] = ->poly
*		R4[B] = prime
* Exit:		R0[A] = ->polyn		The remaining polynomial
*					after linear factors are removed.
*					Preceded by the linear factors.
*					The entire sequence is terminated
*					by a zero byte.
*		CS if polyn is linear too.
* Uses:		A[W] B[W] C[W] D[W] R1[W]
* Stack:
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
BerleLinear	C=R4	B		C[B] = p
		C=C-1	B
berlinloop	R1=C	B		R1[B] = evalpoint (init: p-1)
		A=R0	A
		D1=A			D1 = ->poly
		A=DAT1	B		A[B] = n(poly)
		LC(2)	2
		?A<=C	B
		GOYES	berlinexit	poly is linear - done
		C=0	A		init value=0
		ASL	A
		ASL	A		A[B] = n(poly)
--		D1=D1+	2		*Horner loop*
		A=DAT1	B
		GOSUB	BERMODADD	Add coeff to result
		C=A	B		C[B] = result
		ASR	A
		ASR	A		A[B] = n
		A=A-1	B
		?A=0	B
		GOYES	++		Eval done - C[B] = result
		ASL	A
		ASL	A		A[B1] = n--
		A=R1.F	B		A[B] = evalpoint
		GOSUB	BERMODMULB	C[B] = result*evalpoint
		GOTO	--		Loop until all coefficients in

* If poly(evalpoint) mod p = 0 then found linear factor

++		?C=0	B
		GOYES	berlinwall	Found one - remove it from poly

* Next eval point

berlinnext	C=R1	B
		C=C-1	B		evalpoint--
		GONC	berlinloop
berlinexit	A=R0	A		Done evaluation - terminate sequence
		D0=A			D0 = ->poly
		C=0	A
		C=DAT0	B
		A=A+C	A
		A=A+C	A
		D0=A
		D0=D0+	2		->polytail
		A=0	A
		DAT0=A	B
		C=C-CON	B,3		CS if degree(lastpoly)<3
		RTN			Return flag to caller

* Found linear factor - divide poly by it and continue

berlinwall	A=R0	A
		D0=A			D0 = ->poly
		C=0	A
		C=DAT0	B		C[B] = n(poly)
		C=C+1	A
		C=C+C	A
		A=A+C	A		A[A] = ->polytail
		D0=A
		A=A+CON	A,6		linear poly takes 6 nibbles
		D1=A
		B=A	A		B[A] = ->polytail (new)
		GOSBVL	=MOVEUP		Move poly up to make room for linfact

		LCHEX	0102		lcoeff(linfact)=1, n(linfact)=2
		DAT0=C	A
		D0=D0+	4
		A=R4	B		A[B]=p
		C=R1	B		C[B]=evalpoint
		?C=0	B
		GOYES	+
		C=A-C	B		C[B]=-evalpoint
+		DAT0=C	B
		D0=D0-	4		D0 = ->linfact
		C=B	A
		D1=C			D1 = ->polytail (new)
		GOSUBL	BERDIVMOD
		A=R0	A
		A=A+CON	A,6
		R0=A	A		R0[A] = ->poly (the remaining one)
		D1=A
		D0=A
		D0=D0+	2		Skip the zero remainder
		C=0	A
		C=DAT0	B		C[B] = n(quotient)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN	Move down to fill the 0-remainder
		GOTO	berlinnext	And try next evalpoint

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Split polynomial into factors using the vectors
*		in Q-Matrix.
* Entry:	R0[A]	->$		starts with factor list
*		R1[A]	r		number of splits to do
*		R2[A]	->table		Q-matrix
*		R3[A]	d		Q-matrix has d+1 columns
*		R4[A]	p		prime
* Exit:		R4[A]	p		prime
*		Found factors appended to the factor list
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
BerleSplit

* Find the last polynomial in the list (the one being split)

		A=R0	A
		D0=A
		D0=D0+	10		D0 = ->factor1
		A=R1	A
		GOSUB	Aslw5		A[A1] = r
-		C=0	A
		C=DAT0	B		C[B] = n(factor_i)
		?C=0	A
		GOYES	+
		C=C+1	A
		C=C+C	A
		AD0EX
		D0=A
		A=A+C	A		A[A] = ->factor_i+1
		AD0EX			A[A] = ->factor_i
		GONC	-
+		R1=A	A		R1[A]=->poly	R1[A1]=r

* Setup work address beyond the list

		D0=D0+	2		Skip terminator
		AD0EX
		R0=A	A		R0[A] = ->work memory

* Skip to 1st row in Q-matrix

		A=R2	A		->table
		C=R3	A		d
		C=C+1	A
		C=C+C	A
		A=A+C	A
		A=A+C	A
		A=A+C	A
		R2=A	A		R2[A] = ->vector1

* Reverse the elements in the rows of the Q-matrix and establish
* the length fields of the polynomials using the extra 2 bytes
* between each row. We only have to reverse the first r rows,
* not the full matrix.

		C=R1	W
		GOSUB	Csrw5
		D=C	A		D[A] = r
		C=R2	A
		D0=C			D0 = ->vector1
--		AD0EX
		B=A	A		B[A] = ->row[i]
		D0=A
		C=R3	A		C[A] = d
		A=A+C	A
		A=A+C	A
		D1=A			D1 = ->rowtail
		C=C+1	B
-		C=C-1	B		Count leading zeros to get the
		D1=D1-	2		degree right
		A=DAT1	B
		?A=0	B
		GOYES	-
		D1=D1+	2
		DAT1=C	B		Store n(poly) after first nonzero
-		A=DAT0	B		And reverse from there to the
		C=DAT1	B		start of the row
		DAT1=A	B
		DAT0=C	B
		D0=D0+	2
		D1=D1-	2
		AD0EX
		D0=A
		CD1EX
		D1=C
		?A<C	A		Reverse until addresses cross
		GOYES	-
		A=B	A		A[A] = ->row[i]
		C=R3	A		C[A] = d
		C=C+1	A
		A=A+C	A
		A=A+C	A
		D0=A			D0 = ->row[i+1]
		D=D-1	A		D[A] = r--
		?D#0	A
		GOYES	--		Reverse r rows

* Establish the polynomial pointers for splitting

		A=R1	A
		GOSUB	Aslw5
		A=R1.F	A
		R1=A	W		R1[A]=->poly[i]	R1[A1]=->poly1

* The splitting loop itself tries to split the current polynomial
* with the current row, if it fails then it tries the next polynomial
* until terminator is found. Then a constant is added to the row
* and the loop restarts, unless the additional constant goes to zero
* when all possibilities have been tested. In that case the row
* pointer is advanced to the next row and the loop is started again.
* The process is guaranteed to end during the first r rows.

bersplitloop


* Choose polynomial to split - quadratic ones can be discarded

		A=R1	A		A[A] = ->poly
--		D0=A
		C=0	A
		C=DAT0	B		C[B] = n(poly)
		?C=0	A
		GOYES	bersplitnew	Exhausted - update Q-matrix row
		C=C-CON	B,3		CS if quadratic
		GONC	bertrysplit
		C=C+CON	A,4
		C=C+C	A
		AD0EX
		A=A+C	A
		R1=A.F	A		R1[A] = ->poly++
		GONC	--

* Polynomial list exhausted. Setup new pass and with a new row

bersplitnew	A=R1	W
		GOSUB	Asrw5
		R1=A.F	A		R1[A] = ->poly == ->poly1

		A=R2	A		A[A] = ->vector
		D1=A
		C=0	A
		C=DAT1	B		C[B] = n(vector)
		C=C+C	A
		A=A+C	A
		D1=A			D1 = ->tcoeff(vector)
		A=DAT1	B
		A=A+1	B		tcoeff++
		DAT1=A	B
		C=R4	B		C[B] = p
		?A<C	B
		GOYES	bersplitloop	Accept vector if tcoeff<p

		A=R2	A		Else we take the next row
		C=R3	A
		C=C+1	A
		C=C+C	A
		A=A+C	A
		R2=A.F	A		R2[A] = ->row++
		GOTO	bersplitloop

* Try gcd(poly,vector) to see if we get a split

bertrysplit	C=R2	A		D0 = ->poly
		D1=C			D1 = ->vector
		C=R0	A		C[A] = ->workmem
		C=C+CON	A,4		2 extra bytes for movement aid
		GOSUBL	BERGCDMOD	CS if constant
		GONC	++

* No luck - try next polynomial

bersplitnext	A=R1	A		A[A] = ->poly
		D0=A
		C=0	A
		C=DAT0	B		C[B] = n(poly)
		?C=0	A
		GOYES	bersplitnew	Terminator found - next vector
		C=C+1	A
		C=C+C	A
		A=A+C	A
		R1=A.F	A		R1[A] = ->poly++
		GOTO	bersplitloop

* Found gcd - check it isn't a trivial one

++		C=R1	A
		D1=C			D1 = ->npoly
		C=0	A
		C=DAT0	B		C[B] = n(gcd)
		A=DAT1	B		A[B] = n(poly)
		?C>=A	B
		GOYES	bersplitnext	Trivial gcd - failed after all

* Nontrivial gcd - split polynomial to two parts and append them to the list

		C=C+1	A
		C=C+C	A
		AD0EX
		B=A	A		B[A] = ->gcd
		A=A+C	A
		D1=A			D1 = ->workmem after gcd

* Copy poly to memory after gcd

		A=R1	A
		D0=A			D0 = ->poly
		C=0	A
		C=DAT0	B
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN
		
* And divide

		A=B	A
		D0=A			D0 = ->gcd
		GOSUBL	BERDIVMOD

* Move the polynomials after the one being split up 4 nibbles to
* make room for the 2 new polynomials. The number 4 comes from
* having an extra term and an extra size field.

		D0=D0-	4		D0 = ->tail = ->terminator +2
		AD0EX
		B=A	A		B[A] = ->tail
		D0=A
		C=R1	A		C[A] = ->poly
		C=A-C	A		C[A] = size to move
		A=A+CON	A,4		A[A] = ->tail (new)
		R0=A.F	A		A[A] = ->workmem (new)
		D1=A
		GOSBVL	=MOVEUP

* Move the gcd down to overwrite the old ->poly address

		A=B	A
		AD0EX
		D0=D0+	4		D0 = ->gcd
		D1=A			D1 = ->poly (old)
		C=0	A
		C=DAT0	B		C[B] = n(gcd)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN

* Move the quotient down to overwrite the remaining space

		D0=D0+	2		Skip the 0-remainder from division
		C=0	A
		C=DAT0	B		C[B] = n(quotient)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN

* Update variables

		A=R1	W
		GOSUB	Asrw5
		C=A	A		C[A] = ->poly1
		GOSUB	Asrw5		A[A] = r
		A=A-1	A
		?A=0	A
		RTNYES			DONE r SPLITS!
		GOSUB	Aslw5		A[A1] = r
		A=C	A
		GOSUB	Aslw5		A[A2] = r	A[A1] = ->poly1
		AD1EX
		R1=A	W		R1[A] = ->poly'
		GOTO	bersplitloop	Continue until splitted all

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Calculate Q-matrix for nonlinear polynomial
* Entry:	D0	->poly
*		R4[A]	prime
* Exit:		D0	->matrix
*		R3[A]	degree
* Uses:		A[W] B[W] C[W] D[W] R0[W] R1[W] R2[W] R3[W]
* Stack:
* Notes:	The matrix has d+1 columns, d+2 rows
*
*		Row 1:	u	d+1 elements	polynomial itself
*		Row 2:	a	d   elements
*		Row 3:	c	d   elements	flag vector
*		Row 4:	q0	d   elements	row 1 of Q-matrix
*		...
*		Row N:	qd	d   elements	row d-1 of Q-matrix
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
InitQMatrix	

* Allocate space for the matrix from the end of the work string

		A=R0	A
		D0=A			D0 = ->poly
		A=0	A
		A=DAT0	B		A[B] = n(poly)
		A=A-1	B		A[B] = degree
		R3=A	A		R3[A] = degree

		A=A+1	A		d+1 columns
		C=A	A
		C=C+1	A		d+2 rows
		GOSBVL	=MUL#
		B=B+B	A		B[A] = bytes(matrix)

		GOSUBL	TopTempob	A[A] = ->$
		A=A+CON	A,5
		D1=A
		C=DAT1	A
		A=A+C	A		A[A] = ->$tail
		A=A-B	A
		D1=A			D1 = ->table
		AR0EX	A		R0[A] = ->table
		D0=A			D0 = ->poly

* Now copy the polynomial to the first row in reverse order

		C=0	A
		C=DAT0	B		C[B] = n(poly)
		AD0EX
		A=A+C	A
		A=A+C	A
		AD0EX			D0 = ->tcoeff(poly)
-		A=DAT0	B
		DAT1=A	B
		D0=D0-	2
		D1=D1+	2
		C=C-1	A
		?C#0	A
		GOYES	-		Copy all coefficients

* Initialize the second row and third rows to zero, except a[0]=1

		A=R0	A		A[A] = ->table
		C=R3	A		C[A] = d
		C=C+1	A
		C=C+C	A
		A=A+C	A		A[A] = ->row2 (a)
		D1=A
		A=0	A
		A=A+1	A
		DAT1=A	B		a[0]=1
		D1=D1+	2
		C=C+C	A		C[A] = bytes(2 rows)
*		C=C-CON	A,2		minus a[0] itself
		GOSBVL	=WIPEOUT

* Form the Q-matrix itself

* Variables to be used:
*		R0[A]	->table
*		R1[W]	miscellaneous
*		R2[W]	k,j,i loop counters
*		R3[A]	d
*		R4[B]	p


*	for k from 1 to d1 do
*	    for j from 1 to p do
*	        t = a[d-1]
*		for i from d-1 to 1 step -1 do
*		    a[i] = mods(a[i-1]-t*u[i],p)
*	        next
*	        a[0] = mods(-t*u[0],p)
*	    next
*	    for i from 0 to d-1 do
*	        q[k,i] = a[i]
*	    next
*	    q[k,k] = mods(q[k,k]-1,p)
*	next

		C=0	A
		C=C+1	A		k=1 to d-1
qnextk		GOSUB	Cslw5
		C=C+1	A		j=1 to p
qnextj		GOSUB	Cslw5				
		C=R3.F	A
		C=C-1	A		i=d-1 to 1
		R2=C	W		R2[W] = loop counters
		A=R0.F	A
		A=A-CON	A,4		A[A] = ->table - 4
		C=R3.F	A		C[A] = d
		C=C+1	A
		C=C+C	A
		A=A+C	A
		D0=A			D0 = ->u[d-1]
		A=A+C	A
		D1=A			D1 = ->a[d-1]
		C=DAT1	B
		R1=C.F	B		R1[B] = t = a[d-1]

qnexti		A=DAT0	B		A[B] = u[i]
		D0=D0-	2
		C=R1.F	B		C[B] = t
		GOSUB	BERMODMUL	A[B] = t*u[i]
		D1=D1-	2
		C=DAT1	B		C[B] = a[i-1]
		D1=D1+	2
		C=C-A	B
		GONC	+
		C=C+B	B		C[B] = a[i-1]-t*u[i]
+		DAT1=C	B		= a[i]
		D1=D1-	2
		C=R2.F	A
		C=C-1	A		i=i-1
		R2=C.F	A		
		?C#0	A
		GOYES	qnexti		NEXT I

		A=DAT0	B		A[B] = u[0]
		C=R1.F	B		C[B] = t
		GOSUB	BERMODMULB	A[B] = t*u[0]
		C=B	B
		C=C-A	B		C[B] = -t*u[0]
		?C#B	B
		GOYES	+
		C=0	B
+		DAT1=C	B		= a[0]

		C=R2	W
		GOSUB	Csrw5
		C=C+1	A		j=j+1
		A=R4.F	A		A[A] = p
		?C>A	A
		GOYES	+
		GOTO	qnextj		NEXT J

+		GOSUB	Csrw5
		R2=C	W		R2[A] = k

*	    for i from 0 to d-1 do
*	        q[k,i] = a[i]
*	    next

		A=C	A		A[A]=k
		C=0	A
		GOSUB	QMatrixPosQ
		AD1EX			D1 = ->q[k,0]
		D0=A			D0 = ->a[0]
		C=R3.F	A		C[A] = d
		C=C+C	A
		GOSBVL	=MOVEDOWN

*	    q[k,k] = mods(q[k,k]-1,p)

		A=R2.F	A		A[A] = k
		C=A	A		C[A] = k
		A=A+1	A
		A=A+1	A
		GOSUB	QMatrixGet	A[B] = q[k,k]
		A=A-1	B		A[B] = q[k,k]-1
		GONC	+
		C=R4.F	B		C[B] = p
		A=A+C	B
+		DAT0=A	B		= q[k,k]

		C=R2.F	A
		C=C+1	A		k=k+1
		A=R3.F	A		A[A] = d-1
		?C>=A	A
		RTNYES
		GOTO	qnextk		NEXT K

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Solve nullspace from Q-matrix
* Entry:	R0[A]	->table
*		R3[A]	degree
*		R4[A]	prime
* Exit:		A[A]	r (number of factors)
*		R0[A]	->table
*		R3[A]	degree
*		R4[W]	prime
* Uses:		A[W] B[W] C[W] D[W] R0[W] R1[W] R2[W] R3[W] R4[W]
* Stack:	
* Algorithm:
*	for k from 1 to d-1 do
*	    for j from 0 to d-1 while q[k,j]=0 or c[j]<>0 do
*	    next
*	    if j < d then
*	        mi = mods(-1/q[k,j],p)
*	        q[k,j] = 1
*	        if mi <> 1 then
*	            for i from k+1 to d-1 do
*	                q[i,j] = mods(mi*q[i,j]
*	            next
*	        endif
*	        for jj from 0 to d-1 do
*	             if jj <> j and q[k,jj] <> 0 then
*	                 mi = q[k,jj]
*	                 q[k,jj] = 0
*	                 for i from k+1 to d-1 do
*	                     q[i,jj] = mods(q[i,jj]0mi*q[i,j],p)
*	                  next
*	             endif
*	        next
*	        c[j] = k
*	    else
*	        r = r+1
*	        for i from 0 to d-1 do
*	            q[r,i] = 0
*	        next
*	        for i from 0 to d-1 do
*	            if c[i]<>0 then q[r,c[i]] = q[k,i] endif
*	        next
*	        q[r,k] = 1
*	    endif
*	next
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

SolveQMatrix	C=0	W
		R1=C	W		R1[A1] = r = 0
		C=C+1	A		k=1
xnextk		A=C	A
		GOSUB	Cslw5
		R2=C	W		R2[A1]=k
		C=0	A
		GOSUB	QMatrixPosQ
		D0=A			D0 = ->q[k,0]
		A=0	A
		C=0	A
		GOSUB	QMatrixPosQ
		D1=A			D1 = ->c[0]
		B=0	A		j=0
		C=R3.F	A		C[A] = d
-		A=DAT1	B		A[B] = c[j]
		?A#0	B
		GOYES	+
		A=DAT0	B		A[B] = q[k,j]
		?A#0	B
		GOYES	++
+		D0=D0+	2
		D1=D1+	2
		B=B+1	A		j++
		C=C-1	A		d--
		?C#0	A
		GOYES	-
++		A=B	A		A[A] = j
		GOSUB	Aslw5
		A=R0.F	A
		R0=A	W		R0[A1] = j

		C=R3.F	A		C[A] = d
		?B<C	A
		GOYES	solveqif
		GOTO	solveqelse

solveqif	A=R2	W
		GOSUB	Asrw5		A[A]=k
		DAT1=A	B		c[j]=k

		A=DAT0	B		A[B] = q[k,j]
		GOSUB	BERMODINV	A[B] = 1/q[k,j]
		C=B	B
		C=C-A	B		C[B] = -1/q[k,j]
		?C#B	B
		GOYES	+
		C=0	B
+		R1=C.F	B		R1[B] = mi

		A=B	B		A[B] = p
		A=A-1	B
		DAT0=A	B		q[k,j] = -1
		
		C=C-1	B
		?C=0	B
		GOYES	++		mi==1 - skip next to next phase

		C=R2	W
		GOSUB	Csrw5
		C=C+1	A		i=k+1
		GONC	+
-		R2=C.F	A
		AD0EX
		C=R3.F	A		C[A] = d
		C=C+1	A
		C=C+C	A
		A=A+C	A
		D0=A			D0 = ->q[i,j]
		A=DAT0	B		A[B] = q[i,j]
		C=R1.F	B		C[B] = mi
		GOSUB	BERMODMULB
		DAT0=A	B		q[i,j] = mods(mi*q[i,j])
		C=R2.F	A
		C=C+1	A		i=i+1
+		A=R3.F	A		d
		?C<A	A
		GOYES	-

++		C=0	A		jj=0
xnextjj		R2=C.F	A
		A=R0	W
		GOSUB	Asrw5		A[A] = j
		?A=C	A
		GOYES	xjjnxt
		A=R2	W
		GOSUB	Asrw5		A[A] = k
		A=A+1	A
		A=A+1	A
		GOSUB	QMatrixGet	A[B] = q[k,jj]
		?A#0	B
		GOYES	+
xjjnxt		C=R2.F	A
		C=C+1	A		jj=jj+1
		A=R3.F	A		A[A] = d
		?C<A	A
		GOYES	xnextjj
		GOTO	xknext
+		A=DAT0	B
		R1=A.F	B		R1[B] = mi = q[k,jj]
		A=0	A
		DAT0=A	B		q[k,jj]=0
		A=R0	W
		GOSUB	Asrw5		A[A] = j
		C=R2.F	A		C[A] = jj
		A=A-C	A		A[A] = j-jj
		A=A+A	A
		CD0EX
		D0=C
		C=C+A	A
		D1=C			D1 = ->q[k,j]
		A=R2	W
		GOSUB	Asrw5		A[A] = k
		C=R2	W
		GOSUB	Cslw5
		C=A	A
		C=C+1	A		i=k+1
		GONC	+
-		R2=C	W
		C=R3.F	A		C[A] = d
		C=C+1	A
		C=C+C	A
		AD0EX
		A=A+C	A		->q[i,jj]
		AD0EX
		AD1EX
		A=A+C	A		->q[i,j]
		AD1EX
		A=DAT1	B		A[B] = q[i,j]
		C=R1.F	B		C[B] = mi
		GOSUB	BERMODMUL	A[B] = mi*q[i,j]
		C=DAT0	B		C[B] = q[i,jj]
		GOSUB	BERMODADDB	A[B] = q[ijj]+mi*q[i,j]
		DAT0=A	B
		C=R2	W
		C=C+1	A		i=i+1
+		A=R3.F	A		A[A] = d
		?C<A	A
		GOYES	-
		GOSUB	Csrw5
		R2=C	W
		GOTO	xjjnxt


solveqelse	C=R1	W
		GOSUB	Csrw5
		C=C+1	A		r=r+1
		A=C	A
		GOSUB	Cslw5
		C=R1.F	A
		R1=C	W

		C=0	A
		GOSUB	QMatrixPosQ
		D1=A			D1 = ->q[r,i]
		C=R3.F	A		C[A] = d
		C=C+C	A
		GOSBVL	=WIPEOUT	q[r,...]=0

		A=0	A
		R2=A.F	A		i=0
		C=0	A
		GOSUB	QMatrixPosQ
		D1=A			D1 = ->c[0]
		A=R2	W
		GOSUB	Asrw5
		C=0	A
		GOSUB	QMatrixPosQ
		D0=A			D0 = ->q[k,0]
--		C=0	A
		C=DAT1	B		C[B] = c[i]
		?C=0	B
		GOYES	+
		A=R1	W
		GOSUB	Asrw5		A[A] = r
		GOSUB	QMatrixPosQ	A[A] = ->q[r,c[i]]
		C=DAT0	B		C[B] = q[k,i]
		AD0EX
		DAT0=C	B		q[r,c[i]]=q[k,i]
		D0=A
+		D0=D0+	2		D0 = ->q[k,i+1]
		D1=D1+	2		D1 = ->c[i+1]
		C=R2.F	A
		C=C+1	A		i=i+1
		R2=C.F	A
		A=R3.F	A		A[A]=d
		?C<A	A
		GOYES	--

		A=R1	W
		GOSUB	Asrw5		A[A] = r
		C=R2	W
		GOSUB	Csrw5		C[A] = k
		GOSUB	QMatrixPosQ
		D0=A			D0 = ->q[r,k]
		A=0	A
		A=A+1	A
		DAT0=A	B		q[r,k]=1


xknext		C=R2	W
		GOSUB	Csrw5
		C=C+1	A		k=k+1
		A=R3.F	A		A[A]=d
		?C>=A	A
		GOYES	+
		GOTO	xnextk

* Done - return r in A[A]

+		A=R1	W
		GOTO	Asrw5		A[A] = r

*---------------------------------------------------------------------
* Entry:	R0[A]	->table		A[A]=Y	C[A]=X
* Exit:		D0	->table[Y,X]
*		A[B]	element
*---------------------------------------------------------------------
QMatrixGet	GOSUB	QMatrixPos
		D0=A			->table[Y,X]
		A=DAT0	B		element
		RTN
*---------------------------------------------------------------------
* Desc:		Calculate element position in Q-matrix
* Entry:	R0[A]	->table		A[A]=Y	C[A]=X	R3[A]=d
* Exit:		A[A]	->table[Y,X]
* Uses:		A[A] B[A] C[A]
* Notes:	Position = ->table + 2*(Y*(d+1)+X)
*---------------------------------------------------------------------
QMatrixPosQ	A=A+1	A		Special entry to get elements
		A=A+1	A		in the q vectors

QMatrixPos	B=C	A
		B=B+B	A		B[A] = 2*X
		C=R0.F	A
		B=B+C	A		B[A] = ->table + 2*X
		C=R3.F	A		C[A] = d
		C=C+1	A
		C=C+C	A		C[A] = 2(d+1)
		SB=0
-		ASRB.F	A
		?SB=0
		GOYES	+
		SB=0
		B=B+C	A
+		C=C+C	A
		?A#0	A
		GOYES	-
		A=B	A		A[A] = ->element position
		RTN

*---------------------------------------------------------------------
*		JUMP TABLE
*---------------------------------------------------------------------

Aslw5		GOVLNG	=ASLW5
Asrw5		GOVLNG	=ASRW5
Cslw5		GOVLNG	=CSLW5
Csrw5		GOVLNG	=CSRW5
		
*---------------------------------------------------------------------
*		BYTE ARITHMETIC UTILITIES
*---------------------------------------------------------------------

*---------------------------------------------------------------------
* Desc:		Modular addition
* Entry:	A[B]=X	C[B]=Y	R4[B]=N
* Exit:		A[B]=X+Y MOD N
*---------------------------------------------------------------------
BERMODADD	B=C	B		B[B]=Y
		C=R4.F	B		C[B]=N
		BCEX	B		C[B]=Y	B[B]=N
BERMODADDB	A=A+C	B
		?A<B	B
		RTNYES
		A=A-B	B
		RTN
*---------------------------------------------------------------------
* Desc:		Modular substraction
* Entry:	A[B]=X	C[B]=Y	R4[B]=N
* Exit:		A[B]=X-Y MOD N
*---------------------------------------------------------------------
BERMODSUB	B=C	B		B[B]=Y
		C=R4.F	B		C[B]=N
		BCEX	B		C[B]=Y	B[B]=N
BERMODSUBB	A=A-C	B
		RTNNC
		A=A+B	B
		RTN
*---------------------------------------------------------------------
* Desc:		Modular multiplication
* Entry:	A[B]=X C[B]=Y R4[B]=N
* Exit:		A[B]=C[B]=X*Y MOD N
* Uses:		A[B] B[B] C[B] D[B]
* Notes:	Y is made modular for the benefit of BDERPOLY
*---------------------------------------------------------------------
BERMODMUL	B=C	B		B[B]=Y
		C=R4.F	B		C[B]=N
		BCEX	B		C[B]=Y	B[B]=N
-		?C<B	B		
		GOYES	BERMODMULB
		C=C-B	B
		GONC	-
BERMODMULB	?C<=A	B		Exchange smaller of X,Y to C[B]
		GOYES	+
		ACEX	B
+		D=C	B		D[B]=Y
		C=0	B		C[B]=R=0
		SB=0
bermodmullp	DSRB.F	B		Y>>1
		?SB=0
		GOYES	+
		SB=0
		C=C+A	B		R=R+X
		?C<B	B
		GOYES	+
		C=C-B	B
+		A=A+A	B		X<<1
		?A<B	B
		GOYES	+
		A=A-B	B
+		?D#0	B
		GOYES	bermodmullp
		A=C	B		A[B]=C[B]=R
		RTN

*---------------------------------------------------------------------
* Desc:		Modular inverse
* Entry:	A[B]=X	R4[B]=N
* Exit:		A[B]=1/X MOD N
* Uses:		A[W] B[B] C[W] D[B]
* Stack:	1
* Notes:	We are very very short of registers during the
*		Berlekamp algorithm so that we have hardly any
*		room in registers. Thus we store all of them in C[W].
*
*		Variables in C[W] are	..ZZPPTTRRSSTT
*		Initialized values are	..zzpp..1000..
*		Nibble index		01234567890123
*---------------------------------------------------------------------

* Exit located here to avoid a too long branch

bermodinvok	P=C	8
		C=P	0
		P=C	9
		C=P	1
		P=	0
		A=C	B		A[B]=R
		RTN

* The actual entries

BERMODINV	C=R4.F	B
		B=C	B		B[B] = p
BERMODINVB	C=0	W		S=0
		C=C+1	B
		?A<=C	B
		RTNYES			1/0 = 0 MOD N	1/1 = 1 MOD N
		C=C+1	XS		R=1
		CSL	A
		CSL	A
		C=B	B		P=p
		CSL	W
		CSL	W
		C=A	B		Z=z
		CSL	W
		CSL	W		.ZPTRST
bermodinvlp	P=C	2
		C=P	0
		P=C	3
		C=P	1
		A=C	B		A[B]=Z
		P=C	4
		C=P	0
		P=C	5
		C=P	1		C[B]=P
		?C=0	B
		GOYES	bermodinvok	Done if P=0
		GOSUB	BYTEDIV		C[B]=Q=Z/P
		A=C	B		A[B]=Q
		P=C	0
		C=P	12
		P=C	1
		C=P	13		T=Q
		P=C	10
		C=P	0
		P=C	11
		C=P	1		C[B]=S
		GOSUB	BERMODMULB	A[B]=SQ mod p
		P=C	8
		C=P	0
		P=C	9
		C=P	1		C[B]=R
		C=C-A	B
		GONC	+
		C=C+B	B		C[B]=R-SQ mod p
+		CPEX	0
		CPEX	12
		CPEX	0
		CPEX	1
		CPEX	13		T=R-SQ
		CPEX	1		C[B]=Q
		A=C	B
		P=C	4
		C=P	0
		P=C	5
		C=P	1		C[B]=P
		GOSUB	BERMODMULB	A[B]=QP mod p
		P=C	2
		C=P	0
		P=C	3
		C=P	1		C[B]=Z
		C=C-A	B
		GONC	+
		C=C+B	B		C[B]=Z-QP mod p
+		P=C	0
		C=P	6
		P=C	1
		C=P	7		T=Z-QP
		CSR	W
		CSR	W		New pass
		GOTO	bermodinvlp

*---------------------------------------------------------------------
* Desc:		Byte division
* Entry:	A[B]=X C[B]=Y
* Exit:		A[B]=rem(X,Y)
*		C[B]=quo(X,Y)
*		D[B]=Y
* Uses:		A[B] C[B] D[B]
* Notes:	Binary algorithm not used due to register space shortage
*---------------------------------------------------------------------
BYTEDIV		D=C	B		D[B]=Y
		C=0	B		C[B]=quo=0
		?D=0	B
		RTNYES			Safety against X/0
		P=	1
		?D#0	P
		GOYES	+
		DSL	B		Y<<4
		C=C+1	B
+		P=	0
		CPEX	0		P=loops
		CDEX	B
		GOTO	bytedivlp
bytedivlp2	CSR	B
bytedivlp	D=D+1	P		quodigit++
		A=A-C	B
		GONC	bytedivlp
		D=D-1	P		quodigit--
		A=A+C	B		adjust remainder
		P=P-1
		GONC	bytedivlp2
		P=	0
		CDEX	B		C[B]=quo	D[B]=Y
		RTN

*---------------------------------------------------------------------
*		BYTE POLYNOMIAL UTILITIES
*---------------------------------------------------------------------

*---------------------------------------------------------------------
* Desc:		Pop polynomial from D0 to D1 modulo p
* Entry:	D0	->poly
*		D1	->output
*		R4[W]	prime
* Exit:		D0	->polytail
*		D1	->output tail
* Uses:		A[A] B[A] C[W] D[A]
* Stack:	1 (QLEN)
*---------------------------------------------------------------------
BERPOPPOLY	GOSUBL	QLEN		A[A] = deg+1
		DAT1=A	B		Set size field
		D1=D1+	2
		D0=D0+	5		D0 = ->lcoeff

-		A=DAT0	A		A[A] = prolog
		LC(5)	=DOHSTR
		?A#C	A
		GOYES	+		Abort if not Z coefficient
		C=R4	W
		GOSUBL	ZMODPSHORT	A[B] = coeff mod p
		DAT1=A	B
		D1=D1+	2
		GONC	-
+		D0=D0+	5		D0 = ->poly tail
		RTNCC

*---------------------------------------------------------------------
* Desc:		Trim polynomial
* Entry:	D0	->poly
* Exit:		D0	->polytail (old)
*		D1	->polytail (new)
*		D[S]	<>0 if poly is zero
* Uses:		A[B] C[B] D[S] D0 D1
* Stack:	0
*---------------------------------------------------------------------
BERTRIM		CD0EX
		D1=C			D1 = ->poly
		CD0EX
		A=DAT0	B		A[B] = n
-		A=A-1	B
		GOC	+
		D0=D0+	2
		C=DAT0	B
		?C=0	B
		GOYES	-		Loop until nonzero coeff
+		A=A+1	B		A[B] = true degree
		D=0	S		Flag not zero polynomial
		?A#0	B
		GOYES	+
		D=D+1	S		Flag zero polynomial
+		DAT1=A	B		Store the size
		D1=D1+	2
-		A=A-1	B		coeffs--
		RTNC
		C=DAT0	B		Copy coefficient
		DAT1=C	B
		D0=D0+	2
		D1=D1+	2
		GONC	-
*---------------------------------------------------------------------
* Desc:		Make polynomial monic by dividing with lcoeff
* Entry:	D0	->poly
*		R4[B]	prime
* Exit:		D0	->poly (monic)
*		CS if polynomial is constant
*---------------------------------------------------------------------
BERMONIC	D0=D0+	2
		A=DAT0	B		A[B] = lcoeff
		LC(2)	1
		DAT0=C	B		Set lcoeff=1
		D0=D0-	2
		C=DAT0	B
		C=C-CON	B,2		CS if poly is constant
		RTNC			And return if so
		GOSUB	BERMODINV	A[B] = lc^-1	B[B] = prime

* Help entry for multiplication with constant at exit

bermonicutil	C=0	A
		C=DAT0	B		C[A] = n
		AD0EX
		A=A+C	A
		A=A+C	A
		AD0EX			D0 = ->tcoeff
--		C=C-1	B		C[B] = n--
		?C=0	B
		GOYES	+		Loop until ->lcoeff
		CSL	A
		CSL	A		C[B1] = n
		C=A	B		C[B] = lc^-1
		ASL	A
		ASL	A		A[B] = lc^-1
		A=DAT0	B		A[B] = coeff
		GOSUB	BERMODMUL	A[B] = coeff'
		DAT0=A	B
		ASR	A
		ASR	A		A[B] = lc^-1
		CSR	A
		CSR	A		C[B] = n
		D0=D0-	2
		GONC	--
+		D0=D0-	2		D0 = ->poly
		RTNCC			CC: not constant
*---------------------------------------------------------------------
* Desc:		Calculate derivative of polynomial modulo p
* Entry:	D0	->poly
*		D1	->output
*		R4[B]	modulo
* Exit:		D0	->der
*		D1	->dertail
* Uses:		A[?] B[?] C[?] D[?] D0 D1
* Stack:	2
*---------------------------------------------------------------------
BERDERMOD	CD1EX
		RSTK=C			RSTK = ->pder[p]
		CD1EX
		A=DAT0	B		A[B] = n
		A=A-1	B		A[B] = n-1
		GOC	berderzero	der(0) = 0
		?A=0	B
		GOYES	berderzero	der(const) = 0
		B=A	B		B[B] = n-1
		DAT1=A	B		Set degree of derivative
-		D0=D0+	2
		D1=D1+	2
		A=DAT0	B		A[B] = coefficient
		C=B	B		C[B] = pow
		BSL	A
		BSL	A		B[B1] = pow
		GOSUB	BERMODMUL	A[B] = coeff*pow mod p
		DAT1=A	B
		BSR	A
		BSR	A
		B=B-1	B		B[B] = pow--
		?B#0	B
		GOYES	-		Loop until tcoeff(poly)
		C=RSTK
		RSTK=C
		D0=C			D0 = ->derivative
		GOSUB	BERTRIM		Trim the derivative
-		C=RSTK
		D0=C			D0 = ->derivative
		RTN
berderzero	A=0	A		Set zero derivative
		DAT1=A	B
		D1=D1+	2		D1 = ->zerotail
		GONC	-

*---------------------------------------------------------------------
* Desc:		Modular polynomial GCD with input copy to work area
* Entry:	D0	->A
*		D1	->B
*		C[A]	->work memory
*		R4[B]	prime
* Exit:		D0	->gcd (monic)
*		CS if gcd is constant
* Notes:	Does not handle zero input
*---------------------------------------------------------------------
BERGCDMOD	D=C	A		D[A] = ->workmem
		CD1EX
		B=C	A		B[A] = ->B
		C=0	A
		C=DAT0	B		C[B] = n(A)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN	A copied
		C=B	A
		D0=C			D0 = ->B
		C=0	A
		C=DAT0	B		C[B] = n(B)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN	B copied
		C=D	A
		D0=C			D0 = ->A (the new copy)
*---------------------------------------------------------------------
* Desc:		Modular polynomial GCD
* Entry:	D0	->B
*			->A follows
*			->work memory follows
*		R4[B]	prime
* Exit:		D0	->gcd
*		CS if gcd is constant
* Notes:	Does not handle zero input
*---------------------------------------------------------------------
BERGCDMODNOW	C=0	A
		C=DAT0	B		C[B] = n(B)
		C=C+1	A
		C=C+C	A
		AD0EX
		D0=A
		A=A+C	A
		D1=A			D1 = ->A
		C=0	A
		C=DAT1	B		C[B] = n(A)
		C=C+1	A
		C=C+C	A
		A=A+C	A
		D1=A			D1 = ->workmem
--		GOSUB	BERDIVMOD	D1 = ->quotient
		GOC	++		CS: Remainder is zero

* Move B to old location of A and the remainder to follow it

		AD0EX
		B=A	A		B[A] = ->B
		AD0EX
		C=0	A
		C=DAT0	B		C[B] = n(B)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN	Overwrite quotient
		CD1EX			C[A] = ->tail
		AD0EX			A[A] = ->start (->A)
		D0=A
		C=C-A	A		C[A] = nibbles in B & remainder
		A=B	A
		D1=A			D1 = ->B (old)
		GOSBVL	=MOVEDOWN	Now we have ->rem and ->B order
		A=B	A
		D0=A			D0 = ->rem
		GONC	--		Loop until remainder is zero

* Remainder is zero - the divisor is the gcd

++		GOTO	BERMONIC	CS if gcd is constant

*---------------------------------------------------------------------
* Desc:		Modular polynomial division with remainder
* Entry:	D0	->divisor	(B)
*			->dividend	(A)
*		D1	->output
*		R4[B]	prime
* Exit:		D0	->divisor	(B)
*			->remainder	(R)
*		D1	->quotient	(Q)
*		CS if remainder R=0
* Uses:		A[W] B[W] C[W] D[W]
* Stack:	2 (BERMODINV)
* Notes:	Handles zero polynomials right
*---------------------------------------------------------------------
BERDIVMOD	A=0	A
		DAT1=A	A		Init zero quotient
		A=DAT0	B		A[B] = n(B)

		?A=0	A		If n(B)==0
		GOYES	berdivcc	then we're done (R=A <> 0 assumed)

		CD0EX
		D0=C
		C=C+A	A
		C=C+A	A		C[A] = ->A-2
		CD1EX
		RSTK=C			RSTK = ->output
		D1=D1+	2		D1 = ->A
		C=DAT1	B		C[B] = n(A)
		?C>=A	B
		GOYES	++		Divide only if n(A) >= n(B)

		?C#0	B
		GOYES	+		If n(A)==0
		C=RSTK			then R=0 and Q=0
		A=0	A
		DAT1=A	A
		D1=D1+	2		D1 = ->Q
		RTNSC			CS: remainder is zero
+		C=RSTK			D0 = ->divisor (remainder=A)
		D1=C			D1 = ->quotient = 0
berdivcc	RTNCC			CC: Remainder not zero

* Division must be done

++		C=RSTK
		GOSUB	Cslw5		C[A1] = ->output
		D=C	W		D[A1] = ->output
		D0=D0+	2
		A=DAT0	B		A[B] = lcoeff(B)
		D0=D0-	2
		GOSUB	BERMODINV	A[B] = 1/lc mod p
		C=A	B
		D=C	B		D[B] = 1/lc mod p

* First calculate the new quotient coefficient

berdivloop	D1=D1+	2
		A=DAT1	B		A[B] = lcoeff(R)
		D1=D1-	2
		C=D	B		C[B] = 1/lc mod p
		DSL	A
		DSL	A		D[B1] = 1/lc mod p
		GOSUB	BERMODMUL	C[B] = lcoeff(R)/lc mod p
		D=C	B		D[B] = lcoeff(R)/lc mod p
		C=D	W
		GOSUB	Csrw5		C[A] = ->quotient
		CD1EX
		RSTK=C			RSTK = ->R
		C=0	A
		C=DAT1	B		C[B] = n(quo)
		C=C+1	B
		DAT1=C	B		n(quo)++
		AD1EX
		A=A+C	A
		A=A+C	A
		AD1EX
		DAT1=A	B		Store the new quotient coefficient
		C=RSTK
		D1=C			D1 = ->R

* New remainder R = R - lcoeff(R)/lc * B mod p

		CD0EX
		D0=C
		GOSUB	Cslw5		C[A2] = ->B
		CD1EX
		D1=C
		GOSUB	Cslw5		C[A1] = ->R

		C=DAT0	B		C[B] = n(B)
		A=DAT1	B		A[B] = n(R)
		A=A-1	B
		DAT1=A	B		n(R)--
		A=A+1	B
		A=A-C	B		A[B] = ctr2 = n(R)-n(B)
		ASL	A
		ASL	A		A[B1] = ctr2
		A=C	B		A[B] = ctr1 = n(B)
		D0=D0+	4
--		D1=D1+	4
		A=A-1	B		A[B] = ctr1--
		?A=0	B
		GOYES	++		Done part 1
		ASL	A
		ASL	W		Save ctr2,ctr1
		C=D	B		C[B] = lcoeff(R)/lc mod p
		DSL	A
		DSL	W		Save ->quo,1/lc,lcoeff(R)/lc
		A=DAT0	B		A[B] = coeff(B)
		GOSUB	BERMODMUL	A[B] = lcoeff(R)/lc * coeff(B)
		C=DAT1	B		C[B] = coeff(R)
		C=C-A	B
		GONC	+
		C=C+B	B		C[B] = coeff(R) -lc(R)/lc * coeff(B)
+		D1=D1-	2
		DAT1=C	B		Store to prev position in R
		ASR	W
		ASR	A		Restore A[W]
		DSR	W
		DSR	A		Restore D[W]
		D0=D0+	2
		GONC	--		Loop until part 1 done

* Now B is exhausted - just copy the remaining part of the remainder

++		DSR	A		Restore D[A]
		DSR	A
		ASR	A
		ASR	A		A[B] = ctr2
-		A=A-1	B		ctr2--
		GOC	+
		C=DAT1	B
		D1=D1-	2		n(R) has decreased by 1
		DAT1=C	B
		D1=D1+	4
		GONC	-
+		GOSUB	Csrw5
		D1=C			D1 = ->R
		GOSUB	Csrw5
		D0=C			D0 = ->B
		A=DAT1	B		A[B] = n(R)
		C=DAT0	B		C[B] = n(B)
		?A<C	B
		GOYES	+		Done if n(R)<n(B)
		GOTO	berdivloop	Else do another pass

* Division done. Trim remainder and move quotient down to follow it.

+		CD0EX
		B=C	A		B[A] = ->B
		CD1EX
		D0=C			D0 = ->R
		GOSUB	BERTRIM		D0 = ->Rtail(old) D1=->Rtail D[S]
		
		C=D	W
		GOSUB	Csrw5
		D0=C			D0 = ->quo
		CD1EX
		D=C	A		D[A] = ->quo (new)
		D1=C
		C=0	A
		C=DAT0	B		C[B] = n(quo)
		C=C+1	A
		C=C+C	A
		GOSBVL	=MOVEDOWN	Move down to tail of remainder
		C=D	A
		D1=C			D1 = ->quo (new)
		A=B	A
		D0=A			D0 = ->B (divisor)
		?D#0	S
		RTNYES			CS: remainder zero
		RTNCC			CC: remainder nonzero

*---------------------------------------------------------------------
*		MISCELLANEOUS UTILITIES
*---------------------------------------------------------------------

*---------------------------------------------------------------------
* Desc:		Find next prime > input integer
* Entry:	A[A] = integer
* Exit:		A[A] = R4[A] next prime
*		If overflow then A[A]=0
* Uses:		A[A] B[A] C[A] D[A] R4[A]
* Stack:	2 (IntDiv + RSTK save)
*---------------------------------------------------------------------
PRIME[A]+	C=0	A		Check small primes first
		LC(1)	2
		?A>C	A
		GOYES	+		prime>2 - use division algorithm
		ACEX	A
		?A>C	A
		RTNYES			return 2 for 0,1 input
		A=A+1	A		return 3 for 2 input
		RTN
+		?ABIT=1	0		If input is even make it odd
		GOYES	+		by substracting one
		A=A-1	A
+		R4=A.F	A		Save the number

prime[a]lp	A=R4.F	A		A[A] = trial prime
		A=A+1	A
		RTNC			Return A[A]=0 for overflow
		A=A+1	A
		R4=A.F	A		R4[A] = next trial prime
		C=0	A
		LC(1)	3		Trial division starting from 3
-		RSTK=C
		A=R4.F	A		A[A] = trial prime
		GOSBVL	=IntDiv		A[A] = rem	C[A] = quo
		B=C	A		B[A] = quo
		C=RSTK			C[A] = divisor
		?A=0	A
		GOYES	prime[a]lp	If rem=0 then not prime, try next
		A=R4.F	A		A[A] = trial prime
		?B<C	A
		RTNYES			If quo<divisor then is prime
		C=C+1	A
		C=C+1	A		divisor += 2
		GONC	-		Continue trial division

*---------------------------------------------------------------------
* Desc:		Find topmost object in tempob area
* Entry:
* Exit:		A[A] = ->ob
* Uses:		A[A] C[A] D0
*---------------------------------------------------------------------
TopTempob	D0=(5)	=aTEMPTOP
		A=DAT0	A
		D0=A
		A=DAT0	A
		D0=A			D0 = ->temptop
		D0=D0-	5
		C=DAT0	A		C[A] = toplink
		A=A-C	A
		A=A+1	A		A[A] = ->topob
		RTN
**********************************************************************
ENDCODE
