	STITLE	ALG Meta Matrix
**********************************************************************
*		ALG Library: Meta Matrix Subroutines
**********************************************************************

**********************************************************************
* DEFINITION
*
* Meta matrix is an exploded matrix with rational elements.
* The form is used to avoid excessive overhead due to list operations
* which depending on the elements starts to kick in at about dim 5x5.
*
* The format of the meta matrix is:
*
*	nom11			element 1
*	den11
*	nom12			element 2
*	den12
*	..
*	nom1X
*	den1X
*	nom21
*	..
*	nomYX			last element
*	denYX
*	#Y			number of rows
*	#X			number of columns
*	NULL$			MM identifier
*
* Use of the NULL$ identifier enables using special MM locator subroutines
* which determine the location of the MM on stack at all times. In particular
* most MM low level subroutines do not assume anything on the number of
* objects at the bottom of stack before the MM is found, thus enabling
* easy mathematical calculations with the elements of the matrix. The speed
* difference can be even 30s for 10x10 matrices.
*
* ( MM ? )	is used in stack diagrams to indicate the property
*		of unknown stack position
*
**********************************************************************

**********************************************************************
* EXPORT:
*
*	MMsolve		( MM --> MM' )
*	MMlurank	( MM --> MM' #rank )
*	MMdet		( MM --> nom den )
**********************************************************************

**********************************************************************
* Name:		MMsolve
* Desc:		Solve metamatrix using MS if possible
* Stack:	( MM --> MM' )
**********************************************************************
NULLNAME MMsolve
::
  MM2MS				( MM MS TRUE / MM FALSE )
  NOTcase MMsolve2		( --> MM' )
  MScrout			( MM MS' #rank TRUE / MM MS' FALSE )
  NOTcasedrop MMsolve2		( --> MM' )
  DROP MMdrop			( MS' )
  DUP TOTEMPOB			( MS' MS' )
  MSsolvetri			( MS' MS'' flag )
  NOTcasedrop
  :: MS2MM MMsolvetri ;		( --> MM' )
  SWAPDROP MS2MM		( --> MM' )
;

* This is the plain MM code
NULLNAME MMsolve2
:: MMcrout DROP MMsolvetri ;

**********************************************************************
* Name:		MMlurank
* Desc:		LU decomposition of a meta matrix and rank computation
* Stack:	( MM --> MM' #rank )
**********************************************************************
NULLNAME MMlurank
::
  ERRSET 
   ::
     MM2MS		   ( MM MS TRUE / MM FALSE )
     NOTcase MMcrout	   ( --> MM' #rank )
     MScrout		   ( MM MS' #rank TRUE / MM MS' FALSE / MM #rk + err )
     NOTcasedrop MMcrout   ( --> MM' #rank )
     DROP MMdrop MS2MM	   ( MM' )
     3PICK		   ( MM' #rank )
   ;
  ERRTRAP
   :: ERROR@ #305 #<>case ERRJMP ;		( Inf Res Error )
;

**********************************************************************
* Name:		MMdet
* Desc:		Meta matrix determinant
* Stack:	( MM --> nom den )
**********************************************************************
NULLNAME MMdet
::
  ( Handle special cases for speed )

  OVER #1=case
  :: 3DROP RCanno ;			( --> nom den )

  OVER #2= case
  :: 3DROP				( n11 d11 n12 d12 n21 d21 n22 d22 )
     8ROLL 8ROLL RMul			( n12 d12 n21 d21 n1 d1 )
     6UNROLL 6UNROLL RMul		( n1 d1 n2 d2 )
     RSub RCanno			( nom den )
  ;

  MMlurank				( MM' #rank )
  4PICK #<> case  			( rank<Y --> det=0 ) 
    :: MMdrop HXS 1 0 HXS 2 10 ;	( --> Z0 Z1 )

  ( Multiply the diagonal elements to get determinant )

  HXS 2 10 DUP				( MM' n d )
  5PICK ZERO_DO (DO)			( For 0 to Y-1 )
    INDEX@ DUP MMget RMul		( MM n' d' )
  LOOP
  MMdrop				( n' d' )
  RCanno				( nom den )
;

**********************************************************************
* Name:		MMcrout
* Desc:		LU decomposition of a meta matrix
* Stack:	( MM --> MC #rank )
*
* M: n x m, m >= n
*
*    M = [A|B]	A: n x n   B: n x (m-n) 
*    
* C: n x m
*    
*    |U11 U12 U13 .. U1n Y11 .. Y1(m-n)|
*    |L21 U22 U23 .. U2n Y21 .. Y2(m-n)|
*    |L31 L32 U33 .. U3n Y31 .. Y3(m-n)|
*    | .   .     .    .   .       .    |
*    | .   .       .  .   .       .    |
*    |Ln1 Ln2 Ln3 .. Unn Yn1 .. Yn(m-n)|
*
*
* L: (Lij) n x n        lower triangular matrix with 1's on diagonal
* U: (Uij) n x n        upper triangular matrix 
* Y: (Yij) n x (m-n)  
*
* s.t.  L * U = A * P   P = row permutation matrix
*	L * Y = B * P
*
* Algorithm:
*
*    C:=M
*    for i := 1 to n-1 do
*      [Choose pivot row among Ci..Cn]
*      k:=i
*      for j:=i+1 to n do
*        if Cji "better pivot" than Cki then k:=j 
*      Swap rows Ci<->Ck
*
*      [Gaussian elimination]
*      for k:=i+1 to n do
*	 Cki := Cki/Cii
*	 for j:=i+1 to n do
*	   Ckj := Ckj - Cki*Cij
*    end
*
* Note:
*
*  if the matrix M does NOT have full rank then the algorithm cleans
*  the stack, leaves the current C and rank (current iteration)
*  on the stack and raise an "Infinite Result" error.
*
**********************************************************************
NULLNAME MMcrout
::
  3PICK 				( MM Y )

  DUP#1= case				( * If dim = 1           * )
  ::					( * check single element * )
    6PICK QIsZero? NOT?SEMI		( * non-null -> MM 1     * )
    #1- ErrInfRes			( * null -> MM 0 + error * )
  ;

* Loop for Y-1 rows

  #1- ZERO_DO (DO)			( MM )		( [0,Y-2] )

     ( Get the would-be pivot from indexed row        )
     ( = diagonal element determined by current index )

     INDEX@ DUP#1+			( MM y y+1 )
     SWAPDUP MMget			( MM y nom den )
     
     ( See if there are any better ones in the elements     )
     ( in the same column and below it. The loop stop value )
     ( Y-1 guarantees there is atleast one row below        )

     6PICK INDEX@ #1+			( MM y nom den Y y+1 )
     DO					( MM y nom den )	( [y+1,Y-1] )
      ::
         INDEX@ JINDEX@ MMget		( MM y nom den nom2 den2 )
         BetterPivot? NOTcase2DROP	( * Discard if worse pivot * )
         2SWAP 2DROP			( MM y nom2 den2 )
         ROTDROP			( MM nom2 den2 )
         INDEX@ #1+ UNROT		( MM y2 nom2 den2 )
      ;
     LOOP
					( MM y nom den )

     ( Now update the MM according to the chosen pivot row )

     ROT #1- INDEX@ MMrowswap		( MM nom den )

     ( Check the chosen pivot )

     OVER QIsZero? case2drop
     ::
       MMtrim INDEX@ ErrInfRes		( --> MM #rank )
     ;

     ( Update the rows below the pivot row )

     5PICK INDEX@ #1+			( MM nom den Y y+1 )
     DO								( [y+1,Y-1] )
        INDEX@ JINDEX@ MMget		( MM nom den n' d' )
        3PICK 5PICK RMul RCanno		( MM nom den n d )
        2DUP INDEX@ JINDEX@ MMput	( MM nom den n d )

        JINDEX@				( MM nom den n d y )
        7PICK OVER #1+			( MM nom den n d y X y+1 )
        DO							( [y+1 X-1] )
           DUPINDEX@ MMget		( MM nom den n d y a1 a2 )
           5PICK 5PICK RMul		( MM nom den n d y b1 b2 )
           JINDEX@ INDEX@ MMget
           2SWAP RSub			( MM nom den n d y c1 c2 )
           ( RCanno )
           JINDEX@ INDEX@ MMput		( MM nom den n d y )
        LOOP
        3DROP				( MM nom den )
     LOOP
     2DROP				( MM )

  LOOP

  MMtrim				( MM )

* Check the right-down corner element

  3PICK DUP #1-DUP MMget		( MM Y nom den )
  DROP QIsZero? NOT?SEMI		( --> MM Y=rank )
  #1- ErrInfRes				( --> MM #rank )
;

**********************************************************************
* Name:		BetterPivot?	
* Desc:		True if rational2 is a "better" pivot than rational1
* Stack:	( n1 d1 n2 d2 -> n1 d1 n2 d2 flag )
* Note:		The desirable pivot is the 'simplest' non-zero
*		rational. Maybe should add a test for 1/1 ?
**********************************************************************
NULLNAME BetterPivot?	( n1 d1 n2 d2 -> n1 d1 n2 d2 flag )
::
  OVER QIsZero? caseFALSE
  4PICK QIsZero? caseTRUE
  CK&DISPATCH0
    #BB00 ( hxs hxs any any ) FALSE
      #BB ( any any hxs hxs ) TRUE
    #B000 ( hxs any any any ) FALSE
      #B0 ( any any hxs any ) TRUE
     #B00 ( any hxs any any ) FALSE
       #B ( any any any hxs ) TRUE
    ZERO  ( any any any any ) FALSE
;

**********************************************************************
* Name:		MMsolvetri
* Desc:		Solve triangular system
* Stack:	( MC --> MX )
* Notes:	An error is generated if there is no "right-hand side"
*
* MC = result from Crout decomposition
* MX = the solved right hand side of the matrix
*
* C: n x (n+m), m > 0, C=[U|Y], U=(n x n)  upper triangular, Y=(n x m)
*    
*    |U11 U12 U13 .. U1n Y11 .. Y1m|
*    |    U22 U23 .. U2n Y21 .. Y2m|
*    | .      U33 .. U3n Y31 .. Y3m|
*    | .   .     .    .   .       .|
*    |  any        .  .   .       .|
*    | .   .      .. Unn Yn1 .. Ynm|
*
* 
* X: n x m  s.t.  U*X=Y
*
*
* Algorithm:
*
*   for i:=n downto 1 do
*     for j:=1 to m do
*	Xij := Yij
*	for k := i+1 to n do
*	  Xij := Xij-Uik*Xkj
*	Xij := Xij/Uii
*
**********************************************************************
NULLNAME MMsolvetri
::
  OVER 4PICK			( MM #X #Y )
  2DUP#> NcaseSIZEERR		( * Error if X<=Y * )
  #-				( MM #N )

  ( Loop for each row starting from the bottom one )

  4PICK ZERO_DO (DO)		( MM #N )			( [0,Y-1] )

  ( Loop for each element on the right-hand side )

    DUP ZERO_DO (DO)		( MM #N )			( [0,N-1] )

      JSTOP@ JINDEX@ #- #1-
      5PICK INDEX@ #+		( MM #N #y #x ; for right-hand side element )
      MMget			( MM #N nom den )

      ( Adjust for each element in the upper triangle )
      ( There are none for bottom row, 1 for next etc )

      JINDEX@ DUP#0=ITE DROP
      ::
         #1+_ONE_DO (DO)					( [1,y] )
           6PICK ISTOP@ #-
           7PICK ISTOP-INDEX #-	( MM #N nom den #y #x ; for row element )
           MMget		( MM #N nom den nl dl )
           8PICK ISTOP-INDEX #-
           NINE PICK JINDEX@ #+	( MM #N nom den nl dl #y #x )
           MMget		( MM #N nom den nl dl nr dr )
           RMul RSub		( MM #N nom' den' )
         LOOP
      ;

      ( And divide by the pivot element on the diagonal )

      JSTOP@ JINDEX@ #- #1-DUP	( MM #N nom den #y #x ; for diagonal )
      MMget			( MM #N nom den nd dd )
      SWAP RMul RCanno		( MM #N nom' den' )

      ( Put re-calculated element back to the right-hand side )

      JSTOP@ JINDEX@ #- #1-
      7PICK INDEX@ #+		( MM #N nom' den' #y #x )
      MMput			( MM' #N )
    LOOP
  LOOP
  DROP				( MM' )

  ( Take the right-hand side submatrix )

  ZERO 4PICK			( MM' 0 Y )
  MINUSONE DUP			( MM' 0 Y MAX MAX )
  MMsub				( MM'' )
;

**********************************************************************
* Name:		MA2MM
* Desc:		Convert symbolic matrix to meta matrix
* Stack:	( MA Lvars --> MM )
**********************************************************************
NULLNAME MA2MM
::
  OVER LENCOMP			( MA vars #Y )
  3PICK CARCOMP LENCOMP		( MA vars #Y #X )
  4ROLL FIVE			( .. vars #Y #X MA #off )
  BEGIN
     NEXTCOMPOB			( .. vars #Y #X MA #off' [ob T / F] )
  WHILE
     FIVE			( .. vars #Y #X MA #off' ROW #off )
     BEGIN
        NEXTCOMPOB		( .. vars #Y #X MA #off' ROW #off' [ob T / F )
     WHILE
        8PICK A2R		( .. vars #Y #X MA #off' ROW #off' nom den )
	NINE UNROLL
        NINE UNROLL		( .. nom den vars #Y #X MA #off' ROW #off' )
     REPEAT
     DROP			( .. vars #Y #X MA #off' )
  REPEAT
  DROP				( .. vars #Y #X )
  ROTDROP			( .. #Y #X )
  NULL$				( .. #Y #X NULL$ == MM )
;

**********************************************************************
* Name:		MM2MA
* Desc:		Convert meta matrix to symbolic matrix
* Stack:	( MM ? Lvars --> ? MA )
**********************************************************************
NULLNAME MM2MA
::
  1LAMBIND						( Bind Lvars )
  MMdims SWAP			( MM ? #X #Y )
  ZERO_DO (DO)
     DUP
     ZERO_DO (DO)		( MM ? .. #X )
        JINDEX@ INDEX@ MMget	( MM ? .. #X nom den )
        1GETLAM R2A		( MM ? .. #X symb )
        SWAP			( MM ? .. symb #X )
     LOOP
     {}N DUPLENCOMP		( MM ? .. {} #X )
  LOOP
  DROP				( MM ? row1 .. rowY )
  MMdims DROP			( MM ? row1 .. rowY #Y )
  {}N				( MM ? MA )
  MMdrop			( ? MA )
  ABND							( Abandon Lvars )
;

**********************************************************************
* Name:		MMtrim
* Desc:		Simplify meta matrix elements
* Stack:	( MM ? --> MM' ? )
**********************************************************************
NULLNAME MMtrim
::
  MMdims			( MM ? ##Y #X )
  ZERO_DO (DO)
    DUP ZERO_DO (DO)		( MM ? #Y )
      INDEX@ JINDEX@		( MM ? #Y #y #x )
      2DUP MMget		( MM ? #Y #y #x nom den )
      RCanno
      2SWAP MMput		( MM ? #Y )
    LOOP
  LOOP
  DROP				( MM ? )
;

**********************************************************************
* Name:		MMnull
* Desc:		Create a null n x m Meta Matrix
* Stack:	( n m --> MM )
**********************************************************************
NULLNAME MMnull		
::
  2DUP #*			( n m n*m )
  ZERO_DO (DO)			( .. n m )
    HXS 1 0  HXS 2 10		( .. n m 0 1 )
    2SWAP			( .. 0 1 n m )
  LOOP				( 0 1 .. 0 1 n m )
  NULL$				( MM )
;

**********************************************************************
* Name:		MMdims
* Desc:		Get dimensions of meta matrix
* Stack:	( MM ? --> MM ? #Y #X )
**********************************************************************
NULLNAME MMdims
CODE
		ST=0	10		No GC yet
MMdimrestart	GOSBVL	=SAVPTR
		GOSUB	FindMM		D1 = ->dims
		A=DAT1	10
		GOSBVL	=GETPTR
		D=D-CON	A,2
		GOC	+		Huh, pretty low on memory
		D1=D1-	10		but always faster than POP2# + PUSH2#
		DAT1=A	10
		GOVLNG	=Loop
+		GOSBVL	=DOGARBAGE
		GOTO	MMdimrestart
ENDCODE

**********************************************************************
* Name:		MMdrop
* Desc:		Drop meta matrix from stack
* Stack:	( MM ? --> ? )
**********************************************************************
NULLNAME MMdrop
CODE
		GOSBVL	=SAVPTR
		A=0	A
		R0=A	A		x=0
		R1=A	A		y=0
		
		GOSUB	FindMM
		CD1EX
		D1=C
		D0=C
		D0=D0-	5		->NULL$

		GOSUB	GetMM.1		R2[A]=X R3[A]=Y
		D1=D1+	10		->start of MM

		AD1EX
		GOSBVL	=D1=DSKTOP
		AD1EX

		CD0EX
		D0=C
		C=C-A	A		# of stack levels kept *5

		GOSBVL	=MOVEUP		Handles possible 0 right

		AD1EX			->new bottom of stack
		GOSBVL	=GETPTR
		D1=A			Set it
		GOSBVL	=SAVPTR
		GOSBVL	=ADJMEM		Recalculate free memory
		GOVLNG	=GETPTRLOOP	And exit
ENDCODE

**********************************************************************
* Name:		MMsub
* Desc:		Take submatrix of meta matrix
* Stack:	( MM ? #y1 #x1 #y2 #x2 --> ? MM' )
* Warning:	Only end coordinates are checked for concistency
**********************************************************************
NULLNAME MMsub
::
  MMdims			( MM ? #y1 #x1 #y2 #x2 #Y #X )
  #1- ROT #MIN #1+
  UNROT #1- #MIN #1+		( MM ? #y1 #x1 #x2' #y2' )
  4UNROLL SWAP			( MM ? #y2' #y1 #x2' #x1 )
  2DUP #- 5UNROLL		( MM ? #X' #y2' #y1 #x2' #x1 )
  2SWAP 2DUP #- 6UNROLL		( MM ? #Y' #X' #x2' #x1 #y2' #y1 )
  DO
    2DUP
    DO
       JINDEX@ INDEX@ MMget	( MM ? .. #Y' #X' #x2' #x1' nom den )
       6UNROLL 6UNROLL		( MM ? .. nom den #Y' #X' #x2' #x1' )
    LOOP
  LOOP
  2DROP				( MM ? .. #Y' #X' )
  MMdrop			( ? .. #Y' #X' )
  NULL$				( ? MM' )
;

*---------------------------------------------------------------------
* Name:		MMtrn
* Desc:		Transpose of square meta matrix
* Stack:	( MM ? --> MM' ? )
* Notes:	The diagonal elements are swapped with themselves.
*---------------------------------------------------------------------
* NULLNAME MMtrn
* ::
*   MMdims				( MM ? #Y #X )
*   2DUP#= NOTcase SYNTAXERR
*   ZERO_DO (DO)			( MM ? #Y )
*      DUPINDEX@
*      DO
*         INDEX@ JINDEX@ MMget		( MM ? #Y nom1 den1 )
*         JINDEX@ INDEX@ MMget		( MM ? #Y nom1 den1 nom2 den2 )
*         INDEX@ JINDEX@ MMput		( MM ? #Y nom1 den1 )
*         JINDEX@ INDEX@ MMput		( MM ? #Y )
*     LOOP
*   LOOP				( MM ? #Y )
*   DROP				( MM ?  )
* *  MMdims				( MM ? #Y #X )
* *  SWAP 2DUPSWAP			( MM ? #X #Y #Y #X )
* *  SWAP#1-SWAP MMput			( This pokes over the old dims )
* ;

**********************************************************************
* Name: 	MMrowswap
* Desc: 	Swap rows in meta matrix
* Stack:	( MM ? #row1 #row2 --> MM' ?
* Warning:	Has no protection against invalid row indexes
**********************************************************************
NULLNAME MMrowswap
::
  2DUP#= case2DROP		( --> MM ? )
  MMdims SWAPDROP		( MM ? #row1 #row2 #X )
  ZERO_DO (DO)			( MM ? #row1 #row2 )
     DUPINDEX@ MMget		( MM ? #row1 #row2 nom2 den2 )
     4PICK INDEX@ MMget 	( MM ? #row1 #row2 nom2 den2 nom1 den1 )
     5PICK INDEX@ MMput 	( MM ? #row1 #row2 nom2 den2 )
     4PICK INDEX@ MMput 	( MM ? #row1 #row2 )
  LOOP
  2DROP 			( --> MM ? )
;

*---------------------------------------------------------------------
* Name: 	MMcolswap
* Desc: 	Swap columns in MM
* Stack:	( MM ? #col1 #col2 --> MM' ? )
* Warning:	Has no protection against invalid column indexes
*---------------------------------------------------------------------
* NULLNAME MMcolswap
* ::
*   2DUP#= case2DROP		( --> MM ? )
*   MMdims DROP 		( MM ? #col1 #col2 #Y )
*   ZERO_DO (DO)		( MM ? #col1 #col2 )
*      INDEX@ OVER MMget	( MM ? #col1 #col2 nom2 den2 )
*      INDEX@ 5PICK MMget	( MM ? #col1 #col2 nom2 den2 nom1 den1 )
*      INDEX@ 6PICK MMput	( MM ? #col1 #col2 nom2 den2 )
*      INDEX@ 5PICK MMput	( MM ? #col1 #col2 )
*   LOOP
*   2DROP
* ;

**********************************************************************
* Name:		MMputcol
* Desc:		Put the coeffs of Q in col #x of MM starting at #y
* Stack:	( MM ? Q #y #x --> MM' ? )
**********************************************************************
NULLNAME MMputcol	
::
  ROT DUPTYPELIST? ?SKIP ONE{}N		( MM ? #y #x L )
  DUPLENCOMP ZERO_DO (DO)
    DUP ISTOP-INDEX NTHCOMPDROP		( MM ? #y #x L ai )
    HXS 2 10 				( MM ? #y #x L ai di )
    5PICK INDEX@ #+			( MM ? #y #x L ai di #y' )
    5PICK MMput				( MM' ? #y #x L )
  LOOP
  3DROP
;

**********************************************************************
* Name:		MMgetR
* Desc:		Get rational from column zero of MM 
*		with coefficients in rows #start..#stop
* Stack:	( MM ? #stop+1 #start --> P Q )
**********************************************************************
NULLNAME MMgetR		
::
  HXS 1 0 HXS 2 10  2SWAP		( 0 1 #stop+1 #start )
  DO					( P Q )
    INDEX@ ZERO MMget			( P Q r s )
    ONE CKMkP SWAP			( P Q s' r )
    HXS 1 0 ISTOP-INDEX #1- 		( P Q s' r 0 #xpow )
    NDUPN #1+ CKMkP			( P Q s' r*x^pow )
    SWAP RAdd				( P' Q' )
  LOOP
;

**********************************************************************
* Name:		MMget
* Desc:		Get element of meta matrix
* Stack:	( MM ? #y #x --> MM ? nom den )
* Notes:	y and x run from 0 to dim-1
*		( 0 0 == upper left corner, dim-1 dim-1 low right corner )
* Warnings:	There are no overflow checks on the arguments
**********************************************************************
NULLNAME MMget
CODE
		GOSBVL	=SAVPTR
		GOSBVL	=POP2#
		R0=C			R0[A] = x
		R1=A			R1[A] = y
		GOSUB	GetMM		D1 = ->den(y,x)
		A=DAT1	10
		GOSBVL	=GETPTR
		DAT1=A	10		Overwrite #y #x with nom den
		GOVLNG	=Loop
ENDCODE

**********************************************************************
* Name:		MMput
* Desc:		Puts new element into meta matrix
* Stack:	( MM ? nom den #y #x --> MM ? )
* Notes:	y and x run from 0 to dim-1
*		( 0 0 == upper left corner, dim-1 dim-1 low right corner )
* Warnings:	There are no overflow checks on the arguments
**********************************************************************
NULLNAME MMput
CODE
		GOSBVL	=POP2#
		R0=C			R0[A] = x
		R1=A			R1[A] = y
		A=DAT1	10
		R4=A			R4[0-9] = nom den
		D1=D1+	10		\
		D=D+1	A		 2DROP
		D=D+1	A		/
		GOSBVL	=SAVPTR
		GOSUB	GetMM		D1 = ->den(y,x)
		A=R4
		DAT1=A	10		Overwrite old elements
		GOVLNG	=GETPTRLOOP

*'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
* Desc:		Find element y,x in MM
* Input:	R0[A] = x	D1 = ->stack
*		R1[A] = y
* Output:	D1 = ->den(y,x)
*		R2[A] = Y
*		R3[A] = X
*'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
GetMM		GOSUB	FindMM
GetMM.1		GOSBVL	=POP2#		A[A] = Y	C[A] = X
		R3=A			R2[A] = Y
		R2=C			R3[A] = X
		C=R1			C[A] = y
		A=A-C	A		A[A] = Y-y
		C=R2			C[A] = X
		GOSBVL	=MUL#		B[A] = (Y-y)X
		C=R0			C[A] = x
		B=B-C	A		B[A] = (Y-y)X-x
		B=B-1	A
		AD1EX
		B=B+B	A
		A=A+B	A		+ 2*
		B=B+B	A
		B=B+B	A
		A=A+B	A		+ 8* = + 10*
		D1=A			->den(y,x)
		RTN
*'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
* Desc:		Find MM from stack
* Input:	D1 = ->stack
* Output:	D1 = ->dims of MM
*'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
FindMM		LC(5)	=NULL$		Identifies MM
-		A=DAT1	A		Skip in stack..
		D1=D1+	5
		?A#C	A
		GOYES	-		.. until MM is found
		RTN

ENDCODE

**********************************************************************

