	STITLE	ALG Partial Fractions
**********************************************************************
*		ALG Library: Partial Fraction Decomposition
**********************************************************************

**********************************************************************
* Multivariate partial fraction decomposition is based on the
* note that once we do Euclidian division in a multivariate
* rational polynomial
*
*	Q[x1,x2,x3..]     Q1[x1,x2,x3..]    Q2[x1,x2,x3..]
*	-------------  =  -------------- +  --------------
*	R[x1,x2,x3..]       R1[x2,x3..]     R2[x1,x2,x3..]
*
* we separate the problem into two parts where the first part has
* a reduced number of variables. Thus we can interpret Q1 as a sum
* with rational coefficients not depending on x1, and thus partial
* fraction decomposition can now be done with respect to x2,x3..
* The second part is handled by the traditional single-variable
* partial fraction decomposition algorithm.
*
**********************************************************************


**********************************************************************
* EXPORT:
*
*   ParFrac		( symb {vars} --> symb' )
*   PF2A		( Lpf Lvars --> symb' )
*   QParFrac		( P Q #depth --> Lpf )
*   PFDecomp		( pp qc qp lf --> Lpf )
*
**********************************************************************


**********************************************************************
* Name:		ParFrac
* Desc:		Partial fraction expansion with preferred variables
* Stack:	( symb {vars} --> symb' )
**********************************************************************
NULLNAME ParFrac
::
  DUPLENCOMP UNROT			( #vars symb {vars} )
  RSimVars				( #vars P Q Lvars ) 
  UNROT 4ROLL QParFrac			( Lvars Lpf )
  SWAP PF2A
;

**********************************************************************
* Name:		PF2A
* Desc:		Convert partial fraction list to symbolic
* Stack:	( Lpf Lvars --> symb' )
**********************************************************************

* From algr2sym.s
*
* DEFINE Rvars@ 3GETLAM
* DEFINE Rmain@ 2GETLAM

NULLNAME PF2A
::					( Lpf Lvars )
  OVER NULLCOMP? case2drop %0		( --> %0 )
  ZERO 3PICK				( Lpf Lvars #main Lpf )
  { NULLLAM NULLLAM NULLLAM }
  BIND					( Lpf )
  ZEROSWAP LENCOMP 			
  #1+_ONE_DO (DO)			( Meta )
    1GETLAM INDEX@ NTHCOMPDROP		( Meta pfi )
    INCOMPDROP Frac2SS			( Meta sflag Mterm )
    get1 ITE SSub SAdd			( Meta' )
  LOOP
  ABND
  CKSymbN				( symb' )
;

*---------------------------------------------------------------------
* Name:		Frac2SS	
* Desc:		Partial fraction to sign + symbolic meta 
* Stack:	( num den q pow --> sflag Meta )
* Note:		Assumes LAM vars and main are bound
*---------------------------------------------------------------------
NULLNAME Frac2SS
::
  DUP#0= case2drop 
   :: RGetSign R2S ;		( -> sflag Meta )

  THREE{}N >R			( num )
  HXS 2 10 RGetSign R2S		( sflag Mnum )
  'R HXS 2 10 R2SF		( sflag Mnum Mden )
  'R HXS 2 10 R2SF		( sflag Mnum Mden Mq )
  'RRDROP SPow#			( sflag Mnum Mden Mq^pow )
  DUP#0=ITE DROP SMul		( sflag Mnum Mden' )
  DUP#0=ITE DROP SDiv		( sflag Mterm )
;

**********************************************************************
* Name:		QParFrac
* Desc:		Calculate partial fraction expansion
* Stack:	( P Q #depth --> Lpf )
* Notes:	Lpf is a list of terms in the sum.
*
*		Each element in the list is of the form
*
*			{ Qnom Qden Qpf #pow }
*
*		corresponding to
*
*			   nom
*			----------
*			den*pf^pow
*
*		NULL{} is a special case indicating zero
*		Assume P Q are simplified
**********************************************************************
NULLNAME QParFrac
::

* Decrement depth

  #1- UNROT 					( #d Q P )

* Divide problem into two parts

  EuclidDiv					( #d q1 q2 r1 r2 )

* Partial fraction decomposition of the remainder

  PFDecomp UNROT				( #d Lpf q1 q2 )

* Then handle the quotient part

  ( Handle the case quotient = 0 )

  OVER QIsZero? case2drop SWAPDROP		( --> Lpf )

  ( done if quotient is a polynomial or depth is zero )

  DUPTYPEHSTR? 5PICK #0= ORcase 
  :: 4ROLLDROP HXS 2 10 ZERO FOUR {}N >HCOMP ;	( --> Lpf' )

* Handle rational quotient

  ( denominator has no main variable )

  CARCOMP				( #d Lpf q1 q2 )
  SWAP DUPTYPEHSTR? IT ONE{}N		( #d Lpf q2 q1 )

  toLEN_DO (DO)				( #d Lpf q2 q1 )
   :: 
     DUPINDEX@ NTHCOMPDROP		( #d Lpf q2 q1 q1[i] )
     DupQIsZero? caseDROP		
     3PICK RCanno			( #d Lpf q2 q1 q1[i]' q2' )
     6PICK QParFrac			( #d Lpf q2 q1 Lpf[i] )
     ISTOP-INDEX #1- QParFix		( #d Lpf q2 q1 Lpf[i]' )     
     4ROLL &COMP UNROT			( #d Lpf' q2 q1 )
   ;
  LOOP
  2DROP SWAPDROP			( --> Lpf' )
;

*---------------------------------------------------------------------
* Name:		QParFix
* Desc:		Add new main variable multiplier of given degree to
*		partial fraction expansion
* Stack:	( Lpf #degree --> Lpf' )
*---------------------------------------------------------------------
NULLNAME QParFix
::
  SWAP FIVE				( #d Lpf offset )
  BEGIN
    NEXTCOMPOB
  WHILE					( #d Lpf of Lpf[i] )
    INCOMPDROP 4UNROLL			( #d Lpf of' #pow nom den f )
    ONE CKMkP 4UNROLL			( #d Lpf of' f' #pow nom den )
    ONE CKMkP 4UNROLL			( #d Lpf of' den' f' #pow nom )
    HXS 1 0 8PICK NDUPN #1+ CKMkP	( #d Lpf of' den' f' #pow nom' )
    4UNROLL FOUR {}N			( #d Lpf of' Lpf[i]' )
    4UNROLL				( ... #d Lpf of' )
  REPEAT				
  SWAPDROP LENCOMP {}N			( Lpf' )
;

**********************************************************************
* Name:		PFDecomp
* Desc:		Partial Fraction decomposition 
* Stack:	( p q --> Lpf )
* Notes:	Must have deg(p) < deg(q)
*		Returns NULL{} if p=0
**********************************************************************

NULLNAME PFDecomp
::
  OVER QIsZero? case2drop NULL{}	( -> {} )

  PPP ONE CKMkP				( pp qp qc )
  SWAPDUP PFactor DROP			( pp qc qp lf )

  ( Quick exit if lf has only one factor and pp=const or pow=1 )

  NoDecomp? case 
  :: 
     SWAPDROP				( pp qc lf )
     CARCOMP INCOMPDROP			( pp qc pf pow )
     FOUR {}N ONE{}N			( --> { {pp qc pf m} } )
  ;

  ( Else use regular approach )

  DUPLENCOMP				( pp qc qp lf m )
  3PICK LENCOMP #1-			( pp qc qp lf m dim )
  NULL{}				( pp qc qp lf m dim lpf )

  {{ pp qc qp lf m dim lpf }}		( )

  dim DUP#1+ MMnull			( MM0 )
  ZERO					( MM x )
  m #1+_ONE_DO (DO)			( MM x )
     lf INDEX@ NTHCOMPDROP		( MM x {F k} )
     INCOMPDROP				( MM x F k )
     qp SWAP				( MM x F qp k )
     ZERO_DO (DO)			( MM x F R )
        OVER QDiv			( MM x F R' )
        OVER LENCOMP #1-		( MM x F R' deg )
        ZERO_DO (DO)			( MM x F R' )
           DUPINDEX@			( MM x F R' R' y )
           5ROLL DUP#1+			( MM F R' R' y x x+1 )
           6UNROLL			( MM x+1 F R' R' y x )
           MMputcol			( MM' x+1 F R' )
        LOOP				( MM' x' F R' )
     LOOP				( MM' x' F R' )
     2DROP				( MM' x' )
  LOOP					( MM x )

  pp ZERO ROT MMputcol			( MM' )
  MMsolve				( MX )

  ZERO					( MX y )
  m #1+_ONE_DO (DO)			( MX y )
     lf INDEX@ NTHCOMPDROP		( MX y {F k} )
     INCOMPDROP SWAP			( MX y k F )
     DUPLENCOMP #1-SWAP			( MX y k deg F )
     ROT #1+_ONE_DO (DO)		( MX y deg af )

        HXS 1 0 HXS 2 10		( MX y deg af n d )
        4PICK ZERO_DO (DO)		( MX y deg af n d )
           5ROLL DUP#1+ 6UNROLL		( MX y+1 deg af n d y )
           ZERO MMget			( MX y+1 deg af n d ni di )
           ONE CKMkP SWAP		( MX y+1 deg af n d di' ni )
           HXS 1 0 INDEX@ NDUPN		( MX y+1 deg af n d di' ni Mx^j )
           #1+ CKMkP			( MX y+1 deg af n d di' ni' )
           SWAP RAdd			( MX y+1 deg af n' d' )
        LOOP				( MX y' deg af n' d' )

        OVER QIsZero? ITE
           2DROP			( MX y' deg af )
           ::
             qc QMul			( MX y' deg af n d*qc )
             RCanno			( MX y' deg af n' d' )
             3PICK INDEX@		( MX y' deg af n' d' af #pow )
             FOUR {}N			( MX y' deg af lpf[i] )
             lpf SWAP >TCOMP lpf!	( MX y' deg af )
           ;
     LOOP				( MX y' def af )
     2DROP				( MX y' )
  LOOP					( MX y' )
  DROP MMdrop				( - )
  lpf ABND				( Lpf )
;

*---------------------------------------------------------------------
* Name:		NoDecomp?
* Desc:		Check if only possible decomposition is trivial
* Stack:	( pp qc qp lf --> pp qc qp lf flag ) 
*---------------------------------------------------------------------
NULLNAME NoDecomp?
::
  DUPLENCOMP #1<> caseFALSE		( if |lf| > 1 --> False )
  4PICK TYPEHSTR? caseTRUE		( else if pp=cont --> True )
  DUP CARCOMP TWO NTHCOMPDROP #1=       ( else --> pow =?= 1 )
;					

**********************************************************************

