	STITLE	ALG PRIMES
**********************************************************************
*	ALG Library:	Integer primality testing & factorization
**********************************************************************

**********************************************************************
*	Integer Factorization and Primality Testing
*	===========================================
*
* Most useful factorization algorithms fall into one of two classes:
*
* A. The run time depends mainly on the size of N, the number being
*    factored, and is not strongly dependant on the size of the factor
*    found. Examples are:
*	- Lehman's algorithm with rigorous worst-case run time O(N^1/3)
*	- Shanks' SQUFOF algorithm with expected run time O(N^1/4)
*	- Shanks' Class Group algorithm with run time O(N^(1/5+e))
*	- The Continued Fraction algorithm, O(exp(c(logN log logN)^1/2))
*	- Multiple Polynomial Quadratic Sieve, O(exp(c(logN log logN)^1/2))
*
* B. The run time depends mainly on the size of f, the factor found.
*    (We can assume that f <= N^1/2.) Examples are:
*	- Trial division algorithm, O(f)
*	- Pollard rho algorithm, O(f^1/2)
*	- Lenstra's Elliptic Curve Method (ECM), O(exp((2 logf log logf)^1/2))
*	- Pollard p-1, O(m) where m is the upper bound on factors of p-1
*	  (Worst case time bound is the same as for trial division!)
*
* Usually one must allow an additional factor of the form (logN)^2 for the
* cost of performing arithmetic operations on numbers of size O(N) or O(N^2),
* but could theoretically be replaced by (logN)^(1+e) for any e>0.
*
* For a randomly chosen N there is a 50 percent chance that the second
* largest prime factor < N^0.212. Thus if we have an algorithm in class
* B which can find factors of size 10^22 in a reasonable time, there is
* a 50 percent chance that the algorithm will be able to completely factor
* a random number N of size about 10^100.
*
* Standard method nowadays is to:
*	- Trial division for very small factors (1000-10000?)
*	- Pollard rho or Brent-Pollard to find small factors
*	- ECM to find small/medium size factors
*	- MPQS to find medium/large size factors
*	- NFS to find large size factors
*
* The current practical upper limit on general numbers is about 120
* digits. Note that for example the RSA modulus has about 155 digits.
* Numbers that have a special form can use specialized factorization
* algorithms, numbers up to 155 digits or more have already been factored.
* See for example the Cunningham project n such factorizations.
*
* A list of factorization algorithms for finding references:
*
*	- Trial division
*	- Fermat's Factoring Method
*	- Euler's Factoring Method
*	- Gauss's Factoring Method
*	- Legendre's Factoring Method
*	- Pollard's p-1 Method
*	- Williams' p+1 Method
*	- Pollard's rho Method
*	- Brent-Pollard
*	- Shanks' Factoring Method, SQUFOF
*	- Shanks' Class Group Method
*	- Dixon's Random Squares Method
*	- Valle's Two-Thirds Method
*	- Seysen's Class Group Algorithm
*	- Morrisons and Brillhart's Continued Fraction Method, CFRAC
*	- The Quadratic Sieve, QS
*	- The Multiple Polynomial Quadratic Sieve, MPQS
*	- The Schnorr-Lenstra Method
*	- Lenstra's Elliptic Curve Method, ECM
*	- The Number Field Sieve, NFS
*	- The General Number Field Sieve, GNFS
*
* As opposed to integer factorization primality proving is a pretty much
* solved problem, deterministic algorithms do exist which give the result
* in a reasonable amount of time. ALG48 uses a probabilistic Miller-Rabin
* test though to avoid long execution times due to the slow CPU.
* For a good starting point search the net for 'primality.dvi'.
*
**********************************************************************


**********************************************************************
* Notation convention:
*
*	Zs	= signed infinite precision integer
*	Z	= infinite precision integer (positive)
*	S	= short integer ( <=60 bits )
*	B	= big integer ( >60 bits )
*
*	64 bits is not used so we don't have to worry about
*	special overflows when using W calculations.
**********************************************************************

**********************************************************************
* EXPORT:
*
*	Prime+		( Z --> Z' )
*	Prime-		( Z --> Z' )
*
*	ZFactor		( Zs --> Lf )
*	NFactor		( Z  --> Lf )
*	NFactorSpc	( Z  --> Lf / Z1 )
*	DupTypeS?	( Z --> S TRUE | B FALSE )
*
*	SFactor		( S  --> Lf )
*	SPollard	( S --> S1 S2 )
*
*	BFactor		( B  --> Lf )
*	BrentPow	( Za Z1 Z2 Zn #k --> Z )
*
*	ZPrime?		( Z --> flag )
*	ZIsPrime?	( Z --> flag )
*	SIsPrime?	( S --> flag )
*	BIsPrime?	( B --> flag )
*	BRabin		( Z #base --> Z flag )
*
*	ZTrialPrime?	( Z --> flag )
*	ZTrialDiv	( Z --> Mf Z' )
*
**********************************************************************

**********************************************************************
* Desc:		Return next prime	( Z' > Z )
* Stack:	( Z --> Z' )
**********************************************************************
NULLNAME Prime+
::
  DupZIsNeg?				( For Z<0 )
  OVER QIsZero? OR			( And for Z=0 )
  OVER ZIsOne? OR			( And for Z=1 )
  casedrop HXS 2 20			( Return Z=2 )

  DupZIsEven? IT :: HXS 2 10 QSub ;	( Make trial Z odd )
  BEGIN
	HXS 2 20 QAdd DUP ZPrime?	( Terminates at prime )
  UNTIL
;

**********************************************************************
* Desc:		Return previous prime	( Z' < Z )
* Stack:	( Z --> Z' )
**********************************************************************
NULLNAME Prime-
::
  DupZIsNeg?				( For Z<0 )
  OVER QIsZero? OR			( And for Z=0 )
  OVER ZIsOne? OR			( And for Z=1 )
  casedrop HXS 2 20			( Return Z=2 )

  DupZIsEven? IT :: HXS 2 10 QAdd ;	( Make trial Z odd )
  BEGIN
	HXS 2 20 QSub DUP ZPrime?	( Terminates at 1 or prime )
  UNTIL
  DupZIsOne? casedrop HXS 2 20		( Max[2,prime] )
;

**********************************************************************
* Desc:		Return N'th prime
* Stack:	( # --> Z )
**********************************************************************
*	NULLNAME PrimeN
*	::
*	  DUPTWO #< casedrop HXS 2 20		( Return smallest prime )
*	
*	  HXS 2 10 SWAP				( Z #n )
*	  ONE_DO (DO)
*		HXS 2 20 QAdd			( Z+2 )
*		DUP ZPrime?			( Z+2 flag )
*		ITE ONE ZERO			( Advance loop or not? )
*	  +LOOP
*	;

**********************************************************************
* Desc:		Factor signed long integer
* Stack:	( Zs --> Lf )
**********************************************************************
NULLNAME ZFactor
::
  DupZIsNeg? NOTcase NFactor	( Lf )
  { HXS 2 1F }
  SWAP QNeg
  DupZIsOne? caseDROP		( {-1} for -1 )
  NFactor &COMP			( Lf' )
;

**********************************************************************
* Desc:		Factor positive long integer
* Stack:	( Z --> Lf )
**********************************************************************
NULLNAME NFactor
::
  ZTrialDiv				( Mf Z' )
  DupZIsOne? casedrop {}N		( Trial division did all the work )
  DupTypeS? ITE SFactor	BFactor		( Mf Lf )
  INNERCOMP top& {}N			( Lf' )
;

**********************************************************************
* Desc:		Semi-factor positive long integer
* Stack:	( Z --> Lf )
* Notes:	This is regular factorization with an extra
*		'hopeless?' test.
**********************************************************************
NULLNAME NFactorSpc
::
  ZTrialDiv				( Mf Z' )
  DupZIsOne? casedrop {}N		( Trial division did it all )
  ::
    DupTypeS? case SFactor		( Mf Lf )
    ZBits 100 #> case ONE{}N		( Mf {Z} - seems hopeless )
    BFactor				( Mf Lf )
  ;
  INNERCOMP top& {}N			( Lf' )
;

**********************************************************************
* Desc:		Test if Z is short ( <= 64 bits )
* Stack:	( Z --> Z flag )
**********************************************************************
NULLNAME DupTypeS?
CODE
		C=DAT1	A		->Z
		CD0EX
		D0=D0+	5
		A=DAT0	A		Zsize
		D0=C
		LC(5)	5+16		Sign is included!
		?A<=C	A
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE

**********************************************************************
*		SHORT INTEGER FACTORIZATION
**********************************************************************

**********************************************************************
* Desc:		Factor short integer
* Stack:	( S --> Lf )
* Algorithm:
*
*	Pollard Rho, with the assumption that trial division has been done
*	already. Thus any factor less than 4012009 is known to be a prime,
*	for greater factors a primality test is used before calling the
*	actual Pollard Rho. 
*
*	Pollard Rho does not find the factors in order of magnitude, thus the
*	results will be sorted after full factorization has been achieved.
*
**********************************************************************
NULLNAME SFactor
::
   ONE{}N NULL{}			( Ltodo Lf )
   BEGIN
	SWAP				( Lf Ltodo )
	DUPNULLCOMP? NOT		( Lf Ltodo flag )
   WHILE
	INNERCOMP #1-{}N UNROT		( Ltodo' Lf p )

	DUP HXS 7 9E73D30 ZNLT?		( Ltodo Lf p less? )
	case >TCOMP			( Ltodo Lf' )

	DUP SIsPrime?			( Ltodo Lf p prime? )
	case >TCOMP			( Ltodo Lf' )

	SPollard			( Ltodo Lf p1 p2 )

	4ROLL INNERCOMP #2+ {}N SWAP	( Ltodo' Lf )
   REPEAT
   DROP					( Lf )

   ' :: ZNLT? NOT ;
   SortList

;


**********************************************************************
* Desc:		Factor short integer into 2 parts using
*		Pollard Rho algorithm.
* Stack:	( S --> S1 S2 )
* Notes:
*		Trial division and primality tests should be
*		done prior to calling this subroutine, otherwise
*		an eternal loop is risked.
*
*		The random number generator is modeled after
*		the user level RAND command, although the starting
*		value is different.
*
**********************************************************************
NULLNAME SPollard
CODE

* Register usage:
*
*	B[W]	n (aka S)
*	R0[W]	Y
*	R1[W]	X
*	R2[W]	k
*	R4[W]	random number generator status
*	D1	Pollard Rho loop counter before new %RAN is generated
*	D0	modular exponentiation loop counter

		GOSBVL	=PopASavptr
		D0=A
		D0=D0+	5
		C=DAT0	A
		C=C-CON	A,7			Assume non-zero
		D0=D0+	5
		P=C	0
		C=0	W
		C=DAT0	WP			n
		P=	0
		B=C	W			n

* Initialize random number generator.

		A=0	W
		R4=A	W

* Loop until Pollard succeeds. I have no idea whether 1000 is a good
* value or not, I simply defined some value to make sure another
* random value will be used eventually.

-		D1=(5)	1000-1			1000 retries
		GOSUB	Pollard
		GOC	-			Try again until success

		R2=C	W			p2 = n/p1
		GOSUBL	PushN
		GOSBVL	=SAVPTR
		A=R2	W			p2 = n/p1
		GOLONG	PushNLoop

***
*** Pollard-Rho
***	Input:	B[W]	n
***		R4[W]	rand
***		D1	loop counter
***	Output:
***		CC:	A[W]	n/p
***			C[W]	p
***		CS:	B[W]	n		(failure)

Pollard		A=R4				rand
		?A#0	W
		GOYES	+
		LAHEX	4915EF8AC18A09EC	16 digits
+		LCHEX	35D65A4D7920		multiplier (from %RAN)
		GOSBVL	=MPY
		R4=C				rand'

* No need to take absolute modulo, roughly is enough

-		?C<B	W
		GOYES	+
		CSRB
		GONC	-
+		R0=C				Init Y
		R1=C				Init X

PollardLoop	C=0	W
		C=C+1	A
		R2=C				Init k = 1

* Calculate n size to speed up the modular * loop

		P=	15
-		?B#0	P
		GOYES	+
		P=P-1
		GONC	-
+
		D0=(5)	50-1			^ 50 times

RhoLoop		A=R0				A[W] = Y
		C=A	W
		GOSUB	mulmodrho		^2
		A=A+1	W			+1
		C=A	W
		GOSUB	mulmodrho		^2
		A=A+1	W			+1
		R0=A

		A=R1				A[W] = X
		C=A	W
		GOSUB	mulmodrho		^2
		A=A+1	W			+1
		R1=A

		C=R0				Y
		?A>=C	W
		GOYES	+
		ACEX	W
+		A=A-C	W			|Y-X|
		C=R2				C[W] = k
		GOSUB	mulmodrho
		R2=C
		
		D0=D0-	1
		GONC	RhoLoop
		P=	0			Forget the optimized WP now

		A=B	W			A[W]=n  C[W]=k
		GOSUBL	SHORTGCD		A[W]=gcd(k,n)
		A=A-1	W
		?A#0	W			Done if gcd <> 1
		GOYES	+

		D1=D1-	1			Retry?
		RTNC				No - get new random start
		GOTO	PollardLoop

+		A=A+1	W			p
		R2=A
		C=A	W			p
		A=B	W			n
		?A=C	W
		RTNYES				Huh?
		GOSBVL	=IDIV
		A=R2				p
		RTNCC				CC: factored

** Modular multiplication where P limits modulo size for speed
**	A[W]=C[W] = A*C mod B

mulmodrho	D=C	W
		C=0	W
		SB=0
mulmodrholoop	DSRB
		?SB=0
		GOYES	++
		SB=0
		C=C+A	WP
		GOC	+
		?C<B	WP
		GOYES	++
+		C=C-B	WP
++		A=A+A	WP
		GOC	+
		?A<B	WP
		GOYES	++
+		A=A-B	WP
++		?D#0	B
		GOYES	mulmodrholoop
		?D#0	WP
		GOYES	mulmodrholoop
		A=C	W
		RTNCC

ENDCODE

**********************************************************************
*		LONG INTEGER FACTORIZATION
**********************************************************************


**********************************************************************
* Desc:		Factor long integer
* Stack:	( N --> Lf )
* Algorithm:
*		Brent-Pollard, with the assumption that trial
*		division has been done already. When a small
*		factor is found SFactor is called to get full
*		short factorization. Since the factorization can
*		potentially take a very long time, an execution
*		time test is used to abort factoring very long
*		integers (limit is 60s for each composite).
*
*		The factors are sorted at exit.
**********************************************************************

NULLNAME BFactor
::

  DUP BIsPrime? case ONE{}N	( return Lf if prime )

  SysTime			( N time )
  NULL{} NULL{}			( N time todo done )
  HXS 1 0 DUPDUP DUPDUP		( N time todo done Z X Y Q YS )
  #ZERO#ONE ZERO TEN		( N time todo done Z X Y Q YS K R Iter M )

  {{ n Btime Ltodo Ldone z x y q ys k r Iter m }}

* Main Brent-Pollard loop

*  ONE TestUserFlag IT DOCLLCD	( Clear screen if status is to be shown )

  BEGIN

	SysTime Btime!		( Save start time for current N )

	y n QDivRem SWAPDROP y!			( Make sure Y is mod N )

	HXS 2 10 q!				( Init Q )

	BEGIN

		ZERO k!				( Init K )
		y x!				( X=Y )
		HXS 2 30 y DUP			( Y = [Y*Y+3]^R mod N )
		n r BrentPow y!

		BEGIN
			Iter #1+ Iter!
			y ys!			( YS=Y )
			m r k #- #MIN
			ZERO_DO (DO)
				HXS 2 30 y DUP n ZERO BrentPow y!

				HXS 1 0 y x QSub
				DupZIsNeg? IT :: n QAdd ;
				q n ZERO BrentPow q!
			LOOP
			q n ZGcd z!			( Z=GCD[Q,N] )
			k m #+ k!			( K=K+M )
			k r #<				( K<R )
			z ZIsOne?			( Z==1 )
			AND
		NOT_UNTIL (UNTIL)

		r #2* r!				( R=R*2 )
		z ZIsOne? NOT				( Z<>1 )
		SysTime Btime bit-
		HXS 10 0008700000000000			( 60*8192 = 60s )
		HXS>HXS %1 %=				( exectime>limit? )
		OR
	UNTIL

	( Trace back if necessary )
	z n EQUAL IT
	::
	  BEGIN
		HXS 2 30 ys DUP n ZERO BrentPow ys!
		ys x QSub
		DupZIsNeg? IT :: n QAdd ;
		n ZGcd z!
		z ZIsOne?
	  NOT_UNTIL (UNTIL)
	;

* See if we exceeded execution time limit by checking Z

	z ZIsOne? ITE					( abort? )

	( * Execution time exceeded * )
	::
	  Ldone INNERCOMP				( return factors )
	  Ltodo INNERCOMP top&				( we managed to find )
	  n SWAP#1+					( so far )
	  {}N Ldone!
	  TRUE						( flag we are done )
	;

	( * Found a factorization for N * )
	::
	  z n z QDiv					( Z N/Z )
	  2DUP ZNLT? ?SKIP SWAP				( Zsmall Zbig )

	  Ltodo INNERCOMP #2+ {}N Ltodo!		( add both to todo )

	  ( Get next composite to factor from the to-do list )
	  BEGIN
	    ::
		Ltodo DUPNULLCOMP? casedrop TrueTrue

		INNERCOMP #1-{}N Ltodo!

		DupTypeS? case
		:: SFactor Ldone &COMP Ldone! FALSE ;

		DUP BIsPrime? case
		:: Ldone INNERCOMP #1+ {}N Ldone! FALSE ;

		n! FalseTrue
	    ;
	  UNTIL
	;
  UNTIL

  Ldone ABND

   ' :: ZNLT? NOT ;
   SortList

;

**********************************************************************
* Desc:		Modular * + ^ mod for Brent-Pollard factorization
* Stack:	( Za Z1 Z2 Zn #k --> Z )
* Notes:
*		Output is Z1*Z2+Za mod Zn repeated k times
*		Note that k=0 and k=1 give the same result.
*		Also Z1<>Z2 makes no sense for k<>0.
*
*		All arguments are assumed to be positive.
*		Za is assumed to be < 16.
*
*		In some instances k can be a very high number,
*		thus it might make sense to use Montgomery
*		multiplication.
**********************************************************************
NULLNAME BrentPow
CODE
		GOSBVL	=POP#		k
		GOSBVL	=SAVPTR
		C=A	A
		RSTK=C			RSTK = k

* Calculate and allocate work memory:
*	Twice Zn for * result
*	Then 1*Zn for quotient
*	Then 1*Zn for remainder
*	= 4*Zn

		A=DAT1	A		->Zn
		D0=A
		D0=D0+	5
		A=DAT0	A		Znsize
		A=A+A	A		2*Znsize
		A=A+A	A		4*Znsize
		A=A+A	A		8*Znsize	\ Chicken
		LC(5)	400				/
		C=C+A	A
		GOSBVL	=GETTEMP
		AD0EX
		R4=A			R4[A]=->output
		R3=A			R3[A]=->work memory

		GOSBVL	=D1=DSKTOP
		D1=D1+	10
		A=DAT1	A
		D0=A			->Z1
		D1=D1-	5
		A=DAT1	A
		D1=A			->Z2

* Note: MULZZentry does not use R4[A], the rest is volatile

		GOSUBL	MULZZentry	R3[A]=->product tail

brentpowloop	GOSBVL	=D1=DSKTOP
		D1=D1+	15
		C=DAT1	A		->Za
		D0=C
		D0=D0+	5
		C=DAT0	A
		C=C-CON	A,7
		GOC	brent+ok	Za = 0 - skip addition phase
		D0=D0+	5
		A=0	W
		A=DAT0	1		A[W] = Za (single nibble only!)

		C=R4	A		->Z (the product)
		D0=C
		D0=D0+	5
		C=DAT0	A
		C=C-CON	A,7
		GONC	+		Regular addition to non-zero

		C=C+CON	A,8		Size = 2
		DAT0=C	A
		D0=D0+	5
		DAT0=A	B		Za + positive sign
		D0=D0+	2
		CD0EX
		R3=C	A		->Ztail (for division)
brent+ok	GOTO	brentmod	+Za done - take modulo now

+		D0=D0+	5		Add Za to nonzero Z
		C=DAT0	W
		C=C+A	W
		DAT0=C	W
		GONC	+		No carry from Za addition
-		D0=D0+	16		Add 1 until no more CRY
		C=DAT0	W		Guaranteed to finish as long
		C=C+1	W		as Z is positive!
		DAT0=C	W
		GOC	-
+		C=R3	A		Now check if the size increased
		D0=C			->Ztail
		D0=D0-	1
		A=DAT0	B		sign nibble
		?A=0	P
		GOYES	brentmod	Sign is ok ==> + didn't reach it
		C=C+1	A
		R3=C	A		->Ztail++
		D0=D0+	1
		A=0	A
		DAT0=A	B		sign = +
		C=R4	A
		D0=C			->Z
		D0=D0+	5
		C=DAT0	A
		C=C+1	A		Zsize++
		DAT0=C	A

brentmod	A=R4	A
		D0=A			->Z
		GOSBVL	=ASLW5
		A=R3.F	A		->Ztail (->Output for div)
		R0=A			R0[W] is unchanged over ZQUO
		GOSBVL	=D1=DSKTOP
		A=DAT1	A
		D1=A			->Zn
		GOSUBL	ZQUO
		A=R0
		D0=A			->quotient
		GOSBVL	=ASRW5
		R4=A			->Z
		D1=A			->Z
		D0=D0+	5
		C=DAT0	A
		AD0EX
		A=A+C	A
		D0=A
		D0=D0+	6+5		->remainder size
		C=DAT0	A
		C=C+CON	A,5
		D0=D0-	5
		GOSBVL	=MOVEDOWN	Z = remainder
		CD1EX
		R3=C	A		->Ztail (new one)

		C=RSTK
		C=C-1	A		k--
		GOC	brentpowdone
		?C=0	A
		GOYES	brentpowdone	
		RSTK=C

* Square Z back onto itself

		A=R4	A
		D0=A			->Z
		D1=A			->Z
		GOSUBL	MULZZentry	Output goes to R3[A] = ->Ztail

* Move the result back down. D0=D1 = ->Ztail = ->product
* Could just let it be there for addition & division, and then
* finally move the remainder down..

		A=R4	A		->Z
		D1=A			->Z
		
		D0=D0+	5		->Z*Zsize
		C=DAT0	A
		C=C+CON	A,5
		D0=D0-	5
		GOSBVL	=MOVEDOWN
		CD1EX
		R3=C	A		->Ztail (new)
		GOTO	brentpowloop

brentpowdone	A=R4	A		->Z
		R0=A	A		R0[A] = ->Z
		D0=A
		D0=D0+	5
		C=DAT0	A		Zsize
		C=C+CON	A,5
		C=C+A	A		C[A] = ->Ztail
		GOSUBL	ShrinkQ
		D1=D1+	15		Drop 3DROP
		D=D+CON	A,3
		GOVLNG	=OverWrALoop
ENDCODE


**********************************************************************
*		PRIMALITY TESTING
**********************************************************************

**********************************************************************
* Theory:
*
*	Miller-Rabin primality test with bases 3,2,5,7..
*
*	An _odd_ composite number n (with n-1 = d*2^s, d odd)
*	either is a strong pseudoprime to base a if
*	(i)  a^d       ==  1 (mod n) or
*	(ii) a^(d*2^r) == -1 (mod n), for some r in 0 <= r < s
*
*	Note that if n is prime then either (i) or (ii) must hold.
*
*	The probability of declaring a composite prime is
*	less than 4^(-k).
*
* Data:
*   k	p(k)	h(k)
*   1	  2	2047				(23.89)
*   2	  3	1373653				(829.1657)
*   3	  5	25326001			(2251.11251)
*   4	  7	3215031751			(71.3461.9421)
*   5	 11	2152302898747			(6763.10627.29947)
*   6	 13	3474749660383			(1303.16927.157543)
*   7	 17	341550071728321			(10670053.32010157)
*   8	 19	341550071728321			(10670053.32010157)
*   9	 23	<=41234316135705689041		(4540612081.9081224161)
*  10	 29	<=1553360566073143205541002401	(22754930352733.P14)
*  11	 31	<=56897193526942024370326972321	(137716125329053.P15)
*  12	 37
*
* Reference:	G. Jaeschke
*		Mathematics of Computation
*		Vol 61, No 204, Oct 1993, pp 915-926
*
* Warning:	Miller-Rabin test fails for integers <= 3, so they
*		must be tested separately at the start.
* Note:
*		According to Pinch commercial software packages use
*		only about 5-10 bases by default, maximum around 25.
*		The latest versions usually implement a deterministic
*		algorithm though.
*
**********************************************************************


**********************************************************************
* Desc:		Primality test for a positive integer.
* Stack:	( Z --> flag )
* Notes:	Returns TRUE for 0 and 1 too.
**********************************************************************
NULLNAME ZPrime?
::
  DUP ZTrialPrime? NOTcasedrop FALSE	( trial division says composite )

  DUP HXS 7 9E73D30 ZNLT? casedrptru	( answer known already? )

  ZIsPrime?				( probabilistic test )
;

**********************************************************************
* Desc:		Probabilistic primality test for a positive integer.
* Stack:	( Z --> flag )
**********************************************************************
NULLNAME ZIsPrime?
::
  DupTypeS? case SIsPrime?
  BIsPrime?
;

**********************************************************************
* Desc:		Test if positive short Z is prime.
* Stack:	( S --> flag )
*
* Notes:	M-R test fails for integers <= 3, so we just
*		test them separately at the start. For convenience
*		lets define 0 and 1 to be primes also..
*
**********************************************************************
NULLNAME SIsPrime?
CODE

* Help macro to provide test base and upper limit for confirmation

M-RLIM	MACRO	$base,$limit
	CON(2)	$1
	CON(1)	1		Limit exists-flag
	NIBHEX	$2		Limit. (Digits in reverse order!)
M-RLIM	ENDM

* Help macro to provide test base with unknown limit for confirmation
M-R	MACRO	$base
	CON(2)	$1
	CON(1)	0		No limit exists - flag
M-R	ENDM


		GOSBVL	=PopASavptr	A[A] = ->Z
		D0=A
		D0=D0+	5
		C=DAT0	A		Zsize
		D0=D0+	5
		C=C-CON	A,7
		GONC	+		Continue if non-zero
goSisprime	GOTO	Sisprime	Return TRUE for zero
goSisnotprime	GOTO	Sisnotprime

+		P=C	0
		A=0	W
		A=DAT0	WP		Z
		P=	0
		?ABIT=0	0
		GOYES	goSisnotprime	Divisible by 2 - quick exit
		C=0	W
		LC(1)	3
		?A<=C	W
		GOYES	goSisprime	Return TRUE for Z <= 3

		B=A	W		Z
		GOSUB	+

		M-R	 2			Base  2, 2047
		M-R	 3,5D5F410000000000	Base  3, 1373653
		M-RLIM	 5,1B17281000000000	Base  5, 25326001
		M-RLIM	 7,7CD71AFB00000000	Base  7, 3315031751
		M-RLIM	11,B3EEF3F15F100000	Base 11, 2152302898747
		M-R	13,FDC1837092300000	Base 13, 3474749660383
		M-R	17,1C8C2B253A631000	Base 17, 341550071728321
		M-R	19			Base 19, 341550071728321
		M-R	23			Base 23, ?
		M-R	29			Base 29, ?
		CON(2)	0			Terminate testing

+		C=RSTK
		D0=C				->tests
SIsPrimeLoop	C=0	W
		C=DAT0	B			base
		D0=D0+	2
		?C=0	A
		GOYES	Sisprime		Final test done - assume prime
		GOSUB	MillerRabin
		GOC	Sisnotprime
		A=DAT0	B
		D0=D0+	1
		?A=0	P
		GOYES	SIsPrimeLoop		Not limit to test
		A=DAT0	W
		D0=D0+	16
		?B>A	W
		GOYES	SIsPrimeLoop		Continue if Z > limit
		?B=A	W
		GOYES	Sisnotprime		Z == limit - not prime
Sisprime	GOVLNG	=GPPushTLoop		TRUE for prime
Sisnotprime	GOVLNG	=GPPushFLoop		FALSE for composite

*** Miller-Rabin primality test with base A.
*** Input:	B[W] = n
***		C[W] = A
*** Output:	B[W] = n
***		CS if test detects composite
***
*** Alters:	A[W] C[W] D[W] R0[W] R1[W] R2[W] R3[W] R4[W]


MillerRabin	R4=C	W		R4[W] = A

* Step 1: factor n-1 = d*2^s

		C=B	W		n
		C=C-1	W		d = n-1
		SB=0
-		CSRB			d = d/2
		?SB=0			s is not calculated to save registers
		GOYES	-		it can be inferred from M later
		C=C+C	W
		C=C+1	W		Fix back
		R3=C	W		R3[W] = M

* Step 2: compute a^d mod n

		A=R4	W		A
		GOSUBL	SHORTPOWMOD	A[W] = A^M mod n

* Step 3: if pow == 1 then possible prime

		A=A-1	W		pow-1
		?A=0	W
		GOYES	milrabfail	pow = 1 ==> is prime
		A=A+1	W		pow

* Step 4: If pow == n-1 then probably prime
*	  (for any s > 0 )

milrabloop	R2=A	W		pow
		C=B	W		N
		C=C-1	W		N-1

		A=R3	W		M
		?A=C	W		M == n-1 ?
		RTNYES			Yes - all tested ; CS for prime

		A=R2	W		pow
		?A=C	W
		GOYES	milrabfail	pow == n-1 ==> is prime

		C=A	W		pow
		GOSUBL	SHORTMULMOD	A[W] = pow^2 mod N
		C=R3	W
		C=C+C	W		M *= 2
		R3=C
		GONC	milrabloop

milrabfail	RTNCC			CC: Failure

ENDCODE
**********************************************************************
* Desc:		Test if positive long Z is prime
* Stack:	( S --> flag )
**********************************************************************
NULLNAME BIsPrime?
::
  DupZIsEven? casedrpfls		( Abort if divisible by 2 )

  TWO		BRabin casedrpfls	( Base  2 test, limit 2047 )
  THREE		BRabin casedrpfls	( Base  3 test, limit 1373653 )
  FIVE		BRabin casedrpfls	( Base  5 test, limit 25326001 )
  SEVEN		BRabin casedrpfls	( Base  7 test, limit 2315031751 )
  ELEVEN	BRabin casedrpfls	( Base 11 test, limit 2152302898747 )
*  THIRTEEN	BRabin casedrpfls	( Base 13 test, limit 3474749660383 )
*  SEVENTEEN	BRabin casedrpfls	( Base 17 test, limit 341550071728321 )
*  NINETEEN	BRabin casedrpfls	( Base 19 test, limit 341550071728321 )
*  TWENTYTHREE	BRabin casedrpfls	( Base 23 test )
*  TWENTYNINE	BRabin casedrpfls	( Base 29 test )
*  THIRTYONE	BRabin casedrpfls	( Base 33 test )
*  THIRTYSEVEN	BRabin casedrpfls	( Base 37 test )
*  FORTYONE	BRabin casedrpfls	( Base 41 test )
*  FORTYTHREE	BRabin casedrpfls	( Base 43 test )
*  FORTYSEVEN	BRabin casedrpfls	( Base 47 test )
*  FIFTYTHREE	BRabin casedrpfls	( Base 53 test )
*  FIFTYNINE	BRabin casedrpfls	( Base 59 test )
*  SIXTYONE	BRabin casedrpfls	( Base 61 test )
*  67		BRabin casedrpfls	( Base 67 test )
*  71		BRabin casedrpfls	( Base 71 test )
*  73		BRabin casedrpfls	( Base 73 test )
*  79		BRabin casedrpfls	( Base 79 test )
*  83		BRabin casedrpfls	( Base 83 test )
*  89		BRabin casedrpfls	( Base 89 test )
*  97		BRabin casedrpfls	( Base 97 test )

  DROPTRUE				( Assume prime )
;

**********************************************************************
* Desc:		Perform Miller-Rabin test for long pos integer.
* Stack:	( Z #base --> Z flag )
* Notes:	Returns TRUE if base witnesses composite.
*		Else returns FALSE.
**********************************************************************
NULLNAME BRabin
::

  #>Z					( N A )

* Calculate d: n-1 = d*2^s

   OVER HXS 2 10 QSub			( N A N-1 )

   CODE
		GOSBVL	=SAVPTR		Count the power of two
		A=DAT1	A
		D0=A
		D0=D0+	10
		A=0	A		pow2 = 0
-		C=DAT0	B	<---+
		D0=D0+	1	    |
		?C#0	P	    |
		GOYES	+	--+ |
		A=A+1	A	  | |	pow2++
		GONC	-	----+	This will terminate for nonzero N
+		A=A+A	A	<-+
		A=A+A	A		pow2 << 2
		SB=0
		A=A-1	A
-		A=A+1	A	<---+
		CSRB.F	P	    |
		?SB=0		    |
		GOYES	-	----+
		GOVLNG	=PUSH#ALOOP
   ENDCODE

					( N A N-1 #pow2 )

   HXS 2 20 SWAP PPow# QDiv		( N A D )

* Compute a^d mod n

   SWAPOVER 4PICK ModPow		( N D A^DmodN )

* If pow == 1 then possible prime

   DupZIsOne? case2drop FALSE		( Possible prime )

   ( N D POW )

   3PICK HXS 2 10 QSub UNROT		( N N-1 D POW )

   BEGIN
	3PICK3PICK EQUAL		( N N-1 D POW  N-1==D )
	case2drop
	:: RDROP DROPTRUE ;		( Wittness composite )

	3PICKOVER EQUAL			( N N-1 D POW  N-1==POW )
	case2drop
	:: RDROP DROPFALSE ;		( Possible prime )

	DUP 5PICK ModMul		( N N-1 D POW' )

	SWAPDUP QAdd SWAP		( N N-1 2*D POW' )
   AGAIN
;


**********************************************************************
*		TRIAL DIVISION UTILITY
**********************************************************************

**********************************************************************
* Desc:		Trial division primality test for a positive integer.
* Stack:	( Z --> flag )
**********************************************************************

NULLNAME ZTrialPrime?

CODE

sTRIALPRIME	EQU 9		Flag to indicate test/factor

		ST=1	sTRIALPRIME	Test for composite

		GOTO	trialentry
ENDCODE

**********************************************************************
* Desc:		Trial division of a positive integer.
* Stack:	( Z --> Mf Z' )
* Notes:
*		If Z' is one then full factorization was achieved.
*
*		The long trial division is not too slow, since
*		division by short integer is quite fast.
*
*		The quotient is also checked so that a final factor
*		less than 2000^2 will also be automatically detected.
*
**********************************************************************

NULLNAME ZTrialDiv

CODE

MAXTRIAL	EQU 2000		Trial division up to 2000

		ST=0	sTRIALPRIME	Factorization wanted

trialentry	GOSBVL	=PopASavptr

* First test if 0 <= Z <= 3

		D0=A			->Z
		D0=D0+	5
		A=DAT0	A		Zsize
		D0=D0+	5
		A=A-CON	A,7		CS if zero
		GOC	trialdiv<4
		A=A-1	A		CS if < 16
		GONC	trialdiv>3

		A=DAT0	P
		LC(1)	3
		?A>C	P
		GOYES	trialdiv>3	Regular division for > 3

* 0 <= Z <= 3 - return on stack: ( Z #1 Z1 )

trialdiv<4	?ST=1	sTRIALPRIME	If test only then push TRUE
		GOYES	+
		GOSBVL	=GETPTR		Unpop Z
		D1=D1-	5
		D=D-1	A
		GOSBVL	=SAVPTR
		LA(5)	=ONE
		GOSBVL	=GPPushA
		GOSBVL	=SAVPTR
		A=0	W
		A=A+1	A
		GOLONG	PushNLoop
+		GOVLNG	=GPPushTLoop


* Z > 3 - dispatch to short/long trial division based on size

trialdiv>3	D0=D0-	5		->Zsize
		A=DAT0	A		Zsize
		LC(5)	5+16		Sign is included in the comparison!
		?A<=C	A
		GOYES	trialdivshort
		GOTO	trialdivlong


**************************************************
**		SHORT TRIAL DIVISION		**
**************************************************

* R1[A] = trial division difference pointer
* R2[A] = current trial divider
* R3[W]	= Z
* R4[A] = number of pushed factors

* Note that PUSHhxs uses R1[W], thus it cannot be used.

trialdivshort	D0=D0+	5
		C=A	A		Zsize
		C=C-CON	A,7
		P=C	0
		A=0	W
		A=DAT0	WP		Z
		P=	0
		R3=A			R3[W] = Z

		GOSUB	InitTrial	Init pointer/divider
		
		A=0	A		Init #count to zero
		R4=A	A

* Try division

strialloop	A=R3	W		Z
		C=0	W
		C=R2.F	A		trial

		D=0	B		shifts = 0
		B=0	W		quotient = 0

		?C>=A	W
		GOYES	strialdivok	divisor >= Z - Z must be prime then

strialshift	C=C+C	W		divisor<<1
		D=D+1	B		shifts++
		?C<=A	W
		GOYES	strialshift	shift until divisor > Z
		D=D-1	B		shifts--
strialdivloop	CSRB			d >> 1
		B=B+B	W		quotient << 1
		?C>A	W
		GOYES	+
		A=A-C	W		remainder -= d
		B=B+1	P		quotient++
+		D=D-1	B		shifts--
		GONC	strialdivloop

strialdivok	?A=0	W
		GOYES	strialpush	remainder=0 - found factor

		A=0	W		If quotient < divisor then
		A=R2.F	A		we're done, the last Z is prime too
		?B<A	W
		GOYES	strialquit

		GOSUB	NextTrial
		GONC	strialloop	Continue until MAXTRIAL passed	

* MAXTRIAL exceeded - return ( z1 .. zk #k Z' )

		?ST=1	sTRIALPRIME	If test only then return TRUE
		GOYES	strialquit	(strialprime is too far)
		A=R4	A		number of factors
		R0=A	A
		GOSBVL	=PUSH#
		GOSBVL	=SAVPTR
		A=R3	W		Z'
		GOLONG	PushNLoop

* quotient reached zero - Z must be prime then  ( z1 .. Z #k Z1 )

strialquit	?ST=1	sTRIALPRIME	Return TRUE if test only
		GOYES	strialprime
		A=R3	W		Z'
		GOSUBL	PushN
		GOSBVL	=SAVPTR
		A=R4	A		number of factors
		A=A+1	A		including Z'
		R0=A
		GOSBVL	=PUSH#
		GOSBVL	=SAVPTR
		A=0	W		And push Z1 as the remaining factor
		A=A+1	A
		GOLONG	PushNLoop

* remainder was zero - trial divisor must be a factor
* update variables - push the factor - and continue with the same divisor

strialpush	?ST=1	sTRIALPRIME	If test only then return FALSE
		GOYES	strialcomposite
		A=B	W
		R3=A	W		Z = quotient		
		A=R4	A
		A=A+1	A		factors++
		R4=A	A
		A=0	W
		A=R2.F	A		trial divider
		GOSUBL	PushN
		GOSBVL	=SAVPTR
		GOTO	strialloop	Continue with the same factor

strialprime	GOVLNG	=GPPushTLoop
strialcomposite	GOVLNG	=GPPushFLoop

** Init trial division as follows:
**	R1[A] = trial division difference pointer
**	R2[A] = 1st trial divider

InitTrial	GOSUB	+
		NIBHEX	212240		2,3,5,7,11
+		C=RSTK
		R1=C			->difference pointer
		A=0	A
		R2=A	A		trial divider = 0

* Fall to NextTrial

**
** Get next trial division. CS if MAXTRIAL exceeded
** Note: Only 48 divisions are required to increase divisor by 210
**

NextTrial	A=R1	A			->difference pointer
		D0=A
		A=0	A
		A=DAT0	1			trial offset
		?A#0	P
		GOYES	++			offset is ok - continue
		GOSUB	+			Restart 2*3*5*7 wheel
		NIBHEX	242462642466264264684242
		NIBHEX	486462462664246264242A2A0
+		C=RSTK
		D0=C
		A=DAT0	1
++		D0=D0+	1
		CD0EX
		R1=C	A			->difference pointer'
		C=R2	A			current trial
		A=A+C	A
		R2=A	A			trial'
		LC(5)	MAXTRIAL
		?A>C	A
		RTNYES
		RTNCC

**************************************************
**		LONG TRIAL DIVISION		**
**************************************************

trialdivlong	GOSBVL	=GETPTR		Unpop Z in case GC occurs below
		D1=D1-	5
		D=D-1	A
		GOSBVL	=SAVPTR

* Allocate work memory:
*	Z once for copy
*	Z once for quotient
*	safety for remainder

		A=DAT1	A
		D0=A
		D0=D0+	5
		A=DAT0	A		Zsize
		A=A+A	A		2*Zsize
		A=A+A	A		4*Zsize
		LC(5)	400		Chicken
		C=C+A	A
		GOSBVL	=MAKE$N

* Copy Z into start of work memory

		A=DAT1	A		->Z
		AD0EX
		D1=A			->work memory
		D0=D0+	5
		C=DAT0	A
		D0=D0-	5
		C=C+CON	A,5
		GOSBVL	=MOVEDOWN

* Replace Z on stack with the work string

		A=R0			->$work
		GOSBVL	=D1=DSKTOP
		DAT1=A	A

* Now try quick trial division by 16. (not 2 for easy code..)

		D0=A
		D0=D0+	15		->Zsize
		D0=D0+	5		->mantissa
		A=0	A		div2 counter

trydiv2loop	C=DAT0	B
		D0=D0+	1
		?C#0	P
		GOYES	+		No more full zeros
		A=A+1	A		div2++
		GONC	trydiv2loop	This will terminate as Z <> 0
+		A=A+A	A
		A=A+A	A		*4 for 4-bit nibbles

* Now A[A] = n, where Z is divisible by 2^n

		R4=A	A		R4[A] = factors
		?A#0	A
		GOYES	+
		GOTO	ltrialnon2	None at all - continue

* Push Z2 to stack n times, then divide Z by 2^n (fast method)

+		?ST=0	sTRIALPRIME	If test only then return FALSE
		GOYES	+
		GOVLNG	=GPOverWrFLp
	
+		A=0	W		Push Z2 once first
		A=A+1	A
		A=A+1	A
		GOSUBL	PushN
		GOSBVL	=SAVPTR
		C=R4
		RSTK=C			n
trial2pushloop	A=DAT1	A		Swap Z2 and $work
		D1=D1+	5
		C=DAT1	A
		DAT1=A	A
		D1=D1-	5
		DAT1=C	A
		C=RSTK			Exit loop if pushed enough times
		C=C-1	A
		?C=0	A
		GOYES	trial2pushed
		RSTK=C
		GOSBVL	=GPPushA	Push ->Z2 again
		GOSBVL	=SAVPTR
		GONC	trial2pushloop
trial2pushed

* Divide Z by 2^n, where n = 4*m

		A=DAT1	A
		R0=A			->$work
		A=A+CON	A,10		->Z
		D0=A
		D1=A
		D1=D1+	10		->Zmantissa
		D0=D0+	5
		A=DAT0	A		Zsize
		C=R4.F	A		n
		CSRB.F	A
		CSRB.F	A		m
		A=A-C	A
		DAT0=A	A		Zsize'
		AD0EX
		A=A+C	A		->1st non-zero nibble
		AD0EX
		D0=D0+	5
		C=A	A
*		C=C-CON	A,5
		GOSBVL	=MOVEDOWN	Move down mantissa + sign

* See if done already

		GOSUB	InitTrial	Init for possible switch

		A=R0	A
		D0=A			->$work
		D0=D0+	10		->Z
		A=DAT0	11
		C=A	W
		LCZ1
		?A=C	W
		GOYES	ltrialexit

* See if small now

		D0=D0+	5		->Zsize
		A=DAT0	A
		LC(5)	5+16		Sign is included in the comparison
		?A>=C	A
		GOYES	+
		GOTO	switchtoshort	Switch to short loop
+

* Now trial division by 16 has been done, start the 210-wheel
* Note that the trial divisor is always less than 8 nibbles, thus
* we can call the specialized long/short subroutine inside ZQUO to save
* speed. Naturally this saves registers too, as the algorithm is quite
* simple:
*	ZQUOint:
*	In:	D0	->Z1size
*		B[W]	divisor
*		D[S]	divisor sign
*		R0[A]	->Output
*	Out:	D0	->Z1tail
*		B[W]	remainder
*		D[S]	remainder sign
*		R0[A]	->Output
*	Alters:	A[W] B[W] C[W] D[A]

* R0[A] = ->Ztail (->output)
* R1[A] = trial division difference pointer
* R2[A] = current trial divider
* R3[A] = ->$work
* R4[A] = number of pushed factors

ltrialnon2	A=R0	A		->$work
		R3=A	A		->$work
		D0=A
		D0=D0+	15		->Zlen
		C=DAT0	A
		AD0EX
		A=A+C	A		->Ztail
		R0=A	A		->Output

		GOSUB	InitTrial

* Now start the trial loop.

ltrialloop	A=R3	A		->$work
		D0=A
		D0=D0+	15		->Zsize
		A=0	W
		A=R2.F	A
		B=A	W		trial
		D=0	S		positive
		GOSUBL	ZQUOint

		?B=0	W
		GOYES	ltrialpush	remainder = 0 - push divisor

* No need to test quotient - it is known to be big

		GOSUB	NextTrial
		GONC	ltrialloop	Continue until MAXTRIAL passed	

* MAXTRIAL exceeded - return ( z1 .. zk #k Z' )

ltrialexit	?ST=1	sTRIALPRIME	If test then return TRUE
		GOYES	ltrialprime
		A=R4	A		number of factors
		R0=A	A
		GOSBVL	=PUSH#
		GOSBVL	=SAVPTR

		A=DAT1	A		Swap $work and #k
		D1=D1+	5
		C=DAT1	A
		DAT1=A	A
		D1=D1-	5
		R0=C			->$work
		C=C+CON	A,10		->Z (we leave the Z inside the $)
		DAT1=C	A		->Z
		D0=C			->Z
		GOSBVL	=SKIPOB		->Ztail
		GOSUB	ShrinkZ		Remove extra nibbles from R0-D0
		GOVLNG	=GETPTRLOOP

ltrialprime	GOVLNG	=GPOverWrTLp
ltrialcomposite	GOVLNG	=GPOverWrFLp

* remainder was zero - trial divisor must be a factor
* update variables - push the factor - and continue with the same divisor

ltrialpush	?ST=1	sTRIALPRIME	If test only then return FALSE
		GOYES	ltrialcomposite
		A=R4	A
		A=A+1	A		factors++
		R4=A	A
		A=0	W
		A=R2.F	A		trial divider
		GOSUBL	PushN
		GOSBVL	=SAVPTR
		A=DAT1	A		Swap $work and factor
		D1=D1+	5
		C=DAT1	A
		DAT1=A	A
		D1=D1-	5
		DAT1=C	A
		R3=C			->$work (update in case of GC)

* Move the quotient down to replace old Z

		D1=C
		D1=D1+	10		->target
		D0=C
		D0=D0+	15		->Zsize
		C=DAT0	A
		AD0EX
		A=A+C	A
		D0=A			->quotient
		D0=D0+	5
		C=DAT0	A		quotient size
		C=C+CON	A,5
		D0=D0-	5
		GOSBVL	=MOVEDOWN

* Z has been updated. Update ->Ztail
		CD1EX
		R0=C	A		->Ztail
* Check Zsize to see whether to continue with short or long integers

		A=R3			->$work
		D0=A
		D0=D0+	15		->Zsize
		A=DAT0	A
		LC(5)	5+16		Sign is included in the comparison
		?A<=C	A
		GOYES	switchtoshort
		GOTO	ltrialloop	Continue with the same divisor

* Switch to short trial division loop as follows:
*	Pop Z into R3[W] as used by strialloop
*	Drop $work from the stack
*	Jump to strialloop

switchtoshort	C=A	A
		C=C-CON	A,7		Quotient cannot be zero here
		P=C	0
		D0=D0+	5
		A=0	W
		A=DAT0	WP
		P=	0
		R3=A
		GOSBVL	=GETPTR
		D1=D1+	5		Drop $work
		D=D+1	A
		GOSBVL	=SAVPTR
		GOTO	strialloop	Continue short loop

**********************************************************************
* Substitute for Shrink$ which trashes all objects above $ in tempob.
* Input:	R0[A] = ->Z
*		D0    = ->$end (wanted)
**********************************************************************
ShrinkZ		C=R0	A		->Z
		C=C+CON	A,5
		D1=C
		A=DAT1	A		Zsize (old)
		A=A+C	A
		A=A+CON	A,5
		B=A	A		->Ztail (old)
		AD0EX
		D0=A
		C=A-C	A		Zsize (new)
		DAT1=C	A
		C=R0	A
		A=A-C	A
		A=A+CON	A,6
		DAT0=A	A		Set new link field
		D0=D0+	5
		CD0EX
		C=C-B	A		Removed nibbles
		?C=0	A
		RTNYES
		C=-C	A		Adjust ptrs downwards
		A=B	A		->Ztail (old)
		GOVLNG	=MOVERSD
ENDCODE
**********************************************************************

